/* Copyright (C) 2006 by John Welch
 * Copyright (C) 2004 by Ted Phelps */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef STDC_HEADERS
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_VALUES_H
#include <values.h>
#endif
#include "mcmc.h"

/* Randomly reorder the integers in the array */
static void
shuffle(int *values, int value_count)
{
    int choice, i, temp;

    for (i = 0; i < value_count - 1; i++) {
        /* Choose one of the unshuffled elements */
        choice = lrand48() % (value_count - i);

        /* Swap it with the value at i */
        temp = values[i];
        values[i] = values[i + choice];
        values[i + choice] = temp;
    }
}

/* See mcmc.h for documentation of this function and its arguments */
int mcmc(double t0, double dt, int tscale, long max_rounds, 
          int param_count, double *params,
          double *mins, double *maxs,
          double *step_sizes, int *indices,
          mcmc_log_likelihood_func_t likelihood_func, void *likelihood_rock)
{
    double old_likelihood, likelihood;
    double old_param;
    long round=0, i;
    int last_used = 1;
	double t = t0; // initial temperature
	int keep_going = 1;
	double small_number = 1e-9;
	double saved_likelihood = 0;
	int ctr = 0;
					
	/* Make sure each integer between 0 and param_count - 1 is represented */
    if (indices) {
        for (i = 0; i < param_count; i++) {
            indices[i] = i;
        }

        /* Shuffle them now */
        shuffle(indices, param_count);
    }

    /* Compute the likelihood of the initial parameters */
    old_likelihood = likelihood_func(-1, -1, param_count, params, likelihood_rock);
    
    while (1) {
        
		/* Decide which parameter to change */
        if (indices) {
            i = indices[round % param_count];
        } else {
            i = round % param_count;
		}
		
		if ((round+1) % (tscale*param_count) == 0)
		{
			t = (t < 1e-4) ? 0.0 : t * dt;
			printf("round:%d\ttemp:%f\tlnL:%f\tp:%f\n",round/param_count,t,old_likelihood,params[param_count-1]);
			
			if( t == 0.0 && abs(saved_likelihood-old_likelihood) < small_number )
				ctr++;
			if(ctr > 4)
				return 1;
			saved_likelihood = old_likelihood;
		}
		
        /* Skip bound parameters */
        if (mins[i] == maxs[i]) {
    		round++;
			continue;
        }

        /* Hang on to the old value of the parameter */
        old_param = params[i];

        /* Guess a new value for our parameter, taking care not to go outside limits */
        do {
            params[i] = old_param + step_sizes[i] * (2.0*drand48() - 1.0);
		} while (params[i] < mins[i] || params[i] > maxs[i]);
		       
		/* Compute the likelihood with this new value */
		likelihood = likelihood_func(i, last_used, param_count, params, likelihood_rock);

		/* If the likelihood improves or doesn't get worse by some amount relative to the temperature, 
		 * then keep this value for the parameter.  Otherwise we revert to the value */
		if (likelihood > old_likelihood ||
			likelihood > old_likelihood + t * log( (1.0 - drand48()) ) ) {  
			/* It's good enough so we'll keep it. */
			last_used = 1;
			old_likelihood = likelihood;

			/* Broaden our step size, but make sure it doesn't go infinite. */
			step_sizes[i] = step_sizes[i] / 0.95 < maxs[i] ? step_sizes[i] / 0.95 : maxs[i];
		} else {
			last_used = 0;

			/* Revert to the old param. */
			params[i] = old_param;

			/* Narrow our step size, but make sure it doesn't go to zero. */
			step_sizes[i] = step_sizes[i] * 0.95 < DBL_MIN ? DBL_MIN : step_sizes[i] * 0.95;
		}
		
		/* Have we tweaked all of the parameters */
        if ((round + 1) % param_count == 0 ) {
            if (indices) {
                /* Shuffle the indices */
                shuffle(indices, param_count);
            }

        }
		round++;
		if(round > max_rounds) {
			return -1;
		}
	}
}