/* v2 Copyright (C) 2008 by John Welch
 * v1 Copyright (C) 2006 by John Welch
 * Copyright (C) 2004 by Ted Phelps 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000 Gerard Jungmam */

/* The MK test functionality was written by John Welch containing altered portions 
 * of code by Ted Phelps and G. Jungman from the Gnu Scientific Library  
 * It is partially based on Mathematica code by Nicholas Bierne and Adam Eyre-Walker */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef STDC_HEADERS
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_FLOAT_H
#include <float.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_VALUES_H
#include <values.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#include "freadnums.h"
#include "mcmc.h"

#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_sf_hyperg.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_rng.h>

#include <gsl/gsl_sf_exp.h>
#include <gsl/gsl_sf_bessel.h>
#include <gsl/gsl_sf_laguerre.h>
#include <gsl/gsl_sf_pow_int.h>
#include <gsl/gsl_sf_psi.h>
#include <gsl/gsl_integration.h>

#include <gsl/gsl_sf_elementary.h>

#define INT_THRESHOLD (1000.0*GSL_DBL_EPSILON)
#define _1F1_INT_THRESHOLD (100.0*GSL_DBL_EPSILON)

/* This is error.h required by the altered GNU rountines */
#define OVERFLOW_ERROR(result) do { (result)->val = GSL_POSINF; (result)->err = GSL_POSINF; GSL_ERROR ("overflow", GSL_EOVRFLW); } while(0)
#define UNDERFLOW_ERROR(result) do { (result)->val = 0.0; (result)->err = GSL_DBL_MIN; GSL_ERROR ("underflow", GSL_EUNDRFLW); } while(0)
#define INTERNAL_OVERFLOW_ERROR(result) do { (result)->val = GSL_POSINF; (result)->err = GSL_POSINF; return GSL_EOVRFLW; } while(0)
#define INTERNAL_UNDERFLOW_ERROR(result) do { (result)->val = 0.0; (result)->err = GSL_DBL_MIN; return GSL_EUNDRFLW; } while(0)
#define DOMAIN_ERROR(result) do { (result)->val = GSL_NAN; (result)->err = GSL_NAN; GSL_ERROR ("domain error", GSL_EDOM); } while(0)
#define DOMAIN_ERROR_MSG(msg, result) do { (result)->val = GSL_NAN; (result)->err = GSL_NAN; GSL_ERROR ((msg), GSL_EDOM); } while(0)
#define DOMAIN_ERROR_E10(result) do { (result)->val = GSL_NAN; (result)->err = GSL_NAN; (result)->e10 = 0 ; GSL_ERROR ("domain error", GSL_EDOM); } while(0)
#define OVERFLOW_ERROR_E10(result) do { (result)->val = GSL_POSINF; (result)->err = GSL_POSINF; (result)->e10 = 0; GSL_ERROR ("overflow", GSL_EOVRFLW); } while(0)
#define UNDERFLOW_ERROR_E10(result) do { (result)->val = 0.0; (result)->err = GSL_DBL_MIN; (result)->e10 = 0; GSL_ERROR ("underflow", GSL_EUNDRFLW); } while(0)
#define OVERFLOW_ERROR_2(r1,r2) do { (r1)->val = GSL_POSINF; (r1)->err = GSL_POSINF; (r2)->val = GSL_POSINF ; (r2)->err=GSL_POSINF; GSL_ERROR ("overflow", GSL_EOVRFLW); } while(0)
#define UNDERFLOW_ERROR_2(r1,r2) do { (r1)->val = 0.0; (r1)->err = GSL_DBL_MIN; (r2)->val = 0.0 ; (r2)->err = GSL_DBL_MIN; GSL_ERROR ("underflow", GSL_EUNDRFLW); } while(0)
#define DOMAIN_ERROR_2(r1,r2) do { (r1)->val = GSL_NAN; (r1)->err = GSL_NAN;  (r2)->val = GSL_NAN; (r2)->err = GSL_NAN;  GSL_ERROR ("domain error", GSL_EDOM); } while(0)

#define SERIES_EVAL_OK(a,b,x) ((fabs(a) < 5 && b < 5 && x < 2.0) || (fabs(a) <  10 && b < 10 && x < 1.0))
#define ASYMP_EVAL_OK(a,b,x) (GSL_MAX_DBL(fabs(a),1.0)*GSL_MAX_DBL(fabs(1.0+a-b),1.0) < 0.99*fabs(x))

/* The Monte Carlo Markoff Chain (MCMC) algorithm tweaks parameters
 * and computes the likelihood across a set of data points.  
 * Our MCMC function only needs to know how many parameters we have
 * so they're stored in an array and referenced by index. */  

/* Remember the program name */
static char *progname = NULL;

/* These are flags for the four terms in the likelihood function 
   each of which relates to a different item of data */
#define NUM_TERMS 4
#define P_Dn 0
#define P_Pn 1
#define P_Di 2
#define P_Pi 3

/* These are flags for the four integrals we require when the Sawer-Hartl 
	method. These relate to the expected non-synonymous divergence and 
	polymorphism resulting from -ve and +ve selection coefficients  */
#define NUM_INTEGRALS 3
#define I_Dn_d1 0
#define I_Dn_d2 1
#define I_Pn_d 2

/* These are flags for the types of parameter. Note that omega has been omitted here */
#define NUM_PARAM_TYPES 12
#define P_THETA 0
#define P_TIME 1
#define P_OMEGA 2
#define P_GAMMA_D 3
#define P_ALPHA 4
#define P_BETA_A 5
#define P_BETA_B 6
#define P_DELTA_A0 7
#define P_DELTA_A1 8
#define P_DELTA_P 9
#define P_REG_INTERCEPT 10
#define P_REG_SLOPE 11
char * param_text[] = {"theta","mu_t","f","gamma","alpha","Beta(a)","Beta(b)",
						"Delta(a0)","Delta(a1)","Delta(q)","intercept","slope"};

/* These are flags for the various alpha models */
#define ALPHA_ZERO 0
#define ALPHA_FREE 1
#define ALPHA_BETA 2
#define ALPHA_DELTA 3
#define ALPHA_MANY 4
#define ALPHA_FULL 5
char * alpha_text[] = {"fixed at zero","universal parameter","Beta distributed random variable",
							"Two-spike distributed random variable", "value per class","locus specific values"};

/* These are flags for the various non-alpha models */
#define MODEL_NONE 0
#define MODEL_UNIVERSAL 1
#define MODEL_LOCUSSPECIFIC 2
#define MODEL_CLASSSPECIFIC 3
#define MODEL_XAUTOSOMAL 4
#define MODEL_REGRESSION 5
char * model_text[] = {"not included","universal parameter","locus specific values",
			"one value per class","separate X and autosomal values","linear function of recombination rate"};

typedef struct item *item_t;
struct item
{
    /* The data items: Dn, Pn, Di, Pi */
    int measured[NUM_TERMS];

    /* Number of sites: Dn, Pn, Di, Pi */
    double sites[NUM_TERMS];

	/* The number of alleles sampled */
	int alleles;

	/* Which chromosome is the gene on? */
	int chromosome;
	
	/* An arbitrary flag - refers to index of "flag_list" and used to reference "params"*/
	int flag;
	
	/* Recombination rate */
	double r;
	
    /* The cached item likelihood values: Dn, Pn, Di, Pi  */
    double likelihood[NUM_TERMS];

    /* The old likelihood value: Dn, Pn, Di, Pi */
    double old_likelihood[NUM_TERMS];
	
	/* The cached values related to E[Dn] and E[Pn] */
    double selected_k[NUM_INTEGRALS];

    /* The old values related to E[Dn] and E[Pn] */
    double old_selected_k[NUM_INTEGRALS];
};

/* contains info about the likelihood terms that change when a particular param changes */
struct param_info
{
    /* 0..num_genes-1 if a single gene is affected, -1 for a parameter common to all genes */
    int which_gene; 
	
	/* what kind of parameter is it, i.e. THETA, ALPHA etc. */
	int what_param;
	
	/* which likelihood terms are affected (0 Dn, 1 Pn, 2 Di, 3 Pi) */ 
	int which_terms[NUM_TERMS+1];
};

/* Our data are a set of items */
struct dataset
{
    /* The info about the parameter to be tweaked */
    struct param_info last_param_info;
	
	/* The info about the likelihood models for the five parameter types */
	int models[5];
		
    /* The number of flags applied to diff classes of locus */
    int flag_count;

	/* The flags themselves */
	int * flag_list;

    /* The number of data items */
    int count;

    /* The items themselves */
    struct item *items;
};

/* Used to pass to the numerical integration routine */
struct integrand_params
{
    /* The number of alleles sampled*/
    int n;

    /* 4*Ne*E[|s|] */
    double gamma;
};

/* Used to pass to Kummer's U numerical integration routine */
/*struct KummerU_integrand_params
{
	double a;
	double b;
	double c;
};*/

/* A table mapping from i to log(i!) to speed the computation of log poisson and Beta integral */
static double *log_factorial;

/* Tables mapping sum(1/i,i=1..n-1) and (1+1/n) */
static double *neutral_P_k;
static double *neutral_D_k1;

// For selected Polymorphism integral
gsl_integration_workspace * w;
gsl_function FPn;
gsl_function FDn;
struct integrand_params i_params;

/* Variables used if gamma_d is universal */
double ** old_univ_Pn_d_k; 
double ** univ_Pn_d_k;
double ** old_univ_Dn_d_k1; 
double ** univ_Dn_d_k1;
double old_univ_Dn_d_k2[2];
double univ_Dn_d_k2[2];
int n_values_count = 0;
int * n_values;

/* the number of parameters */
int param_count = 0; 

/* Global flags */
/* BEW=0 includes sampling corrections, BEW=1 excludes them (like Bierne and Eyre-Walker 2004) */
int BEW;
/* A_NOT_ALPHA=0 calculate adaptive subs/site rather than alpha */
int A_NOT_ALPHA;
/* Flag used if results for a single locus are required */
int single_gene = 0;
/* The number of times that the Kummer's U function crashes out */
static int approx_ctr = 0;
/* Whether the dist of selection coefficients is reflected or not */
int M_DIST_IS_REFLECTED;
/* Shape parameter for the distribution (not implemented) */
int M_BETA_IS_ONE;
/* m=1 for pairs of lineages, m=2 for div. along a single lineage */
int M_CORRECTION_COEFF; 

void disp_to_screen( int no_conv, double lnL, int data_count, int param_count, double * params, int alpha_status, int * flag_list, int flag_count) {

	int i;
	double temp;
	
	printf("\n  done; log(likelihood) = %0.12g", lnL);
	if(no_conv < 0)
		printf(", but with possible convergence problems");
	printf("\n  AIC  = %0.12g", -2.0*lnL+2.0*(double)param_count);
	printf("\n  AICc = %0.12g", -2.0*lnL+2.0*(double)param_count+
		2.0*(double)param_count*((double)param_count+1.0)/(4.0*(double)(data_count)-(double)param_count-1.0));
	printf("\n  BIC  = %0.12g\n\n", -2.0*lnL+log(4.0*(double)(data_count))*(double)param_count);
	
	switch(alpha_status) {
		case ALPHA_FREE:
			printf("alpha = %0.4g\n\n",params[param_count-1]);	
			break;
		case ALPHA_BETA:
			temp = 	params[param_count-2]*params[param_count-1]/(params[param_count-2]+params[param_count-1])/
				(params[param_count-2]+params[param_count-1])/(params[param_count-2]+params[param_count-1]+1.0);	
			printf("a = %0.4g, b = %0.4g (E[alpha] = %0.4g, Var[alpha] = %0.4g)\n\n",params[param_count-2],params[param_count-1],
				params[param_count-2]/(params[param_count-2]+params[param_count-1]),temp);
			break;
		case ALPHA_DELTA:
			temp = (1.0-params[param_count-1])*params[param_count-1]*(params[param_count-3]-params[param_count-2])*(params[param_count-3]-params[param_count-2]); 
			printf("a0 = %0.4g, a1 = %0.4g, q = %0.4g (E[alpha] = %0.4g, Var[alpha] = %0.4g)\n\n",
				params[param_count-3],params[param_count-2],params[param_count-1],params[param_count-3]*params[param_count-1]+params[param_count-2]*(1.0-params[param_count-1]),temp);	
			break;
		case ALPHA_MANY:
			printf("alpha values for classes:\n");
			for(i=0; i<flag_count;i++) {
				printf("%d\t",flag_list[i]);
			}
			printf("\n");
			for(i=0; i<flag_count;i++) {
				printf("%0.4g\t",params[param_count-flag_count+i]);
			}
			printf("\n\n");
			break;
		case ALPHA_FULL:
			temp = 0.0;
			for(i=0; i<data_count;i++) {
				temp += params[param_count - data_count + i];
			}
			temp /= (double)data_count;
			printf("Mean[alpha] = %0.4g\n\n",temp);
			break;
	}		

}

/* Sort bootstrapped values for calculating confidence intervals */
static void 
boot_sort(double boot[], int left, int right) 
{
	int i, last;
	void swap(double boot[], int i, int j);
	
	if (left >= right)
		return;
	swap(boot, left, (left+right)/2);
	last = left;
	for (i = left+1; i <= right; i++)
		if (boot[i] < boot[left])
			swap(boot, ++last, i);
	swap(boot, left, last);
	boot_sort(boot, left, last-1);
	boot_sort(boot, last+1, right);	
}

void swap(double boot[], int i, int j)
{
	double temp;
	temp = boot[i];
	boot[i] = boot[j];
	boot[j] = temp;	
}

/* Calculate Smith and Eyre-Walker's and Fay, Wyckoff and Wu's alpha_bar and Mantel-Haenszel's z */
static void 
do_Smith_Eyre_Walker( struct dataset * data )
{
	int i, j, index;
	
	double dDn, dDi, dPn, dPi;
		
	double Dn_bar  = 0.0; 
	double Di_bar  = 0.0;
	double Pn_bar  = 0.0; 
	double Pi_bar  = 0.0;
	
	double LDn_bar = 0.0; 
	double LDi_bar = 0.0;
	double LPn_bar = 0.0;	    
	double LPi_bar = 0.0;
	
	double z_bar   = 0.0;
	
	double sum_d_m_delta = 0.0;
	double sum_V = 0.0;
	
	double c2,r2,N;
	/* alpha_bar estimates of Smith/Eyre-Walker and Fay/Wykcoff/Wu, and Mantel-Haenszel's z */
	double sew, fww, mhz;  
	double sew_boot[10000]; /* for bootstrap values */
	double fww_boot[10000]; /* for bootstrap values */
	
	for ( i=0; i < data->count; i++ ) {
		
		dDn = (double)(data->items[i].measured[P_Dn]);
		dDi = (double)(data->items[i].measured[P_Di]);
		dPn = (double)(data->items[i].measured[P_Pn]);
		dPi = (double)(data->items[i].measured[P_Pi]);
		
		Dn_bar  += dDn;
		Di_bar  += dDi;
		Pn_bar  += dPn;
		Pi_bar  += dPi;
				
		LDn_bar += data->items[i].sites[P_Dn];
		LDi_bar += data->items[i].sites[P_Di];
		LPn_bar += data->items[i].sites[P_Pn];
		LPi_bar += data->items[i].sites[P_Pi];
		
		/* With site corrections*/
		/*z_bar   += dPn * data->items[i].sites[P_Pi] / ( dPi + 1.0 );*/
		/* Without site corrections, as published */
		z_bar   += dPn / ( dPi + 1.0 );
	
		/* For Mantel-Haenszel z statistic */
		r2 = dDn + dPn;
		c2 = dDn + dDi;
		N  = c2 + dPn + dPi;
		if ( N > 1.0 ) {
			sum_d_m_delta   += dDn - r2*c2/N;
			sum_V			+= c2 * r2 * (dPi + dPn) * (dPi + dDi) / N / N / (N - 1.0);
		}
	}
	/* Smith Eyre-Walker */ 
	/*with corrections for differences in sites*/
	/*sew = 1.0 - Di_bar * LDn_bar / Dn_bar / LDi_bar * z_bar / LPn_bar;*/
	/* Without site corrections, as published */
	sew = 1.0 - Di_bar / Dn_bar * z_bar / data->count;
		
	/* Fay, Wyckoff and Wu */
	fww = 1.0 - Di_bar * LDn_bar / Dn_bar / LDi_bar * Pn_bar * LPi_bar  / Pi_bar / LPn_bar;
	
	mhz = (fabs( sum_d_m_delta ) - 0.5) / sqrt( sum_V );
	/* For Mantel-Haenszel z statistic */
	printf("\nMHz: %0.6g\n", mhz ); 
			
	/* bootstrap */
	for ( j=0; j < 10000; j++ ) {
		Dn_bar  = 0.0;
		Di_bar  = 0.0;
		Pn_bar  = 0.0;
		Pi_bar  = 0.0;
		
		LDn_bar = 0.0;
		LDi_bar = 0.0;
		LPn_bar = 0.0;
		LPi_bar = 0.0;
		
		z_bar   = 0.0;
		
		for ( i=0; i < data->count; i++ ) {
			index = floor( drand48() * data->count ); /* randomly chosen data item */
			
			Dn_bar  += (double)(data->items[index].measured[P_Dn]);
			Di_bar  += (double)(data->items[index].measured[P_Di]);
			Pn_bar  += (double)(data->items[index].measured[P_Pn]);
			Pi_bar  += (double)(data->items[index].measured[P_Pi]);
						
			LDn_bar += data->items[index].sites[P_Dn];
			LDi_bar += data->items[index].sites[P_Di];
			LPn_bar += data->items[index].sites[P_Pn];
			LPi_bar += data->items[index].sites[P_Pi];
			
			/* Site corrections */
			/*z_bar   += (double)(data->items[index].measured[P_Pn]) * data->items[index].sites[P_Pi] /
						( (double)(data->items[index].measured[P_Pi]) + 1.0 );*/
			z_bar   += (double)(data->items[index].measured[P_Pn]) / ( (double)(data->items[index].measured[P_Pi]) + 1.0 );
		}	
		/*sew_boot[j] = 1.0 - Di_bar * LDn_bar / Dn_bar / LDi_bar * z_bar / LPn_bar;*/
		sew_boot[j] = 1.0 - Di_bar / Dn_bar * z_bar / data->count;
		fww_boot[j] = 1.0 - Di_bar * LDn_bar / Dn_bar / LDi_bar * Pn_bar * LPi_bar  / Pi_bar / LPn_bar;
	}
	/* 95% confidence intervals and proportion < 0 for Smith Eyre-Walker */
	boot_sort(sew_boot, 0, 9999);
	i=0;
	while(sew_boot[i]<0.0)
		i++;
	printf("SEW: %0.5g\t(%0.5g, %0.5g)\t[%0.5g]\n",sew,sew_boot[250],sew_boot[9749],0.001*i);
	
	/* 95% confidence intervals and proportion < 0 for Fay, Wyckoff and Wu */
	boot_sort(fww_boot, 0, 9999);
	i=0;
	while(fww_boot[i]<0.0)
		i++;
	printf("FWW: %0.5g\t(%0.5g, %0.5g)\t[%0.5g]\n",fww, fww_boot[250],fww_boot[9749],0.001*i);
} 

/* how many parameters are associated with alpha */
int get_nalphap(int alpha_model, int flag_count, int num_genes) {
	switch (alpha_model) {
		case ALPHA_ZERO:
			return 0;
		case ALPHA_FREE:
			return 1;
		case ALPHA_BETA:
			return 2;
		case ALPHA_DELTA:
			return 3;
		case ALPHA_MANY:
			return flag_count;
		case ALPHA_FULL:
			return num_genes;
		default:
			return -4*num_genes;
	}
}
/* how many parameters are associated with theta and time */
int get_nnonalphap(int model_num, int flag_count, int num_genes ) {
	switch (model_num) {
		case MODEL_NONE:
			return 0;
		case MODEL_UNIVERSAL:
			return 1;
		case MODEL_LOCUSSPECIFIC:
			return num_genes;
		case MODEL_CLASSSPECIFIC:
			return flag_count;
		case MODEL_XAUTOSOMAL:
		case MODEL_REGRESSION:
			return 2;
		default:
			return -4*num_genes;
	}
}

/* how many parameters does each model have? */
int get_np(int * models, int num_genes, int flag_count)
{
	int np;
	
	np  = get_nnonalphap( models[P_THETA],   flag_count, num_genes );
	np += get_nnonalphap( models[P_TIME],    flag_count, num_genes );
	np += get_nnonalphap( models[P_OMEGA],   flag_count, num_genes );
	np += get_nnonalphap( models[P_GAMMA_D], flag_count, num_genes );
	np += get_nalphap(	  models[P_ALPHA],   flag_count, num_genes );
	
	return np;
}
	
/* the log of a Poisson pdf with argument "expected"  */
static double
calc_log_poisson(double expected, int observed)
{
   	if(expected > 0.0) {
		return -expected + observed * log(expected) - log_factorial[observed];
	} else if (expected == 0.0 && observed == 0) {
		return 0;
	} else {
	 	//fprintf(stderr,"\nI am HERE\n");
		return GSL_NEGINF;
	}
	//return -expected + observed * log(expected) - log_factorial[observed];
}

/* Calculate the log of the following integral: Integrate( P(k;c/(1-x)) { p Dirac(x-a0) + (1-p) Dirac(x-a1) } , x=0..1 ) 
 * where P is a Poisson distribution 
 * this is the LnL for Dn under when the proportion of adaptive substitutions is distributed with two weighted spikes*/
static double
calc_log_delta_integral(double a0, double a1, double p, double c, int k) {
	double LnL;						/* the Log likelihood value */
	double lnpre1_val, lnpre2_val, max_lnpre_val;
	double t1, t2, lp1, lp2;
	
	LnL = k * log(c) - log_factorial[k];
	
	lnpre1_val = -c/(1.0-a0) - k*log(1.0-a0);
	lnpre2_val = -c/(1.0-a1) - k*log(1.0-a1);
	max_lnpre_val = GSL_MAX(lnpre1_val, lnpre2_val);
	lp1 = lnpre1_val - max_lnpre_val;
	lp2 = lnpre2_val - max_lnpre_val;
	t1  = p*exp(lp1);
	t2  = (1.0-p)*exp(lp2);
			
	LnL+= log(t1+t2) + max_lnpre_val;
	return LnL;
}

/* Calculate the log of the following integral: Integrate( P(k;c/(1-x)) B(x;a,b) , x=0..1 ) 
 * where P is a Poisson distribution, and B is a Beta distribution 
 * this is the LnL for Dn under when the proportion of adaptive substitutions is Beta-distributed */
static double
calc_log_beta_integral(double a, double b, double c, int int_k)
{
	gsl_sf_result_e10	U_r; /* result structure for Kummer's U */
	gsl_sf_result 	lnG_r;	 /* result structure for lnGammas */
    int lnG_status;			 /* Status flag for lnGamma calculations */
	int U_status;			 /* Status flag for Kummer's U calculation */
	double a_arg, b_arg;	 /* Arguments for Kummer's U */
	double LnL;				 /* the Log likelihood value */
	double k = (double)int_k;
	
	double crap, t1, t2;
	double lnG_sgn;
	int t1e10, t2e10, te10;
	
	LnL = - c - log_factorial[int_k];
	
	lnG_status = gsl_sf_lngamma_e(a + b, &lnG_r);
	if(lnG_status) {
		fprintf(stderr,"internal error: gnu lnGamma: %s (a=%g; b=%g)", gsl_strerror(lnG_status),a,b);
		exit(1);
	}
	LnL += lnG_r.val;
	
	lnG_status = gsl_sf_lngamma_e(b, &lnG_r);
	if(lnG_status) {
		fprintf(stderr,"internal error: gnu lnGamma: %s (a=%g)", gsl_strerror(lnG_status),a);
		exit(1);
	}
	LnL -= lnG_r.val;
	
	/* Kummer transform if necessary for greater accuracy */
	if ( b > k ) { /* transform to ensure second argument is non-negative */
		a_arg = b - k + a ;
		b_arg = b - k + 1.0;
		LnL += b * log(c);
	}
	else {
		a_arg = a;
		b_arg = k - b + 1.0;
		LnL += k * log(c);
	}
	
	/* Calculate Kummer's U function */
	U_status = jsl_sf_hyperg_U_e10_e(a_arg, b_arg, c, &U_r);
	if ( U_status==GSL_SUCCESS && gsl_finite(U_r.val) && U_r.val!=0 ) {
		LnL += log( U_r.val ) + (double)U_r.e10 * M_LN10;
	} 
	else { // If the function crashes, use an approximate form of Kummer's U
		approx_ctr++;
		
		if (b_arg > 2.75 && c < 0.00001 ) {
			
			LnL += (1.0 - b_arg) * log(c);
			crap = (1.0 - b_arg) * log(c);
			lnG_status = gsl_sf_lngamma_e(b_arg-1.0, &lnG_r);
			
			if(lnG_status) {
				fprintf(stderr,"internal error: gnu lnGamma(b-1) %s (b=%g)", gsl_strerror(lnG_status),b_arg);
				exit(1);
			}
			LnL += lnG_r.val;
			crap += lnG_r.val; 
			lnG_status = gsl_sf_lngamma_e(a_arg, &lnG_r);
			if(lnG_status) {
				fprintf(stderr,"internal error: gnu lnGamma(a): %s (a=%g)", gsl_strerror(lnG_status),a_arg);
				exit(1);
			}
			LnL -= lnG_r.val;	
			crap -= lnG_r.val;
			//fprintf(stderr,"\nCrude approximation of Ln(U(%g,%g,%0.12g)) = %g",a_arg,b_arg,c,crap);
		} else {
			lnG_status = gsl_sf_lngamma_e(b_arg-1.0, &lnG_r);
			if(lnG_status) {
				fprintf(stderr,"internal error: gnu lnGamma(b-1): %s (b=%g)", gsl_strerror(lnG_status),b_arg);
				exit(1);
			}
			crap = lnG_r.val;
			
			lnG_status = gsl_sf_lngamma_e(a_arg, &lnG_r);
			if(lnG_status) {
				fprintf(stderr,"internal error: gnu lnGamma(a): %s (a=%g)", gsl_strerror(lnG_status),a_arg);
				exit(1);
			}
			crap -= lnG_r.val;
			
			U_status = gsl_sf_exp_mult_e10_e( crap+(1.0-b_arg)*log(c), 1.0+(1.0+a_arg-b_arg)*c/(2.0-b_arg),&U_r);
			t1 = U_r.val;
			t1e10 = U_r.e10;
						
			lnG_status = gsl_sf_lngamma_sgn_e( 1.0 - b_arg , &lnG_r, &lnG_sgn); 
			if(lnG_status) {
				fprintf(stderr,"internal error: gnu lnGamma(1-b): %s (b=%g)", gsl_strerror(lnG_status),b_arg);
				exit(1);
			}
			crap = lnG_r.val;
			
			lnG_status = gsl_sf_lngamma_e( 1.0 - b_arg + a_arg, &lnG_r);
			if(lnG_status) {
				fprintf(stderr,"internal error: gnu lnGamma(1-b+a): %s (a=%g b=%g)", gsl_strerror(lnG_status),a_arg,b_arg);
				exit(1);
			}
			crap -= lnG_r.val;
			
			U_status = gsl_sf_exp_mult_e10_e( crap , lnG_sgn * (1.0 + a_arg*c/b_arg), &U_r); 
			t2 = U_r.val;
			t2e10 = U_r.e10;
			
			if ( t1e10 == 0 && t2e10 == 0) {
				te10 = 0;
			} else {
				te10 = (t1e10 < t2e10) ? t1e10: t2e10;
				t1 *= gsl_pow_int(10.0,t1e10 - te10);
				t2 *= gsl_pow_int(10.0,t2e10 - te10);
			}
							
			LnL += log( t1 + t2 ) + (double)te10 * M_LN10;
		}
	}
	return LnL;
}

/* Functions for numerical integration */
double 
Pn_d_k_integrand_exp_shift( double x, void * params )
{
	struct integrand_params * iparams;
	double psi, f;
	
	iparams = (struct integrand_params *)params;
	
	psi = gsl_sf_psi( x + 1.0/iparams->gamma );
	/* Should I add lines for x~0 or x~1 ?? */
	f = psi * ( 1.0 - gsl_sf_pow_int(x,iparams->n) - gsl_sf_pow_int(1.0-x,iparams->n) ) /x /(1.0-x); 
	return f;
}

/*double 
Pn_d_k_integrand_gamma_shift( double x, void * params )
{
	struct integrand_params * iparams;
	double zeta, f;
	
	iparams = (struct integrand_params *)params;
	
	zeta = gsl_sf_hzeta( 0.5, x + 0.5/iparams->gamma );
	// Should I add lines for x~0 or x~1 ??
	f = zeta * ( 1.0 - gsl_sf_pow_int(x,iparams->n) - gsl_sf_pow_int(1.0-x,iparams->n) ) /x /(1.0-x); 
	return f;
}*/

double 
Pn_d_k_integrand_exp_reflect( double x, void * params )
{
	struct integrand_params * iparams;
	double psi, f;
	
	iparams = (struct integrand_params *)params;
	
	psi  = gsl_sf_psi( 1.0 - 0.5*x + 0.5/iparams->gamma );
	psi -= gsl_sf_psi(       0.5*x + 0.5/iparams->gamma );
	/* Should I add lines for x~0 or x~1 ?? */
	f = psi * ( 1.0 - gsl_sf_pow_int(x,iparams->n) - gsl_sf_pow_int(1.0-x,iparams->n) ) /x /(1.0-x); 
	return f;
}

/*double 
Pn_d_k_integrand_gamma_reflect( double x, void * params )
{
	struct integrand_params * iparams;
	double zeta, f;
	
	iparams = (struct integrand_params *)params;
	zeta  = gsl_sf_hzeta( 0.5, 0.5 - 0.5*x + 0.25/iparams->gamma );
	zeta -= gsl_sf_hzeta( 0.5, 0.5 + 0.5*x + 0.25/iparams->gamma );
	// Should I add lines for x~0 or x~1 ??
	f = zeta * ( 1.0 - gsl_sf_pow_int(x,iparams->n) - gsl_sf_pow_int(1.0-x,iparams->n) ) /x /(1.0-x); 
	return f;
}*/

double 
Dn_d_k1_integrand_exp_reflect( double x, void * params )
{
	struct integrand_params * iparams;
	double psi, f;
	
	iparams = (struct integrand_params *)params;
	
	psi  = gsl_sf_psi( 0.5 + 0.5*x + 0.5/iparams->gamma );
	psi -= gsl_sf_psi( 0.5 - 0.5*x + 0.5/iparams->gamma );
	f = psi * gsl_sf_pow_int(1.0-x,iparams->n -1 )/x; 
	return f;
}

double 
Dn_d_k1_integrand_exp_shift( double x, void * params )
{
	struct integrand_params * iparams;
	double psi, f;
	
	iparams = (struct integrand_params *)params;
	
	psi  = gsl_sf_psi( 1.0 + 1.0/iparams->gamma );
	psi -= gsl_sf_psi( 1.0 + 1.0/iparams->gamma - x );
	f = psi * gsl_sf_pow_int(1.0-x,iparams->n -1) / x; 
	return f;
}


/*************** k2 Calculate the contribution to the expected divergence ****************/
double calc_Dn_d_k2( double gamma_d )
{
	int d_status;	
	gsl_sf_result d_r;
	double result;

	if ( !M_BETA_IS_ONE ) {
		fprintf(stderr,"\n No available implementation for beta = 1/2");
		exit(1);
	}

	if ( M_DIST_IS_REFLECTED ) { /* Gamma shape param. Beta = 1; reflected Exponential dist.*/
		d_status = gsl_sf_psi_1_e( 0.5 + 0.5/gamma_d, &d_r );
		if ( d_status==GSL_SUCCESS ) {
			result = d_r.val;
			d_status = gsl_sf_psi_1_e( 1.0 + 1.0/gamma_d, &d_r );
			if ( d_status==GSL_SUCCESS ) {
				result += 4.0 * d_r.val;
				d_status = gsl_sf_psi_1_e( 1.0 + 0.5/gamma_d, &d_r );
				if ( d_status==GSL_SUCCESS ) {		
					result -= d_r.val;
					return result * 0.25 / gamma_d;
				}
			}
		}
	} 
	else {					 /* Gamma shape param. Beta = 1; Exponential shift model */
		d_status = gsl_sf_psi_1_e( 1.0 + 1.0/gamma_d, &d_r );
		if ( d_status==GSL_SUCCESS ) {
			return d_r.val/ gamma_d;
		}
	}
	/*else if ( M_DIST_IS_REFLECTED ) {			 // Gamma shape param. Beta = 1/2; reflected Gamma dist.
		d_status = gsl_sf_hzeta_e( 1.5, 0.25 / gamma_d, &d_r );
		if ( d_status==GSL_SUCCESS ) {
			result = d_r.val;
			d_status = gsl_sf_hzeta_e( 1.5, 1.0 + 0.25 / gamma_d, &d_r );
			if ( d_status==GSL_SUCCESS ) {
				result += d_r.val;
				return result * 0.25 * sqrt( 0.25 / gamma_d );
			}
		}
	} 
	else {										//  Gamma shape param. Beta = 1/2; Gamma shift
		d_status = gsl_sf_hzeta_e( 1.5, 1.0 + 0.5 / gamma_d, &d_r );
		if ( d_status==GSL_SUCCESS ) {
			return d_r.val * 0.5 *  sqrt( 0.5 / gamma_d );
		}
	}*/
	
	fprintf(stderr,"\nError in deleterious s D_k2 calculation");
	exit(1);
}


/*************** k1 Calculate the contribution to the expected divergence ***************/
double calc_Dn_d_k1( double gamma_d, int n )
{
	
	double i_result, i_error, result;
	size_t i_neval;
	int i_status;
	
	if ( !M_BETA_IS_ONE ) {
		fprintf(stderr,"\n No available implementation for beta = 1/2");
		exit(1);
	}
	
	/* Carry out integration F(n) */
	i_params.n = n;
	i_params.gamma = gamma_d;
	i_status = gsl_integration_qng( &FDn, 0.00, 1.00, 0.001, 0.001, &i_result, &i_error, &i_neval );	
	if ( i_status!=GSL_SUCCESS ) {
		/* Try the more long winded integration method */
		i_status = gsl_integration_qag( &FDn, 0.00, 1.00, 0.001, 0.001, 1000, 6, w, &i_result, &i_error ); 
		if ( i_status!=GSL_SUCCESS ) {
			fprintf(stderr,"\n calc_Dn_d_k1(%0.4g, %d): Integration accuracy error!: %s\n", gamma_d,n,gsl_strerror(i_status));
			exit(1);
		}
	}
	result = i_result;
	
	/* Carry out second integration F(1) */
	/*i_params.n = 1;
	i_params.gamma = gamma_d;
	i_status = gsl_integration_qng( &FPn, 0.00, 1.00, 0.001, 0.001, &i_result, &i_error, &i_neval );	
	if ( i_status!=GSL_SUCCESS ) {
		// Try the more long winded integration method
		i_status = gsl_integration_qag( &FPn, 0.00, 1.00, 0.001, 0.001, 1000, 6, w, &i_result, &i_error ); 
		if ( i_status!=GSL_SUCCESS ) {
			fprintf(stderr,"\n calc_Dn_d_k1(%0.4g, %d): Integration accuracy error!: %s\n", gamma_d,n,gsl_strerror(i_status));
			exit(1);
		}
	}
	result = i_result;*/
	
	result /= gamma_d;
	if ( M_DIST_IS_REFLECTED ) {
		result *= 0.5;
	}
	return result;
}

/******************** Calculate the contribution to the expected polymorphism *****************/
double calc_Pn_d_k( double gamma_d, int n ) 
{
	double i_result, i_error, result;
	size_t i_neval;
	int i_status;
	
	gsl_sf_result d_r;
	int r_status;
	
	i_params.n = n;
	i_params.gamma = gamma_d;

	/* Carry out numerical integration*/
	i_status = gsl_integration_qng( &FPn, 0.00, 1.00, 0.001, 0.001, &i_result, &i_error, &i_neval );	
	if ( i_status!=GSL_SUCCESS ) {
		/* Try the more long winded integration method */
		i_status = gsl_integration_qag( &FPn, 0.00, 1.00, 0.001, 0.001, 1000, 6, w, &i_result, &i_error ); 
		if ( i_status!=GSL_SUCCESS ) {
			fprintf(stderr,"\n calc_Pn_integral(%0.4g, %d): Integration accuracy error!: %s\n", gamma_d,n,gsl_strerror(i_status));
			exit(1);
		}
	}

	if ( !M_BETA_IS_ONE ) {
		fprintf(stderr,"\n No available implementation for beta = 1/2");
		exit(1);
	}

	/* Complete the calculation */
	if ( M_DIST_IS_REFLECTED ) {	/* Gamma shape param Beta = 1; Reflected Exponential */
		result = 0.5/gamma_d * i_result;
	}
	else {						/* Gamma shape param Beta = 1; Exponential shift */ 
		result = gsl_sf_psi( 1.0 + 1.0/gamma_d ) * 2.0 * neutral_P_k[n-2];
		result -= i_result;
		result /= gamma_d;	
	}
	/*else if ( M_DIST_IS_REFLECTED ) {				// Gamma shape param Beta = 0.5; Reflected Gamma
		result = 2.0 * neutral_P_k[n-2];
		result -= sqrt( 0.25/gamma_d ) * i_result; 
	}
	else {											// Gamma shape param Beta = 0.5; Gamma shift
		result = i_result;
		result -= gsl_sf_hzeta( 0.5, 1.0 + 0.5/gamma_d ) * 2.0 * neutral_P_k[n-2];
		fprintf(stderr,"\nzeta_func(%g) = %g",1.0 + 0.5/gamma_d,gsl_sf_hzeta( 0.5, 1.0 + 0.5/gamma_d ) );
		result *= sqrt( 0.5 / gamma_d );
		fprintf(stderr,"\nmultiplier = %g",sqrt( 0.5 / gamma_d ));
	}*/
	return result;
}

void universal_gamma_d( double gamma_d )
{
	int j;
	double gamma_d_sex = gamma_d * 0.75;
	
	old_univ_Dn_d_k2[0] = univ_Dn_d_k2[0];
	old_univ_Dn_d_k2[1] = univ_Dn_d_k2[1];
	univ_Dn_d_k2[0] = calc_Dn_d_k2( gamma_d_sex ); /* For the X */
	univ_Dn_d_k2[1] = calc_Dn_d_k2( gamma_d );	   /* For autosomes */
	
	for(j = 0; j < n_values_count; j++) {
		
		old_univ_Pn_d_k[j][0] = univ_Pn_d_k[j][0];
		old_univ_Pn_d_k[j][1] = univ_Pn_d_k[j][1];
		univ_Pn_d_k[j][0] = calc_Pn_d_k( gamma_d_sex, n_values[j] ); /* For the X */
		univ_Pn_d_k[j][1] = calc_Pn_d_k( gamma_d,     n_values[j] ); /* For autosomes */
		
		old_univ_Dn_d_k1[j][0] = univ_Dn_d_k1[j][0];
		old_univ_Dn_d_k1[j][1] = univ_Dn_d_k1[j][1];
		univ_Dn_d_k1[j][0] = calc_Dn_d_k1( gamma_d_sex, n_values[j] ) + calc_Dn_d_k1( gamma_d_sex, M_CORRECTION_COEFF ); // For the X 
		univ_Dn_d_k1[j][1] = calc_Dn_d_k1( gamma_d,     n_values[j] ) + calc_Dn_d_k1( gamma_d,     M_CORRECTION_COEFF ); // For autosomes
	}
}

void replace_universal_gamma_d( void )
{
	int j;
	univ_Dn_d_k2[0] = old_univ_Dn_d_k2[0];
	univ_Dn_d_k2[1] = old_univ_Dn_d_k2[1];
	for(j = 0; j < n_values_count; j++) {
		
		univ_Pn_d_k[j][0] = old_univ_Pn_d_k[j][0];
		univ_Pn_d_k[j][1] = old_univ_Pn_d_k[j][1];
		
		univ_Dn_d_k1[j][0] = old_univ_Dn_d_k1[j][0];
		univ_Dn_d_k1[j][1] = old_univ_Dn_d_k1[j][1];
	}
}

double get_univ_selected_P_k(struct dataset * data, int which_gene)
{
	int j;
	int autosome = (data->items[which_gene].chromosome == 0) ? 0 : 1;
	
	for (j=0; j < n_values_count; j++) {
		if( data->items[which_gene].alleles == n_values[j] )
			return univ_Pn_d_k[j][autosome];
	}
}

double get_univ_selected_D_k1(struct dataset * data, int which_gene)
{
	int j;
	int autosome = (data->items[which_gene].chromosome == 0) ? 0 : 1;
	
	for (j=0; j < n_values_count; j++) {
		if( data->items[which_gene].alleles == n_values[j] )
			return univ_Dn_d_k1[j][autosome];
	}
}

double get_univ_selected_D_k2(struct dataset * data, int which_gene)
{
	int autosome = (data->items[which_gene].chromosome == 0) ? 0 : 1;
	
	return univ_Dn_d_k2[autosome];
}


/* Retrieve LLterm for the specified likelihood term gene and model */
static double 
calc_log_likelihood_term(double *params, struct dataset *data, int which_gene, int which_term)
{	
	/* things required for almost all models */
	int num_genes = data->count;
	int measured  = data->items[which_gene].measured[which_term];
	double sites  = data->items[which_gene].sites[which_term];
	int i;
	
	double theta, time, omega;
	
	/* Get the relevant theta parameter */
	switch( data->models[P_THETA] ) {
		case MODEL_UNIVERSAL:
			theta = params[0];
			theta *= (data->items[which_gene].chromosome == 0) ? 0.75 : 1.00;
			break;
		case MODEL_LOCUSSPECIFIC:
			theta = params[which_gene];
			break;
		case MODEL_CLASSSPECIFIC:	
			theta = params[data->items[which_gene].flag];
			theta *= (data->items[which_gene].chromosome == 0) ? 0.75 : 1.00;
			break;
		case MODEL_XAUTOSOMAL:	
			theta = (data->items[which_gene].chromosome == 0) ? params[1] : params[0];
			break;
		case MODEL_REGRESSION:
			//theta = exp(params[0] + params[1]*data->items[which_gene].r);	
			theta = params[0] + params[1]*data->items[which_gene].r;
			break;
		default:
			fprintf(stderr,"\nbad theta model number %d\n",data->models[P_THETA] );
	}
	/* we now have enough info to calculate the neutral diversity if necessary */
	if( which_term == P_Pi ) {
		return calc_log_poisson( theta * neutral_P_k[data->items[which_gene].alleles - 2] * sites, measured );
	}	
	/* Get the relevant time parameter */
	i = get_nnonalphap( data->models[P_THETA], data->flag_count, num_genes);
	switch( data->models[P_TIME] ) {
		case MODEL_UNIVERSAL:
			time = params[i];
			break;
		case MODEL_LOCUSSPECIFIC:
			time = params[i + which_gene];
			break;
		case MODEL_CLASSSPECIFIC:		
			time = params[i + data->items[which_gene].flag];
			break;
		case MODEL_XAUTOSOMAL:	
			time = (data->items[which_gene].chromosome == 0) ? params[i+1] : params[i];
			break;
		default:
			fprintf(stderr,"\nbad time model number %d\n",data->models[P_TIME]);
			exit(1);
	}	
	/* we now have enough info to calculate the neutral divergence if necessary */
	if( which_term == P_Di ) {	
		if (!BEW)
			time += theta*neutral_D_k1[data->items[which_gene].alleles - 2];
		return calc_log_poisson( time * sites, measured );
	}	
	/* Get the relevant constraint parameter */ 
	i += get_nnonalphap( data->models[P_TIME], data->flag_count, num_genes);
	switch (data->models[P_OMEGA])	{
		case MODEL_NONE:
			omega = 1.0;
			break;
		case MODEL_UNIVERSAL:
			omega = params[i];
			break;
		case MODEL_LOCUSSPECIFIC:
			omega = params[i + which_gene];
			break;
		case MODEL_CLASSSPECIFIC:		
			omega = params[i + data->items[which_gene].flag];
			break;
		case MODEL_XAUTOSOMAL:	
			omega = (data->items[which_gene].chromosome == 0) ? params[i+1] : params[i];
			break;
		default:
			fprintf(stderr,"\nbad constraint model number %d\n",data->models[P_OMEGA]);
			exit(1);
	}
	/* Calculate Pn if necessary */
	if( which_term == P_Pn ) {
		/* multiply theta to account for other factors */
		switch (data->models[P_GAMMA_D])	{
			case MODEL_NONE:
				theta *= neutral_P_k[data->items[which_gene].alleles - 2];
				break;
			case MODEL_UNIVERSAL:
				theta *= get_univ_selected_P_k(data,which_gene);
				break;
			case MODEL_LOCUSSPECIFIC:
				theta *= data->items[which_gene].selected_k[I_Pn_d];
				break;
			default:
				fprintf(stderr,"\nbad gamma model number %d\n",data->models[P_GAMMA_D]);
				exit(1);
		}
		return calc_log_poisson( theta * omega * sites, measured );  
	}
	/* Calculate Dn - this is the difficult one */
	if ( which_term == P_Dn ) {
		/* alter time (wow) to account for other factors */
		switch (data->models[P_GAMMA_D])	{
			case MODEL_NONE:
				if (!BEW)
					time += theta*neutral_D_k1[data->items[which_gene].alleles - 2]; // Don't do for BEW-style model
				break;	
			case MODEL_UNIVERSAL:
				time *= get_univ_selected_D_k2(data,which_gene);
				if (!BEW)
					time += theta*get_univ_selected_D_k1(data,which_gene); // Don't do for BEW-style model
				break;
			case MODEL_LOCUSSPECIFIC:
				time *= data->items[which_gene].selected_k[I_Dn_d2];
				if (!BEW)
					time += theta*data->items[which_gene].selected_k[I_Dn_d1];  // Don't do for BEW-style model
				break;
			default:
				fprintf(stderr,"\nbad gamma model number %d\n",data->models[P_GAMMA_D]);
				exit(1);
		}
		
		/* Assemble a grand parameter */
		time *= omega * sites;
		switch ( data->models[P_ALPHA] ) {
			case ALPHA_ZERO:
				return calc_log_poisson( time, measured );
			case ALPHA_FREE: // Poisson(param/(1-alpha))
				if(A_NOT_ALPHA)
					return calc_log_poisson( time + params[param_count-1]*sites, measured);
					//return calc_log_poisson( time*(1.0+params[param_count-1]/omega), measured);
				return calc_log_poisson( time/( 1.0 - params[param_count-1]), measured );
			case ALPHA_BETA: // BetaDistribution(a,b,param)
				return calc_log_beta_integral( params[param_count-2], params[param_count-1], time, measured );	
			case ALPHA_DELTA: // DeltaDistribution(a0,a1,p,param)
				return calc_log_delta_integral( params[param_count-3], params[param_count-2], params[param_count-1], time, measured );
			case ALPHA_MANY: // pick the correct alpha value
				if(A_NOT_ALPHA)
					return calc_log_poisson( time + params[param_count - data->flag_count + data->items[which_gene].flag]*sites, measured );
					//return calc_log_poisson( time*(1.0+params[param_count - data->flag_count + data->items[which_gene].flag]/omega), measured );
				return calc_log_poisson( time/( 1.0 - params[param_count - data->flag_count + data->items[which_gene].flag]), measured );
			case ALPHA_FULL: // Locus-specific alpha value
				// Now hacked so that the perfect ML value is always calculated
				if(A_NOT_ALPHA) {
					params[param_count - num_genes + which_gene] = ((double)measured - time)/sites;
					//return calc_log_poisson( time + params[param_count - num_genes + which_gene]*sites, measured );
				} else {
					params[param_count - num_genes + which_gene] = (measured == 0) ? GSL_NEGINF : 1.0 - time/((double)measured);
					//return calc_log_poisson( time/( 1.0 - params[param_count - num_genes + which_gene]), measured );
				}
				return calc_log_poisson( (double)measured, measured );
			default:
				fprintf(stderr,"\nWeird alpha encountered in calc_log_likelihood_term");
				exit(1);
		}
	}
	fprintf(stderr,"\nShould not be here in calc_log_likelihood_term");
	exit(1);
}

/* specifies which terms of the likelihood equation contain a given parameter in a given model */
static void
get_likelihood_info(int which_param, int * models, struct param_info * info, int num_genes, int flag_count)
{
	int i;
	info->which_gene = -1;
	
	/* Alpha */
	if( models[P_ALPHA] == ALPHA_FREE && which_param == param_count-1 ) {
		info->what_param = P_ALPHA;
	} else if ( models[P_ALPHA] == ALPHA_FULL && which_param >= param_count-num_genes ) {
		info->what_param = P_ALPHA;
		info->which_gene = which_param - param_count + num_genes;
	} else if( models[P_ALPHA] == ALPHA_MANY && which_param >= param_count-flag_count ) {
		info->what_param = P_ALPHA;
	} else if ( models[P_ALPHA] == ALPHA_BETA && which_param >= param_count-2 ) {
		info->what_param = ( which_param == param_count-1 ) ? P_BETA_B : P_BETA_A;
	} else if ( models[P_ALPHA] == ALPHA_DELTA && which_param >= param_count-3 ) {
		info->what_param = ( which_param == param_count-1 ) ? P_DELTA_P : 
			( which_param == param_count-2 ) ?  P_DELTA_A1 : P_DELTA_A0;
 	
	} else { /* Theta */
		if ( models[P_THETA] == MODEL_UNIVERSAL && which_param == 0 ) {	
			info->what_param = P_THETA;
		} else if ( models[P_THETA] == MODEL_XAUTOSOMAL && which_param < 2) {
			info->what_param = P_THETA;
		} else if ( models[P_THETA] == MODEL_REGRESSION && which_param < 2) {
			info->what_param = (which_param == 0) ? P_REG_INTERCEPT : P_REG_SLOPE;
		} else if ( models[P_THETA] == MODEL_CLASSSPECIFIC && which_param < flag_count ) {
			info->what_param = P_THETA;
		} else if ( models[P_THETA] == MODEL_LOCUSSPECIFIC && which_param < num_genes ) {
			info->what_param = P_THETA;
			info->which_gene = which_param;
		
		} else { /* Time */
			
			i = get_nnonalphap( models[P_THETA],  flag_count, num_genes );
			if ( models[P_TIME] == MODEL_UNIVERSAL && which_param == i ) {	
				info->what_param = P_TIME;
			} else if ( models[P_TIME] == MODEL_XAUTOSOMAL && which_param < i+2 ) {
				info->what_param = P_TIME;
			} else if ( models[P_TIME] == MODEL_CLASSSPECIFIC && which_param < i+flag_count ) {
				info->what_param = P_TIME;
			} else if ( models[P_TIME] == MODEL_LOCUSSPECIFIC && which_param < i+num_genes ) {
				info->what_param = P_TIME;
				info->which_gene = which_param-i;
			
			} else { /* Omega */
			
				i += get_nnonalphap( models[P_TIME],  flag_count, num_genes );
				if ( models[P_OMEGA] == MODEL_UNIVERSAL && which_param == i ) {	
					info->what_param = P_OMEGA;
				} else if ( models[P_OMEGA] == MODEL_XAUTOSOMAL && which_param < i+2 ) {
					info->what_param = P_OMEGA;
				} else if ( models[P_OMEGA] == MODEL_CLASSSPECIFIC && which_param < i+flag_count ) {
					info->what_param = P_OMEGA;
				} else if ( models[P_OMEGA] == MODEL_LOCUSSPECIFIC && which_param < i+num_genes ) {
					info->what_param = P_OMEGA;
					info->which_gene = which_param-i;
				
				} else { /* Gamma */
					i += get_nnonalphap( models[P_OMEGA],  flag_count, num_genes );
					if ( models[P_GAMMA_D] == MODEL_UNIVERSAL && which_param == i ) {	
						info->what_param = P_GAMMA_D;
					} else if ( models[P_GAMMA_D] == MODEL_LOCUSSPECIFIC && which_param < i+num_genes ) {
						info->what_param = P_GAMMA_D;
						info->which_gene = which_param-i;
					} else {
						fprintf(stderr, "internal error: in likelihoodinfo\n");
						exit(1);
					}
				}
			}
		}
	}
		
	switch(info->what_param) 
	{
		case P_THETA:
		case P_REG_INTERCEPT:
		case P_REG_SLOPE:
			info->which_terms[0]=P_Pi;
			info->which_terms[1]=P_Pn;
			info->which_terms[2]=P_Di;	
			info->which_terms[3]=P_Dn;
			break;
		case P_TIME:
			info->which_terms[0]=P_Di;	
			info->which_terms[1]=P_Dn;
			info->which_terms[2]= -1;
			break;
		case P_OMEGA:
		case P_GAMMA_D:
			info->which_terms[0]=P_Pn;	
			info->which_terms[1]=P_Dn;
			info->which_terms[2]= -1;
			break;
		default:
			info->which_terms[0]=P_Dn;
			info->which_terms[1]= -1;
			break;
	}
}
/* Populate initial param values etc.*/
void initialise_params(struct dataset * data, double **params,double **step_sizes,double param_start_mins[NUM_PARAM_TYPES],double param_start_maxs[NUM_PARAM_TYPES] ) {

	int i,j;
	/* Start value and initial step size for each individual parameter */
	for (i = 0; i < param_count; i++ ) {
		get_likelihood_info(i, data->models, &(data->last_param_info), data->count, data->flag_count);
		(*params)[i]= param_start_mins[data->last_param_info.what_param] + drand48() *
			(param_start_maxs[data->last_param_info.what_param]-param_start_mins[data->last_param_info.what_param]);  
		(*step_sizes)[i] = (*params)[i] == 0 ? 1e-6 : (*params)[i] * 0.1;
	}

	/* Calculate initial values of integrals involving gamma_d if required */
	if( data->models[P_GAMMA_D] == MODEL_LOCUSSPECIFIC ) {	
		for(i=0; i < param_count; i++) {
			get_likelihood_info(i, data->models, &(data->last_param_info), data->count, data->flag_count); 
			if ( data->last_param_info.what_param == P_GAMMA_D ) { /* if we have a GAMMA_D param, then calculate integrals */
				data->items[data->last_param_info.which_gene].selected_k[I_Pn_d] = 
						calc_Pn_d_k( (*params)[i], data->items[data->last_param_info.which_gene].alleles );	
				data->items[data->last_param_info.which_gene].selected_k[I_Dn_d1] = 
						calc_Dn_d_k1( (*params)[i], data->items[data->last_param_info.which_gene].alleles )
							+ calc_Dn_d_k1( (*params)[i], M_CORRECTION_COEFF );
				data->items[data->last_param_info.which_gene].selected_k[I_Dn_d2] = 
						calc_Dn_d_k2( (*params)[i] );
			}
		}
	} else if ( data->models[P_GAMMA_D] == MODEL_UNIVERSAL )	{
		i=-1;
		do { // Find the gamma_d parameter
			i++;
			get_likelihood_info(i, data->models, &(data->last_param_info), data->count, data->flag_count);
		} while ( data->last_param_info.what_param != P_GAMMA_D );
		univ_Dn_d_k2[0] = calc_Dn_d_k2( (*params)[i] * 0.75 ); // For the X
		univ_Dn_d_k2[1] = calc_Dn_d_k2( (*params)[i] );		// For autosomes
		for(j = 0; j < n_values_count; j++) {
			univ_Pn_d_k[j][0]  = calc_Pn_d_k(  (*params)[i] * 0.75, n_values[j] ); // For the X
			univ_Pn_d_k[j][1]  = calc_Pn_d_k(  (*params)[i],        n_values[j] ); // For autosomes
			univ_Dn_d_k1[j][0] = calc_Dn_d_k1( (*params)[i] * 0.75, n_values[j] ) + calc_Dn_d_k1( (*params)[i] * 0.75, M_CORRECTION_COEFF ); // For the X 
			univ_Dn_d_k1[j][1] = calc_Dn_d_k1( (*params)[i],        n_values[j] ) + calc_Dn_d_k1( (*params)[i],        M_CORRECTION_COEFF ); // For autosomes
		}
	}
}

/* Set up and arrays whose size depends on the model specified */
void setup_model(struct dataset * data, double **params, 
					double **mins, double **maxs, double **step_sizes,int **indices,
					double param_mins[NUM_PARAM_TYPES], double param_maxs[NUM_PARAM_TYPES]) {
	int i;
	
	/* Get param count suitable for model, and display info to screen */
	if ( (param_count = get_np( data->models, data->count, data->flag_count )) > 1 ) {
		printf("\nNeutral diversity (theta) %d: %s", data->models[P_THETA], model_text[data->models[P_THETA]] );
		printf("\nNeutral divergence (mu_t) %d: %s", data->models[P_TIME], model_text[data->models[P_TIME]] );
		printf("\nSelective constraint  (f) %d: %s", data->models[P_OMEGA], model_text[data->models[P_OMEGA]] );
		if(data->models[P_GAMMA_D] != MODEL_NONE ) 
			printf("gamma=4Ne_s for weakly selected mutations (%d): %s\n", data->models[P_GAMMA_D], model_text[data->models[P_GAMMA_D]] );
		if(A_NOT_ALPHA) {
			printf("\nAdaptive subs (a NOT alpha) %d: %s\n", data->models[P_ALPHA], alpha_text[data->models[P_ALPHA]] );
		} else {
		printf("\nAdaptive subs.    (alpha) %d: %s\n", data->models[P_ALPHA], alpha_text[data->models[P_ALPHA]] );
		}
		printf("\nTotal number of parameters = %d\n", param_count);
	} else {
		fprintf(stderr, "%s: Cannot specify model %d\n", progname, param_count );
		exit(1);
	}
	/* Allocate room for the mins, maxs and params */
    if ((*params = calloc(param_count, sizeof(double))) == NULL ||
        (*mins = calloc(param_count, sizeof(double))) == NULL ||
        (*maxs = calloc(param_count, sizeof(double))) == NULL ||
        (*step_sizes = calloc(param_count, sizeof(double))) == NULL ||
		(*indices = calloc(param_count, sizeof(int))) == NULL) {
        fprintf(stderr, "%s: unable to allocate memory\n", progname);
        exit(1);
    }
	/* Bounds for each individual parameter */
	for (i = 0; i < param_count; i++ ) {
		get_likelihood_info(i, data->models, &(data->last_param_info), data->count, data->flag_count);
		(*mins)[i]=	param_mins[data->last_param_info.what_param];
		(*maxs)[i]=	param_maxs[data->last_param_info.what_param];
	}
}

/* Sum all of the log likelihood terms, to calculate the full log likelihood value */
static double tally_log_likelihood(void * rock)
{
	struct dataset *data = (struct dataset *)rock;
	double sum = 0.0;
	int g, t;
	
	/* sum the log likelihoods for each term and for all genes */
    for (g = 0; g < data->count; g++) {
		for (t = 0; t< NUM_TERMS; t++) {
			sum += data->items[g].likelihood[t];
		} 
    }
	return sum;
}

/* Populate all likelihood terms, and return complete loglikelihood*/
static double calc_full_log_likelihood(double *params, void *rock)
{
    struct dataset *data = (struct dataset *)rock;
	int g, t;
	
	for (g = 0; g < data->count; g++) {
		for (t = 0; t< NUM_TERMS; t++) {
			data->items[g].likelihood[t] = calc_log_likelihood_term( params, data, g, t );
		} 
    }
	return tally_log_likelihood(data);
}

/* Calculates the log likelihood */
static double
calc_log_likelihood(int which_param, int last_used,
                    int param_count, double *params,
                    void *rock)
{
    struct dataset *data = (struct dataset *)rock;
	int g, t, j, num_terms_altered;
	
	if (which_param < 0) {
		return calc_full_log_likelihood(params,data);
	}
	
	/* 1) Deal with old parameter if it was rejected */
    if (!last_used) { /* parameter was rejected */
		
		num_terms_altered = (data->last_param_info.which_terms[1] == -1) ? 1 : (data->last_param_info.which_terms[2] == -1) ? 2 : 4;
		
		if (data->last_param_info.which_gene < 0) { /* a parameter applying to all genes */
			
			if (data->last_param_info.what_param == P_GAMMA_D) {
				replace_universal_gamma_d();
			}
			for (g = 0; g < data->count; g++) {
				for (j = 0; j < num_terms_altered; j++) {
					t = data->last_param_info.which_terms[j];
					data->items[g].likelihood[t] = data->items[g].old_likelihood[t];
				}
			}
		} else {									/* a parameter applying to one gene only */
			g = data->last_param_info.which_gene;
			
			if ( data->last_param_info.what_param == P_GAMMA_D ) {
				data->items[g].selected_k[I_Dn_d1] = data->items[g].old_selected_k[I_Dn_d1];
				data->items[g].selected_k[I_Dn_d2] = data->items[g].old_selected_k[I_Dn_d2]; 	
				data->items[g].selected_k[I_Pn_d]  = data->items[g].old_selected_k[I_Pn_d];
			}
			for (j = 0; j < num_terms_altered; j++) {
					t = data->last_param_info.which_terms[j];
					data->items[g].likelihood[t] = data->items[g].old_likelihood[t];
			}
		}
    } // end if parameter was rejected
		
	/* 2) Deal with new parameter  */
	/* get information on current parameter - populating last_param_info */
	get_likelihood_info(which_param, data->models, &(data->last_param_info), data->count, data->flag_count);
	num_terms_altered = (data->last_param_info.which_terms[1] == -1) ? 1 : (data->last_param_info.which_terms[2] == -1 ) ? 2 : 4;
	
	if (data->last_param_info.which_gene < 0) { /* a parameter applying to all genes */
		// Store and recalculate integrals if necessary
		if (data->last_param_info.what_param == P_GAMMA_D) {
			universal_gamma_d( params[which_param] );
		} 
		// Store and recalculate likelihood terms
		for (g = 0; g < data->count; g++) {
			for (j = 0; j < num_terms_altered; j++) {
				t = data->last_param_info.which_terms[j];
				data->items[g].old_likelihood[t] = data->items[g].likelihood[t];
				data->items[g].likelihood[t] = calc_log_likelihood_term( params, data, g, t );
			}
		}
	} else {									/* a parameter applying to one gene only */
		g=data->last_param_info.which_gene;
		// Store and recalculate integrals if necessary
		if (data->last_param_info.what_param == P_GAMMA_D) {
			data->items[g].old_selected_k[I_Pn_d]  = data->items[g].selected_k[I_Pn_d];
			data->items[g].old_selected_k[I_Dn_d1] = data->items[g].selected_k[I_Dn_d1];
			data->items[g].old_selected_k[I_Dn_d2] = data->items[g].selected_k[I_Dn_d2];
			data->items[g].selected_k[I_Pn_d]  = calc_Pn_d_k(  params[which_param], data->items[g].alleles );
			data->items[g].selected_k[I_Dn_d1] = calc_Dn_d_k1( params[which_param], data->items[g].alleles ) 
											   + calc_Dn_d_k1( params[which_param], M_CORRECTION_COEFF );
			data->items[g].selected_k[I_Dn_d2] = calc_Dn_d_k2( params[which_param] );
		}
		// Store and recalculate likelihood terms
		for (j = 0; j < num_terms_altered; j++) {
			t = data->last_param_info.which_terms[j];
			data->items[g].old_likelihood[t] = data->items[g].likelihood[t];
			data->items[g].likelihood[t] = calc_log_likelihood_term( params, data, g, t );
		}
	}
	
	return tally_log_likelihood(data);
}

void output_header(struct dataset * data,int param_count, FILE * out) {
	int i,j, old_param_type;
	fprintf(out, "LnL");
	old_param_type = -1;
	for(i=0;i<param_count; i++) {
    	get_likelihood_info(i, data->models, &(data->last_param_info), data->count, data->flag_count);
		if(data->last_param_info.what_param != old_param_type) 
			j = 0;
		fprintf(out, "\t%s", param_text[data->last_param_info.what_param]);
		switch(data->last_param_info.what_param) {
			case P_THETA:
			case P_TIME:
			case P_OMEGA:
			case P_GAMMA_D:
				switch(data->models[data->last_param_info.what_param]) {
					case MODEL_LOCUSSPECIFIC:
						fprintf(out, "%d", data->last_param_info.which_gene+1);	
						break;
					case MODEL_CLASSSPECIFIC:
						fprintf(out, "C%d", data->flag_list[j]);	
						break;
					case MODEL_XAUTOSOMAL:
						if(j==0) {
							fprintf(out, "[auto]", data->flag_list[j]);	
						} else {
							fprintf(out, "[X]", data->flag_list[j]);	
						}
						break;
				}
				break;
			case P_ALPHA:
				switch(data->models[data->last_param_info.what_param]) {
					case ALPHA_FULL:
						fprintf(out, "%d", data->last_param_info.which_gene+1);	
						break;
					case ALPHA_MANY:
						fprintf(out, "C%d", data->flag_list[j]);	
						break;
				}
				break;
		}
		old_param_type = data->last_param_info.what_param; 
		j++;
	}
    fprintf(out, "\n");
}

static void output_line(double likelihood,int param_count,double *params, FILE *out)
{
    int i;

    /* Write the likelihood and the current value of each parameter */
    fprintf(out, "%0.12g", likelihood);
    for (i = 0; i < param_count; i++) {
        fprintf(out, "\t%0.12g", params[i]);
    }
    fprintf(out, "\n");
}

/* Read data from the input file */
static void
read_data(FILE *in, char *filename, int *count_out, struct item **items_out, int *flag_count_out, int **flag_list_out, int x, int D_ONLY_FLAG, int D_EXCLUDE_FLAG)
{
    double values[12];
    int result;
    int line = 0;
    struct item *items = NULL;
    struct item *new_items;
    int count = 0;
	int *flag_list = NULL;
	int *new_flag_list;
	int flag_count = 0;
	int seen_flag;
	int i;
	
    while (!feof(in)) {
        /* Watch for errors */
        if (ferror(in)) {
            fprintf(stderr, "%s: error reading file %s: %s\n",
                    progname, filename, strerror(errno));
            exit(1);
        }

        /* Read a line of numbers */
        line++;
        result = freadnums(in, values, 12);

        if (result < 0) {
            fprintf(stderr, "%s: non-numeric input: %s:%d\n",
                    progname, filename, line);
            exit(1);
        } else if (result == 0) {
            continue;
        } else if (result < 9 || result > 12) {
            fprintf(stderr, "%s: wrong number of values %s:%d\n",
                    progname, filename, line);
            exit(1);
        }

        if ( (int)values[6] > x ) {	/* more than x intron polymorphisms? */
		if ( result < 11 || ((D_ONLY_FLAG == -1 || D_ONLY_FLAG == (int)values[10]) && D_EXCLUDE_FLAG != (int)values[10]) ) { /* keep this class? */
		if ( single_gene == line || (single_gene <= 0 && single_gene != -line ) ) { /* keep this gene? */
			
			/* Allocate more room for the new item */
			new_items = realloc(items, (count + 1) * sizeof(struct item));
			if (new_items == NULL) {
				free(items);
				fprintf(stderr, "%s: out of memory\n", progname);
				exit(1);
			}
			
			/* Record the values - check that this order is correct! */
			items = new_items;
			items[count].measured[P_Dn] = (int)values[0];
			items[count].sites[P_Dn]    = values[1];
		
			items[count].measured[P_Pn] = (int)values[2];
			items[count].sites[P_Pn]    = values[3];
		
			items[count].measured[P_Di] = (int)values[4];
			items[count].sites[P_Di]    = values[5];
		
			items[count].measured[P_Pi] = (int)values[6];
			items[count].sites[P_Pi]    = values[7];

			items[count].alleles		= (int)values[8];
			/* Record optional data if present */
			items[count].chromosome		= (result < 10) ? 1 : (int)values[9];
			items[count].flag			= (result < 11) ? 0 : (int)values[10];
			items[count].r				= (result < 12) ? 0.0 : values[11];
			
			/* Sanity check the data */
			if (items[count].measured[P_Dn] < 0) {
				fprintf(stderr, "%s: negative Dn at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].measured[P_Pn] < 0) {
				fprintf(stderr, "%s: negative Pn at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].measured[P_Di] < 0) {
				fprintf(stderr, "%s: negative Di at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].measured[P_Pi] < 0) {
				fprintf(stderr, "%s: negative Pi at %s:%d\n",
					progname, filename, line);
				exit(1);
			}	
			if (items[count].sites[P_Dn] < 1.0) {
				fprintf(stderr, "%s: incorrect L(Dn) at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].sites[P_Pn] < 1.0) {
				fprintf(stderr, "%s: incorrect L(Pn) at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].sites[P_Di] < 1.0) {
				fprintf(stderr, "%s: incorrect L(Di) at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].sites[P_Pi] < 1.0) {
				fprintf(stderr, "%s: incorrect L(Pi) at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].alleles < 2) {
				fprintf(stderr, "%s: incorrect n at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].flag < 0 ) {
				fprintf(stderr, "%s: incorrect flag at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].chromosome < 0 ) {
				fprintf(stderr, "%s: -ve chromosome number at %s:%d\n",
					progname, filename, line);
				exit(1);
			}
			if (items[count].r < 0.0 ) {
				fprintf(stderr, "%s: -ve recombination rate at %s:%d\n",
					progname, filename, line);
				exit(1);
			}			
			/* Check whether we have a new flag or not */
			seen_flag = 0;
			for( i = 0; i < flag_count; i++ ) {
				if( seen_flag == 0 && flag_list[i] == items[count].flag ) {
					seen_flag = 1;
					items[count].flag = i; /* Replace with an integer that can reference params */
				}
			}
			if(seen_flag == 0) { /* Add a new flag to the flag_list */
				new_flag_list = realloc(flag_list, (flag_count + 1) * sizeof(int));
				if (new_flag_list == NULL) {
					free(flag_list);
					exit(1);
				}
				flag_list = new_flag_list;
				flag_list[flag_count] = items[count].flag;
				items[count].flag = flag_count; /* Replace with an integer that can reference params */
				flag_count++;
			}
			count++;
		}
		}
		}
    } 
	*flag_count_out = flag_count;
	*flag_list_out	= flag_list;
    *items_out		= items;
    *count_out		= count;
}
/* setup the bootstrapped data set by sampling from the real data with replacement */
void create_bootstrapped_dataset(struct dataset * boot_data, struct dataset * data, int * class_size, int ** class_list) {
	int i,j,k,l,m;
	l = 0;
	
	for(i=0; i < data->flag_count; i++) {
		for(j=0; j < class_size[i]; j++ ) {
			k = floor( drand48() * class_size[i] ); /* randomly chosen member of the class */
			boot_data->items[l].alleles    = data->items[class_list[i][k]].alleles;
			boot_data->items[l].chromosome = data->items[class_list[i][k]].chromosome;
			boot_data->items[l].flag       = data->items[class_list[i][k]].flag;
			/*fprintf(stderr," %d(%d):",i,boot_data->items[l].flag);*/
			boot_data->items[l].r          = data->items[class_list[i][k]].r;
			for(m=0;m < NUM_TERMS;m++) {
				boot_data->items[l].measured[m] = data->items[class_list[i][k]].measured[m];
				boot_data->items[l].sites[m]    = data->items[class_list[i][k]].sites[m];
			}
			l++;
		}
	}	
}

/* Handler for sigint */
#ifdef HAVE_SIGNAL
static RETSIGTYPE handle_sigint(int sig)
{
    /* Flush stdout */
    fflush(stdout);

    /* And exit */
    fprintf(stderr, "  aborted\n%s: exiting on signal %d\n", progname, sig);
    exit(0);
}
#endif /* HAVE_SIGNAL */

#define OPTIONS "c:BRL:X:K:k:m:g:a:p:f:t:ho:r:P:i:vx:A"
void usage()
{
    fprintf(stderr,
            "usage: %s [OPTION]... data-file\n"
            "    -o file\t\tWrite output to file [stdout]\n\n"
            "    -a alpha model\tZero, Universal, Beta, Delta, per-class, or per-locus OR heuristic estimators etc. (0-5,999) [1]\n"
			"    -p theta model\tUniversal, per-locus, per-class, X-autosome, linear function of recomb rate ( 1-5 ) [1]\n"
			"    -f constraint model\tZero, Universal, per-locus, per-class, X-autosome ( 0-4 ) [1]\n"
            "    -g E[4Nes] model\tZero, Universal, per-locus ( 0-2 ) [0]\n"
			"    -t mu_t model\tUniversal, per-locus, per-class or X-autosome ( 1-4 ) [1]\n\n"
			"    -L alleles\t\tDivergence was polarized with this many alleles\n"
			"    -c alpha value\tFix alpha at this value [-a 1 set automatically]\n"
			"    -A\t\t\tcalculate adaptive subs/site rather than alpha\n"
			"    -B\t\t\tDon't include sampling corrections like Bierne and Eyre-Walker\n\n"
			"    -X class\t\tExclude loci in this class [NA]\n"
			"    -K class\t\tKeep only loci in this class [NA]\n"
			"    -x locus\t\tExclude this locus [NA]\n"
			"    -k locus\t\tKeep only this locus [NA]\n\n"
			"    -P num\t\tPermute the classes (+ve value) or Bootstrap the loci (-ve value) (a=4 set automatically) [0]\n"
			"    -m min(Ps)\t\tAccept only genes with greater than min(Ps) neutral polymorphisms [-1]\n"		
            "    -i num\t\tMaximum number of iterations to run before giving up [20000]\n"
			"    -r dt\t\tCool at this rate t = t*dt [0.50]\n"
			"    -R\t\t\tUse a single-sided, not a reflected exponential dist. of sel. coeffs.\n"
			"    -h\t\t\tPrint this help\n"
            "    -v\t\t\tPrint version information\n", progname);
}

/* PART 2 - hyperg_U */
/****************************************************************************************************************************************
 ****************************************************************************************************************************************
 ****************************************************************************************************************************************/
/* specfunc/hyperg_U.c 
 Copyright (C) 1996, 1997, 1998, 1999, 2000 Gerard Jungmam */

/* Evaluate u_{N+1}/u_N by Steed's continued fraction method.
 * u_N := Gamma[a+N]/Gamma[a] U(a + N, b, x)
 * u_{N+1}/u_N = (a+N) U(a+N+1,b,x)/U(a+N,b,x) */ // Unchanged from gnu
static
int
hyperg_U_CF1(const double a, const double b, const int N, const double x,
             double * result, int * count)
{
  const double RECUR_BIG = GSL_SQRT_DBL_MAX;
  const int maxiter = 20000;
  int n = 1;
  double Anm2 = 1.0;
  double Bnm2 = 0.0;
  double Anm1 = 0.0;
  double Bnm1 = 1.0;
  double a1 = -(a + N);
  double b1 =  (b - 2.0*a - x - 2.0*(N+1));
  double An = b1*Anm1 + a1*Anm2;
  double Bn = b1*Bnm1 + a1*Bnm2;
  double an, bn;
  double fn = An/Bn;

  while(n < maxiter) {
    double old_fn;
    double del;
    n++;
    Anm2 = Anm1;
    Bnm2 = Bnm1;
    Anm1 = An;
    Bnm1 = Bn;
    an = -(a + N + n - b)*(a + N + n - 1.0);
    bn =  (b - 2.0*a - x - 2.0*(N+n));
    An = bn*Anm1 + an*Anm2;
    Bn = bn*Bnm1 + an*Bnm2;
    
    if(fabs(An) > RECUR_BIG || fabs(Bn) > RECUR_BIG) {
      An /= RECUR_BIG;
      Bn /= RECUR_BIG;
      Anm1 /= RECUR_BIG;
      Bnm1 /= RECUR_BIG;
      Anm2 /= RECUR_BIG;
      Bnm2 /= RECUR_BIG;
    }
    
    old_fn = fn;
    fn = An/Bn;
    del = old_fn/fn;
    
    if(fabs(del - 1.0) < 10.0*GSL_DBL_EPSILON) break;
  }
  
  *result = fn;
  *count  = n;

  if(n == maxiter)
    GSL_ERROR ("error", GSL_EMAXITER);
  else
    return GSL_SUCCESS;
}

/* Large x asymptotic for  x^a U(a,b,x)
 * Based on SLATEC D9CHU() [W. Fullerton]
 *
 * Uses a rational approximation due to Luke.
 * See [Luke, Algorithms for the Computation of Special Functions, p. 252] [Luke, Utilitas Math. (1977)]
 *     
 * z^a U(a,b,z) ~ 2F0(a,1+a-b,-1/z)
 *
 * This assumes that a is not a negative integer and that 1+a-b is not a negative integer. If one of them
 * is, then the 2F0 actually terminates, the above relation is an equality, and the sum should be
 * evaluated directly [see below]. */ // Unchanged from gnu
static
int
d9chu(const double a, const double b, const double x, gsl_sf_result * result)
{
  const double EPS   = 8.0 * GSL_DBL_EPSILON;  /* EPS = 4.0D0*D1MACH(4)   */
  const int maxiter = 500;
  double aa[4], bb[4];
  int i;

  double bp = 1.0 + a - b;
  double ab = a*bp;
  double ct2 = 2.0 * (x - ab);
  double sab = a + bp;
  
  double ct3 = sab + 1.0 + ab;
  double anbn = ct3 + sab + 3.0;
  double ct1 = 1.0 + 2.0*x/anbn;

  bb[0] = 1.0;
  aa[0] = 1.0;

  bb[1] = 1.0 + 2.0*x/ct3;
  aa[1] = 1.0 + ct2/ct3;
  
  bb[2] = 1.0 + 6.0*ct1*x/ct3;
  aa[2] = 1.0 + 6.0*ab/anbn + 3.0*ct1*ct2/ct3;

  for(i=4; i<maxiter; i++) {
    int j;
    double c2;
    double d1z;
    double g1, g2, g3;
    double x2i1 = 2*i - 3;
    ct1   = x2i1/(x2i1-2.0);
    anbn += x2i1 + sab;
    ct2   = (x2i1 - 1.0)/anbn;
    c2    = x2i1*ct2 - 1.0;
    d1z   = 2.0*x2i1*x/anbn;
    
    ct3 = sab*ct2;
    g1  = d1z + ct1*(c2+ct3);
    g2  = d1z - c2;
    g3  = ct1*(1.0 - ct3 - 2.0*ct2);
    
    bb[3] = g1*bb[2] + g2*bb[1] + g3*bb[0];
    aa[3] = g1*aa[2] + g2*aa[1] + g3*aa[0];
    
    if(fabs(aa[3]*bb[0]-aa[0]*bb[3]) < EPS*fabs(bb[3]*bb[0])) break;
    
    for(j=0; j<3; j++) {
      aa[j] = aa[j+1];
      bb[j] = bb[j+1];
    }
  }
  
  result->val = aa[3]/bb[3];
  result->err = 8.0 * GSL_DBL_EPSILON * fabs(result->val);
  
  if(i == maxiter) {
    GSL_ERROR ("error", GSL_EMAXITER);
  }
  else {
    return GSL_SUCCESS;
  }
}

/* Evaluate asymptotic for z^a U(a,b,z) ~ 2F0(a,1+a-b,-1/z)
 * We check for termination of the 2F0 as a special case. Assumes x > 0.
 * Also assumes a,b are not too large compared to x. */ // Unchanged from gnu
static
int
hyperg_zaU_asymp(const double a, const double b, const double x, gsl_sf_result *result)
{
  const double ap = a;
  const double bp = 1.0 + a - b;
  const double rintap = floor(ap + 0.5);
  const double rintbp = floor(bp + 0.5);
  const int ap_neg_int = ( ap < 0.0 && fabs(ap - rintap) < INT_THRESHOLD );
  const int bp_neg_int = ( bp < 0.0 && fabs(bp - rintbp) < INT_THRESHOLD );

  if(ap_neg_int || bp_neg_int) {
    /* Evaluate 2F0 polynomial */
    double mxi = -1.0/x;
    double nmax = -(int)(GSL_MIN(ap,bp) - 0.1);
    double tn  = 1.0;
    double sum = 1.0;
    double n   = 1.0;
    double sum_err = 0.0;
    while(n <= nmax) {
      double apn = (ap+n-1.0);
      double bpn = (bp+n-1.0);
      tn  *= ((apn/n)*mxi)*bpn;
      sum += tn;
      sum_err += 2.0 * GSL_DBL_EPSILON * fabs(tn);
      n += 1.0;
    }
    result->val  = sum;
    result->err  = sum_err;
    result->err += 2.0 * GSL_DBL_EPSILON * (fabs(nmax)+1.0) * fabs(sum);
    return GSL_SUCCESS;
  }
  else {
    return d9chu(a,b,x,result);
  }
}


/* Evaluate finite sum which appears below. */ // Unchanged from gnu
static
int
hyperg_U_finite_sum(int N, double a, double b, double x, double xeps,
                    gsl_sf_result * result)
{
  int i;
  double sum_val;
  double sum_err;

  if(N <= 0) {
    double t_val = 1.0;
    double t_err = 0.0;
    gsl_sf_result poch;
    int stat_poch;

    sum_val = 1.0;
    sum_err = 0.0;
    for(i=1; i<= -N; i++) {
      const double xi1  = i - 1;
      const double mult = (a+xi1)*x/((b+xi1)*(xi1+1.0));
      t_val *= mult;
      t_err += fabs(mult) * t_err + fabs(t_val) * 8.0 * 2.0 * GSL_DBL_EPSILON;
      sum_val += t_val;
      sum_err += t_err;
    }

    stat_poch = gsl_sf_poch_e(1.0+a-b, -a, &poch);

    result->val  = sum_val * poch.val;
    result->err  = fabs(sum_val) * poch.err + sum_err * fabs(poch.val);
    result->err += fabs(poch.val) * (fabs(N) + 2.0) * GSL_DBL_EPSILON * fabs(sum_val);
    result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
    result->err *= 2.0; /* FIXME: fudge factor... why is the error estimate too small? */
    return stat_poch;
  }
  else {
    const int M = N - 2;
    if(M < 0) {
      result->val = 0.0;
      result->err = 0.0;
      return GSL_SUCCESS;
    }
    else {
      gsl_sf_result gbm1;
      gsl_sf_result gamr;
      int stat_gbm1;
      int stat_gamr;
      double t_val = 1.0;
      double t_err = 0.0;

      sum_val = 1.0;
      sum_err = 0.0;
      for(i=1; i<=M; i++) {
        const double mult = (a-b+i)*x/((1.0-b+i)*i);
        t_val *= mult;
        t_err += t_err * fabs(mult) + fabs(t_val) * 8.0 * 2.0 * GSL_DBL_EPSILON;
        sum_val += t_val;
        sum_err += t_err;
      }

      stat_gbm1 = gsl_sf_gamma_e(b-1.0, &gbm1);
      stat_gamr = gsl_sf_gammainv_e(a,  &gamr);

      if(stat_gbm1 == GSL_SUCCESS) {
        gsl_sf_result powx1N;
        int stat_p = gsl_sf_pow_int_e(x, 1-N, &powx1N);
        double pe_val = powx1N.val * xeps;
        double pe_err = powx1N.err * fabs(xeps) + 2.0 * GSL_DBL_EPSILON * fabs(pe_val);
        double coeff_val = gbm1.val * gamr.val * pe_val;
        double coeff_err = gbm1.err * fabs(gamr.val * pe_val)
                         + gamr.err * fabs(gbm1.val * pe_val)
                         + fabs(gbm1.val * gamr.val) * pe_err
                         + 2.0 * GSL_DBL_EPSILON * fabs(coeff_val);

        result->val  = sum_val * coeff_val;
        result->err  = fabs(sum_val) * coeff_err + sum_err * fabs(coeff_val);
        result->err += 2.0 * GSL_DBL_EPSILON * (M+2.0) * fabs(result->val);
        result->err *= 2.0; /* FIXME: fudge factor... why is the error estimate too small? */
        return stat_p;
      }
      else {
        result->val = 0.0;
        result->err = 0.0;
        return stat_gbm1;
      }
    }
  }
}


/* Based on SLATEC DCHU() [W. Fullerton]
 * Assumes x > 0.
 * This is just a series summation method, and it is not good for large a.
 * I patched up the window for 1+a-b near zero. [GJ] */ // Unchanged from gnu
static
int
hyperg_U_series(const double a, const double b, const double x, gsl_sf_result * result)
{
  const double EPS      = 2.0 * GSL_DBL_EPSILON;  /* EPS = D1MACH(3) */
  const double SQRT_EPS = M_SQRT2 * GSL_SQRT_DBL_EPSILON;

  if(fabs(1.0 + a - b) < SQRT_EPS) {
    /* Original Comment: ALGORITHM IS BAD WHEN 1+A-B IS NEAR ZERO FOR SMALL X */
    /* We can however do the following:
     * U(a,b,x) = U(a,a+1,x) when 1+a-b=0 and U(a,a+1,x) = x^(-a). */
    double lnr = -a * log(x);
    int stat_e =  gsl_sf_exp_e(lnr, result);
    result->err += 2.0 * SQRT_EPS * fabs(result->val);
    return stat_e;
  }
  else {
    double aintb = ( b < 0.0 ? ceil(b-0.5) : floor(b+0.5) );
    double beps  = b - aintb;
    int N = aintb;
    
    double lnx  = log(x);
    double xeps = exp(-beps*lnx);

    /* Evaluate finite sum. */
    gsl_sf_result sum;
    int stat_sum = hyperg_U_finite_sum(N, a, b, x, xeps, &sum);

    /* Evaluate infinite sum. */
    int istrt = ( N < 1 ? 1-N : 0 );
    double xi = istrt;

    gsl_sf_result gamr;
    gsl_sf_result powx;
    int stat_gamr = gsl_sf_gammainv_e(1.0+a-b, &gamr);
    int stat_powx = gsl_sf_pow_int_e(x, istrt, &powx);
    double sarg   = beps*M_PI;
    double sfact  = ( sarg != 0.0 ? sarg/sin(sarg) : 1.0 );
    double factor_val = sfact * ( GSL_IS_ODD(N) ? -1.0 : 1.0 ) * gamr.val * powx.val;
    double factor_err = fabs(gamr.val) * powx.err + fabs(powx.val) * gamr.err
                      + 2.0 * GSL_DBL_EPSILON * fabs(factor_val);

    gsl_sf_result pochai;
    gsl_sf_result gamri1;
    gsl_sf_result gamrni;
    int stat_pochai = gsl_sf_poch_e(a, xi, &pochai);
    int stat_gamri1 = gsl_sf_gammainv_e(xi + 1.0, &gamri1);
    int stat_gamrni = gsl_sf_gammainv_e(aintb + xi, &gamrni);
    int stat_gam123 = GSL_ERROR_SELECT_3(stat_gamr, stat_gamri1, stat_gamrni);
    int stat_gamall = GSL_ERROR_SELECT_4(stat_sum, stat_gam123, stat_pochai, stat_powx);

    gsl_sf_result pochaxibeps;
    gsl_sf_result gamrxi1beps;
    int stat_pochaxibeps = gsl_sf_poch_e(a, xi-beps, &pochaxibeps);
    int stat_gamrxi1beps = gsl_sf_gammainv_e(xi + 1.0 - beps, &gamrxi1beps);

    int stat_all = GSL_ERROR_SELECT_3(stat_gamall, stat_pochaxibeps, stat_gamrxi1beps);

    double b0_val = factor_val * pochaxibeps.val * gamrni.val * gamrxi1beps.val;
    double b0_err =  fabs(factor_val * pochaxibeps.val * gamrni.val) * gamrxi1beps.err
                   + fabs(factor_val * pochaxibeps.val * gamrxi1beps.val) * gamrni.err
                   + fabs(factor_val * gamrni.val * gamrxi1beps.val) * pochaxibeps.err
                   + fabs(pochaxibeps.val * gamrni.val * gamrxi1beps.val) * factor_err
                   + 2.0 * GSL_DBL_EPSILON * fabs(b0_val);

    if(fabs(xeps-1.0) < 0.5) {
      /* C  X**(-BEPS) IS CLOSE TO 1.0D0, SO WE MUST BE C  CAREFUL IN EVALUATING THE DIFFERENCES. */
      int i;
      gsl_sf_result pch1ai;
      gsl_sf_result pch1i;
      gsl_sf_result poch1bxibeps;
      int stat_pch1ai = gsl_sf_pochrel_e(a + xi, -beps, &pch1ai);
      int stat_pch1i  = gsl_sf_pochrel_e(xi + 1.0 - beps, beps, &pch1i);
      int stat_poch1bxibeps = gsl_sf_pochrel_e(b+xi, -beps, &poch1bxibeps);
      double c0_t1_val = beps*pch1ai.val*pch1i.val;
      double c0_t1_err = fabs(beps) * fabs(pch1ai.val) * pch1i.err
                       + fabs(beps) * fabs(pch1i.val)  * pch1ai.err
                       + 2.0 * GSL_DBL_EPSILON * fabs(c0_t1_val);
      double c0_t2_val = -poch1bxibeps.val + pch1ai.val - pch1i.val + c0_t1_val;
      double c0_t2_err =  poch1bxibeps.err + pch1ai.err + pch1i.err + c0_t1_err
                       + 2.0 * GSL_DBL_EPSILON * fabs(c0_t2_val);
      double c0_val = factor_val * pochai.val * gamrni.val * gamri1.val * c0_t2_val;
      double c0_err =  fabs(factor_val * pochai.val * gamrni.val * gamri1.val) * c0_t2_err
                     + fabs(factor_val * pochai.val * gamrni.val * c0_t2_val) * gamri1.err
                     + fabs(factor_val * pochai.val * gamri1.val * c0_t2_val) * gamrni.err
                     + fabs(factor_val * gamrni.val * gamri1.val * c0_t2_val) * pochai.err
                     + fabs(pochai.val * gamrni.val * gamri1.val * c0_t2_val) * factor_err
                     + 2.0 * GSL_DBL_EPSILON * fabs(c0_val);
      /* C  XEPS1 = (1.0 - X**(-BEPS))/BEPS = (X**(-BEPS) - 1.0)/(-BEPS) */
      gsl_sf_result dexprl;
      int stat_dexprl = gsl_sf_exprel_e(-beps*lnx, &dexprl);
      double xeps1_val = lnx * dexprl.val;
      double xeps1_err = 2.0 * GSL_DBL_EPSILON * (1.0 + fabs(beps*lnx)) * fabs(dexprl.val)
                       + fabs(lnx) * dexprl.err
                       + 2.0 * GSL_DBL_EPSILON * fabs(xeps1_val);
      double dchu_val = sum.val + c0_val + xeps1_val*b0_val;
      double dchu_err = sum.err + c0_err
                      + fabs(xeps1_val)*b0_err + xeps1_err * fabs(b0_val)
                      + fabs(b0_val*lnx)*dexprl.err
                      + 2.0 * GSL_DBL_EPSILON * (fabs(sum.val) + fabs(c0_val) + fabs(xeps1_val*b0_val));
      double xn = N;
      double t_val;
      double t_err;

      stat_all = GSL_ERROR_SELECT_5(stat_all, stat_dexprl, stat_poch1bxibeps, stat_pch1i, stat_pch1ai);

      for(i=1; i<2000; i++) {
        const double xi  = istrt + i;
        const double xi1 = istrt + i - 1;
        const double tmp = (a-1.0)*(xn+2.0*xi-1.0) + xi*(xi-beps);
        const double b0_multiplier = (a+xi1-beps)*x/((xn+xi1)*(xi-beps));
        const double c0_multiplier_1 = (a+xi1)*x/((b+xi1)*xi);
        const double c0_multiplier_2 = tmp / (xi*(b+xi1)*(a+xi1-beps));
        b0_val *= b0_multiplier;
        b0_err += fabs(b0_multiplier) * b0_err + fabs(b0_val) * 8.0 * 2.0 * GSL_DBL_EPSILON;
        c0_val  = c0_multiplier_1 * c0_val - c0_multiplier_2 * b0_val;
        c0_err  =  fabs(c0_multiplier_1) * c0_err
                 + fabs(c0_multiplier_2) * b0_err
                 + fabs(c0_val) * 8.0 * 2.0 * GSL_DBL_EPSILON
                 + fabs(b0_val * c0_multiplier_2) * 16.0 * 2.0 * GSL_DBL_EPSILON;
        t_val  = c0_val + xeps1_val*b0_val;
        t_err  = c0_err + fabs(xeps1_val)*b0_err;
        t_err += fabs(b0_val*lnx) * dexprl.err;
        t_err += fabs(b0_val)*xeps1_err;
        dchu_val += t_val;
        dchu_err += t_err;
        if(fabs(t_val) < EPS*fabs(dchu_val)) break;
      }

      result->val  = dchu_val;
      result->err  = 2.0 * dchu_err;
      result->err += 2.0 * fabs(t_val);
      result->err += 4.0 * GSL_DBL_EPSILON * (i+2.0) * fabs(dchu_val);
      result->err *= 2.0; /* FIXME: fudge factor */

      if(i >= 2000) {
        GSL_ERROR ("error", GSL_EMAXITER);
      }
      else {
        return stat_all;
      }
    }
    else {
      /* C  X**(-BEPS) IS VERY DIFFERENT FROM 1.0, SO THE C  STRAIGHTFORWARD FORMULATION IS STABLE. */
      int i;
      double dchu_val;
      double dchu_err;
      double t_val;
      double t_err;
      gsl_sf_result dgamrbxi;
      int stat_dgamrbxi = gsl_sf_gammainv_e(b+xi, &dgamrbxi);
      double a0_val = factor_val * pochai.val * dgamrbxi.val * gamri1.val / beps;
      double a0_err =  fabs(factor_val * pochai.val * dgamrbxi.val / beps) * gamri1.err
                     + fabs(factor_val * pochai.val * gamri1.val / beps) * dgamrbxi.err
                     + fabs(factor_val * dgamrbxi.val * gamri1.val / beps) * pochai.err
                     + fabs(pochai.val * dgamrbxi.val * gamri1.val / beps) * factor_err
                     + 2.0 * GSL_DBL_EPSILON * fabs(a0_val);
      stat_all = GSL_ERROR_SELECT_2(stat_all, stat_dgamrbxi);

      b0_val = xeps * b0_val / beps;
      b0_err = fabs(xeps / beps) * b0_err + 4.0 * GSL_DBL_EPSILON * fabs(b0_val);
      dchu_val = sum.val + a0_val - b0_val;
      dchu_err = sum.err + a0_err + b0_err
        + 2.0 * GSL_DBL_EPSILON * (fabs(sum.val) + fabs(a0_val) + fabs(b0_val));

      for(i=1; i<2000; i++) {
        double xi = istrt + i;
        double xi1 = istrt + i - 1;
        double a0_multiplier = (a+xi1)*x/((b+xi1)*xi);
        double b0_multiplier = (a+xi1-beps)*x/((aintb+xi1)*(xi-beps));
        a0_val *= a0_multiplier;
        a0_err += fabs(a0_multiplier) * a0_err;
        b0_val *= b0_multiplier;
        b0_err += fabs(b0_multiplier) * b0_err;
        t_val = a0_val - b0_val;
        t_err = a0_err + b0_err;
        dchu_val += t_val;
        dchu_err += t_err;
        if(fabs(t_val) < EPS*fabs(dchu_val)) break;
      }

      result->val  = dchu_val;
      result->err  = 2.0 * dchu_err;
      result->err += 2.0 * fabs(t_val);
      result->err += 4.0 * GSL_DBL_EPSILON * (i+2.0) * fabs(dchu_val);
      result->err *= 2.0; /* FIXME: fudge factor */

      if(i >= 2000) {
        GSL_ERROR ("error", GSL_EMAXITER);
      }
      else {
        return stat_all;
      }
    }
  }
}


/* Assumes b > 0 and x > 0. */ // Unchanged from gnu
static
int
hyperg_U_small_ab(const double a, const double b, const double x, gsl_sf_result * result)
{
  if(a == -1.0) {
    /* U(-1,c+1,x) = Laguerre[c,0,x] = -b + x
     */
    result->val  = -b + x;
    result->err  = 2.0 * GSL_DBL_EPSILON * (fabs(b) + fabs(x));
    result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
    return GSL_SUCCESS;
  }
  else if(a == 0.0) {
    /* U(0,c+1,x) = Laguerre[c,0,x] = 1
     */
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(ASYMP_EVAL_OK(a,b,x)) {
    double p = pow(x, -a);
    gsl_sf_result asymp;
    int stat_asymp = hyperg_zaU_asymp(a, b, x, &asymp);
    result->val  = asymp.val * p;
    result->err  = asymp.err * p;
    result->err += fabs(asymp.val) * GSL_DBL_EPSILON * fabs(a) * p;
    result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
    return stat_asymp;
  }
  else {
    return hyperg_U_series(a, b, x, result);
  }
}

/* We use this to keep track of large dynamic ranges in the recursions.This can be important because sometimes
 * we want to calculate a very large and a very small number and the answer is the product, of order 1. This happens,
 * for instance, when we apply a Kummer transform to make b positive and both x and b are large. */ // Unchanged
#define RESCALE_2(u0,u1,factor,count)      \
do {                                       \
  double au0 = fabs(u0);                   \
  if(au0 > factor) {                       \
    u0 /= factor;                          \
    u1 /= factor;                          \
    count++;                               \
  }                                        \
  else if(au0 < 1.0/factor) {              \
    u0 *= factor;                          \
    u1 *= factor;                          \
    count--;                               \
  }                                        \
} while (0)

/* Assumes b > 0 and x > 0. /* This is the one that formerly crashed now calls jsl_sf_hyperg_U_large_b_e */
static
int
hyperg_U_small_a_bgt0(const double a, const double b, const double x,
                      gsl_sf_result * result,
                      double * ln_multiplier )
{
  if(a == 0.0) {
    result->val = 1.0;
    result->err = 1.0;
    *ln_multiplier = 0.0;
    return GSL_SUCCESS;
  }
  else if(   (b > 5000.0 && x < 0.90 * fabs(b)) || (b >  500.0 && x < 0.50 * fabs(b))   ) {
	// The function below is changed - the old one was faulty
	int stat = jsl_sf_hyperg_U_large_b_e(a, b, x, result, ln_multiplier);
    //printf("(%s)",gsl_strerror(stat));
	if(stat == GSL_EOVRFLW)
      return GSL_SUCCESS;
    else
      return stat;
  }
  else if(b > 15.0) {
    /* Recurse up from b near 1.*/
    double eps = b - floor(b);
    double b0  = 1.0 + eps;
    gsl_sf_result r_Ubm1;
    gsl_sf_result r_Ub;
    int stat_0 = hyperg_U_small_ab(a, b0,     x, &r_Ubm1);
    int stat_1 = hyperg_U_small_ab(a, b0+1.0, x, &r_Ub);
    double Ubm1 = r_Ubm1.val;
    double Ub   = r_Ub.val;
    double Ubp1;
    double bp;

	// John added this
	int scale_count = 0;
	
    for(bp = b0+1.0; bp<b-0.1; bp += 1.0) {
      Ubp1 = ((1.0+a-bp)*Ubm1 + (bp+x-1.0)*Ub)/x;
      Ubm1 = Ub;
      Ub   = Ubp1;
      
	  // John added this
	  RESCALE_2(Ub,Ubm1,GSL_SQRT_DBL_MAX,scale_count);		
	}
    
	// John added this
	//*ln_multiplier = 0.0;
	*ln_multiplier = GSL_LOG_DBL_MAX / 2.0 * (double)scale_count;
	
	result->val  = Ub;
    result->err  = (fabs(r_Ubm1.err/r_Ubm1.val) + fabs(r_Ub.err/r_Ub.val)) * fabs(Ub);
    result->err += 2.0 * GSL_DBL_EPSILON * (fabs(b-b0)+1.0) * fabs(Ub);
    return GSL_ERROR_SELECT_2(stat_0, stat_1);
  }
  else {
    *ln_multiplier = 0.0;
    return hyperg_U_small_ab(a, b, x, result);
  }
}

/* This function has been substantially altered from gsl_sf_hyperg_U_large_b_e */
int
jsl_sf_hyperg_U_large_b_e(const double a, const double b, const double x,
                             gsl_sf_result * result,
                             double * ln_multiplier )
{
  double N   = floor(b);  /* b = N + eps */
  double eps = b - N;
  
  //printf("\nin the big B function");
  
  if(fabs(eps) < GSL_SQRT_DBL_EPSILON) { /* if b is close to an integer */
    double lnpre_val;
    double lnpre_err;
    gsl_sf_result M;
    if(b > 1.0) {
      double tmp = (1.0-b)*log(x);
      gsl_sf_result lg_bm1;
      gsl_sf_result lg_a;
      gsl_sf_lngamma_e(b-1.0, &lg_bm1);
      gsl_sf_lngamma_e(a, &lg_a);
      lnpre_val = tmp + x + lg_bm1.val - lg_a.val;
      lnpre_err = lg_bm1.err + lg_a.err + GSL_DBL_EPSILON * (fabs(x) + fabs(tmp));
      /* The approximate form commented out below gave poor results for large x 
	   * full function altered to avoid circularity (i.e. calculating U with M and then M with U */
	  //gsl_sf_hyperg_1F1_large_b_e(1.0-a, 2.0-b, -x, &M);
	  jsl_sf_hyperg_1F1_e(1.0-a, 2.0-b, -x, &M);
    }
    else {
      gsl_sf_result lg_1mb;
      gsl_sf_result lg_1pamb;
      gsl_sf_lngamma_e(1.0-b, &lg_1mb);
      gsl_sf_lngamma_e(1.0+a-b, &lg_1pamb);
      lnpre_val = lg_1mb.val - lg_1pamb.val;
      lnpre_err = lg_1mb.err + lg_1pamb.err;
      /* The approximate form commented out below gave poor results for large x 
	   * full function altered to avoid circularity (i.e. calculating U with M and then M with U */
	  //gsl_sf_hyperg_1F1_large_b_e(a, b, x, &M);
      jsl_sf_hyperg_1F1_e(a, b, x, &M);
	}

    if(lnpre_val > GSL_LOG_DBL_MAX-10.0) {
      result->val  = M.val;
      result->err  = M.err;
      *ln_multiplier = lnpre_val;
      GSL_ERROR ("overflow", GSL_EOVRFLW);
    }
    else {
      gsl_sf_result epre;
      int stat_e = gsl_sf_exp_err_e(lnpre_val, lnpre_err, &epre);
      result->val  = epre.val * M.val;
      result->err  = epre.val * M.err + epre.err * fabs(M.val);
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
      *ln_multiplier = 0.0;
      return stat_e;
    }
  }
  else { /* if b is not close to an integer */
    double omb_lnx = (1.0-b)*log(x);
    gsl_sf_result lg_1mb;    double sgn_1mb;
    gsl_sf_result lg_1pamb;  double sgn_1pamb;
    gsl_sf_result lg_bm1;    double sgn_bm1;
    gsl_sf_result lg_a;      double sgn_a;
    // I've added these
	gsl_sf_result lg_b;		 double sgn_b;
    gsl_sf_result lg_mbm1;	 double sgn_mbm1;
	gsl_sf_result lg_bp2;    double sgn_bp2;
    gsl_sf_result lg_2mb;    double sgn_2mb;
	
	gsl_sf_result M1, M2;
    double lnpre1_val, lnpre2_val;
    double lnpre1_err, lnpre2_err;
    double sgpre1, sgpre2;
    
	 /* The approximate form commented out below gave poor results for large x 
	   * full function altered to avoid circularity (i.e. calculating U with M and then M with U */
	//gsl_sf_hyperg_1F1_large_b_e(    a,     b,  x, &M1);
    //gsl_sf_hyperg_1F1_large_b_e(1.0+a-b, 2.0-b,x,&M2);
	jsl_sf_hyperg_1F1_e(a,           b,x,&M1);
	jsl_sf_hyperg_1F1_e(1.0+a-b, 2.0-b,x,&M2);
	//printf("%g %g\n",M1.val,M2.val);
	
	gsl_sf_lngamma_sgn_e(1.0+a-b, &lg_1pamb, &sgn_1pamb);
	gsl_sf_lngamma_sgn_e(b,		  &lg_b,     &sgn_b);
	gsl_sf_lngamma_sgn_e(a,		  &lg_a,     &sgn_a);
	gsl_sf_lngamma_sgn_e(2.0-b,   &lg_2mb,   &sgn_2mb);
	gsl_sf_lngamma_sgn_e(-b-1.0,  &lg_mbm1,  &sgn_mbm1);
	gsl_sf_lngamma_sgn_e(b+2.0,   &lg_bp2,   &sgn_bp2);
	
    lnpre1_val = lg_mbm1.val + lg_bp2.val - lg_1pamb.val - lg_b.val;
    lnpre1_err = lg_mbm1.err + lg_bp2.err + lg_1pamb.err + lg_b.err;
	sgpre1 =     sgn_mbm1    * sgn_bp2    * sgn_1pamb    * sgn_b; 
	
	lnpre2_val = lg_mbm1.val + lg_bp2.val - lg_2mb.val - lg_a.val + omb_lnx ;
	lnpre2_err = lg_mbm1.err + lg_bp2.err + lg_2mb.err + lg_a.err + GSL_DBL_EPSILON * fabs(omb_lnx);
    sgpre2 =     -sgn_mbm1   * sgn_bp2    * sgn_2mb    * sgn_a;
		
	//printf("\nM(%g,%g,%g) = %g, M(%g,%g,%g) = %g",a,b,x,M1.val,1.0+a-b,2.0-b,x,M2.val);
	//printf("\nlnpre1:%g lnpre2:%g t1!:%g t2!:%g",lnpre1_val,lnpre2_val,sgpre2*exp(lnpre2_val)*M2.val, sgpre1*exp(lnpre1_val)*M1.val);
					
	/* // This was incorrect, I think
	gsl_sf_hyperg_1F1_large_b_e(1.0-a, 2.0-b, x, &M2);
	gsl_sf_hyperg_1F1_large_b_e(    a,     b, x, &M1);
	gsl_sf_lngamma_sgn_e(1.0-b,   &lg_1mb,   &sgn_1mb);
    gsl_sf_lngamma_sgn_e(1.0+a-b, &lg_1pamb, &sgn_1pamb);
	gsl_sf_lngamma_sgn_e(b-1.0, &lg_bm1, &sgn_bm1);
    gsl_sf_lngamma_sgn_e(a,     &lg_a,   &sgn_a);

    lnpre1_val = lg_1mb.val - lg_1pamb.val;
    lnpre1_err = lg_1mb.err + lg_1pamb.err;
	lnpre2_val = lg_bm1.val - lg_a.val - omb_lnx - x;
    lnpre2_err = lg_bm1.err + lg_a.err + GSL_DBL_EPSILON * (fabs(omb_lnx)+fabs(x));
    sgpre1 = sgn_1mb * sgn_1pamb;
    sgpre2 = sgn_bm1 * sgn_a; */
	
	// This changed by John
    //if(lnpre1_val > GSL_LOG_DBL_MAX-10.0 || lnpre2_val > GSL_LOG_DBL_MAX-10.0) {
    
	double t1 = sgpre1*exp(lnpre1_val);
    double t2 = sgpre2*exp(lnpre2_val);
    result->val  = t1*M1.val + t2*M2.val;
	
	if ( gsl_finite(result->val * 1e5) ) {
	  result->err  = fabs(t1) * M1.err + fabs(t2)*M2.err;
      result->err += GSL_DBL_EPSILON * (exp(lnpre1_err)*fabs(t1*M1.val) + exp(lnpre2_err)*fabs(t2*M2.val));
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
      *ln_multiplier = 0.0;
      return GSL_SUCCESS;
	}
	else {
	  double max_lnpre_val = GSL_MAX(lnpre1_val,lnpre2_val);
      double max_lnpre_err = GSL_MAX(lnpre1_err,lnpre2_err);
      double lp1 = lnpre1_val - max_lnpre_val;
      double lp2 = lnpre2_val - max_lnpre_val;
      double t1  = sgpre1*exp(lp1);
      double t2  = sgpre2*exp(lp2);
      //printf("\nlp1:%g lp2:%g t1:%g t2:%g",lp1,lp2,t1,t2);
	  result->val  = t1*M1.val + t2*M2.val;
      result->err  = fabs(t1)*M1.err + fabs(t2)*M2.err;
      result->err += GSL_DBL_EPSILON * exp(max_lnpre_err) * (fabs(t1*M1.val) + fabs(t2*M2.val));
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
      *ln_multiplier = max_lnpre_val;
	  //printf("\nU(%g,%g,%g) = %g x 10^(%g)", a, b, x, result->val, M_LOG10E*max_lnpre_val );
	  GSL_ERROR ("overflow", GSL_EOVRFLW); // This is just ignored by the calling function
    }
  }
}

/****************************************************************************************************
 ****************************************************************************************************
 ****************************************************************************************************/
/* This is the function that is actually called - formerly known as "hyperg_U_bge1" */
int jsl_sf_hyperg_U_e10_e(const double a, const double b, const double x,
                         gsl_sf_result_e10 * result)
{
  
  // I've removed checks for a==0, a = -ve integer, a < 0
  if(ASYMP_EVAL_OK(a,b,x)) {
    //printf("\n(1)\n");
	const double ln_pre_val = -a*log(x);
    const double ln_pre_err = 2.0 * GSL_DBL_EPSILON * fabs(ln_pre_val);
    gsl_sf_result asymp;
    int stat_asymp = hyperg_zaU_asymp(a, b, x, &asymp);
    int stat_e = gsl_sf_exp_mult_err_e10_e(ln_pre_val, ln_pre_err,
                                              asymp.val, asymp.err,
                                              result);
    return GSL_ERROR_SELECT_2(stat_e, stat_asymp);
  }
  else if(fabs(a) <= 1.0) {
    //printf("\n(2)\n");
	gsl_sf_result rU;
    double ln_multiplier;
	// (2) crashes on the next line if b big and x small
    int stat_U = hyperg_U_small_a_bgt0(a, b, x, &rU, &ln_multiplier);
    int stat_e = gsl_sf_exp_mult_err_e10_e(ln_multiplier, 2.0*GSL_DBL_EPSILON*fabs(ln_multiplier), rU.val, rU.err, result);
    return GSL_ERROR_SELECT_2(stat_U, stat_e);
  }
  else if(SERIES_EVAL_OK(a,b,x)) {
    //printf("\n(3)\n");
	gsl_sf_result ser;
    const int stat_ser = hyperg_U_series(a, b, x, &ser);
    result->val = ser.val;
    result->err = ser.err;
    result->e10 = 0;
    return stat_ser;
  }
  else if(b >= 2*a + x) {
    //printf("\n(4)\n");
	/* Recurse forward from a near zero. Note that we cannot cross the singularity at 
     * the line b=a+1, because the only way we could be in that little wedge is if a < 1. But we
     * have already dealt with the small a case. */
    int scale_count = 0;
    const double a0 = a - floor(a);
    const double scale_factor = GSL_SQRT_DBL_MAX;
    double lnscale;
    double lm_0, lm_1, lm_max;
    gsl_sf_result r_Uam1;
    gsl_sf_result r_Ua;
    int stat_0 = hyperg_U_small_a_bgt0(a0-1.0, b, x, &r_Uam1, &lm_0);
    int stat_1 = hyperg_U_small_a_bgt0(a0,     b, x, &r_Ua,   &lm_1);
    int stat_e;
    gsl_sf_result lnm;
    gsl_sf_result y;
    double Uam1 = r_Uam1.val;
    double Ua   = r_Ua.val;
    double Uap1;
    double ap;
    lm_max = GSL_MAX(lm_0, lm_1);
    Uam1 *= exp(lm_0-lm_max);
    Ua   *= exp(lm_1-lm_max);

    for(ap=a0; ap<a-0.1; ap += 1.0) {
      Uap1 = -(Uam1 + (b-2.0*ap-x)*Ua)/(ap*(1.0+ap-b));
      Uam1 = Ua;
      Ua   = Uap1;
      RESCALE_2(Ua,Uam1,scale_factor,scale_count);
    }

    lnscale = log(scale_factor);
    lnm.val = lm_max + scale_count * lnscale;
    lnm.err = 2.0 * GSL_DBL_EPSILON * (fabs(lm_max) + fabs(scale_count * lnscale));
    y.val  = Ua;
    y.err  = fabs(r_Uam1.err/r_Uam1.val) * fabs(Ua);
    y.err += fabs(r_Ua.err/r_Ua.val) * fabs(Ua);
    y.err += 2.0 * GSL_DBL_EPSILON * (fabs(a-a0) + 1.0) * fabs(Ua);
    y.err *= fabs(lm_0-lm_max) + 1.0;
    y.err *= fabs(lm_1-lm_max) + 1.0;
    stat_e = gsl_sf_exp_mult_err_e10_e(lnm.val, lnm.err, y.val, y.err, result);
    return GSL_ERROR_SELECT_3(stat_e, stat_0, stat_1);
  }
  
  else {
    if(b <= x) {
      //printf("\n(5)\n");
	  /* Recurse backward to a near zero. */
      const double a0 = a - floor(a);
      const double scale_factor = GSL_SQRT_DBL_MAX;
      int scale_count = 0;
      gsl_sf_result lnm;
      gsl_sf_result y;
      double lnscale;
      double lm_0;
      double Uap1;
      double Ua;
      double Uam1;
      gsl_sf_result U0;
      double ap;
      double ru;
      double r;
      int CF1_count;
      int stat_CF1 = hyperg_U_CF1(a, b, 0, x, &ru, &CF1_count);
      int stat_U0;
      int stat_e;
      r = ru/a;
      Ua   = GSL_SQRT_DBL_MIN;
      Uap1 = r * Ua;
      for(ap=a; ap>a0+0.1; ap -= 1.0) {
        Uam1 = -((b-2.0*ap-x)*Ua + ap*(1.0+ap-b)*Uap1);
        Uap1 = Ua;
        Ua   = Uam1;
        RESCALE_2(Ua,Uap1,scale_factor,scale_count);
      }

      stat_U0 = hyperg_U_small_a_bgt0(a0, b, x, &U0, &lm_0);

      lnscale = log(scale_factor);
      lnm.val = lm_0 - scale_count * lnscale;
      lnm.err = 2.0 * GSL_DBL_EPSILON * (fabs(lm_0) + fabs(scale_count * lnscale));
      y.val  = GSL_SQRT_DBL_MIN*(U0.val/Ua);
      y.err  = GSL_SQRT_DBL_MIN*(U0.err/fabs(Ua));
      y.err += 2.0 * GSL_DBL_EPSILON * (fabs(a0-a) + CF1_count + 1.0) * fabs(y.val);
      stat_e = gsl_sf_exp_mult_err_e10_e(lnm.val, lnm.err, y.val, y.err, result);
      return GSL_ERROR_SELECT_3(stat_e, stat_U0, stat_CF1);
    }
    else { /* b > x*/
	  //printf("\n(6)\n");
	  /* Recurse backward to near the b=2a+x line, then forward from a near zero to get the normalization. */
      int scale_count_for = 0;
      int scale_count_bck = 0;
      const double scale_factor = GSL_SQRT_DBL_MAX;
      const double eps = a - floor(a);
      const double a0 = ( eps == 0.0 ? 1.0 : eps );
      const double a1 = a0 + ceil(0.5*(b-x) - a0);
      gsl_sf_result lnm;
      gsl_sf_result y;
      double lm_for;
      double lnscale;
      double Ua1_bck;
      double Ua1_for;
      int stat_for;
      int stat_bck;
      int stat_e;
      int CF1_count;

	  //six_ctr++;	
		
      {
      	/* Recurse back to determine U(a1,b), sans normalization. */
        double Uap1;
        double Ua;
        double Uam1;
        double ap;
        double ru;
        double r;
        int stat_CF1 = hyperg_U_CF1(a, b, 0, x, &ru, &CF1_count);
        r = ru/a;
        Ua   = GSL_SQRT_DBL_MIN;
        Uap1 = r * Ua;
		for(ap=a; ap>a1+0.1; ap -= 1.0) {
          Uam1 = -((b-2.0*ap-x)*Ua + ap*(1.0+ap-b)*Uap1);
          Uap1 = Ua;
          Ua   = Uam1;
          RESCALE_2(Ua,Uap1,scale_factor,scale_count_bck);
        }
        Ua1_bck = Ua;
        stat_bck = stat_CF1;
	  }
      {
        /* Recurse forward to determine U(a1,b) with absolute normalization. */
        gsl_sf_result r_Uam1;
        gsl_sf_result r_Ua;
        double lm_0, lm_1;
        // (6) crashes on the next line if b big, x small
		int stat_0 = hyperg_U_small_a_bgt0(a0-1.0, b, x, &r_Uam1, &lm_0);
        int stat_1 = hyperg_U_small_a_bgt0(a0,     b, x, &r_Ua,   &lm_1);
    	
		double Uam1 = r_Uam1.val;
        double Ua   = r_Ua.val;
        double Uap1;
        double ap;

        lm_for = GSL_MAX(lm_0, lm_1);
        Uam1 *= exp(lm_0 - lm_for);
        Ua   *= exp(lm_1 - lm_for);

        for(ap=a0; ap<a1-0.1; ap += 1.0) {
          Uap1 = -(Uam1 + (b-2.0*ap-x)*Ua)/(ap*(1.0+ap-b));
          Uam1 = Ua;
          Ua   = Uap1;
          RESCALE_2(Ua,Uam1,scale_factor,scale_count_for);
        }
        Ua1_for = Ua;
        stat_for = GSL_ERROR_SELECT_2(stat_0, stat_1);
      }

      lnscale = log(scale_factor);
      lnm.val = lm_for + (scale_count_for - scale_count_bck)*lnscale;
      lnm.err = 2.0 * GSL_DBL_EPSILON * (fabs(lm_for) + fabs(scale_count_for - scale_count_bck)*fabs(lnscale));
      y.val = GSL_SQRT_DBL_MIN*Ua1_for/Ua1_bck;
      y.err = 2.0 * GSL_DBL_EPSILON * (fabs(a-a0) + CF1_count + 1.0) * fabs(y.val);
      
	  // This added by John - sorts things out
	  if(y.val < GSL_DBL_MIN) {
		y.val = Ua1_for/Ua1_bck;
		lnm.val += GSL_LOG_DBL_MIN/2.0;
	  } 
	  	  
	  stat_e = gsl_sf_exp_mult_err_e10_e(lnm.val, lnm.err, y.val, y.err, result);
      return GSL_ERROR_SELECT_3(stat_e, stat_bck, stat_for);
    }
  }
}

/* PART 3 - hyperg_1F1 */
/****************************************************************************************************************************************
 ****************************************************************************************************************************************
 ****************************************************************************************************************************************/
/* specfunc/hyperg_1F1.c
/* Author:  G. Jungman */

/* Asymptotic result for 1F1(a, b, x)  x -> -Infinity. Assumes b-a != neg integer and b != neg integer. */
static
int
hyperg_1F1_asymp_negx(const double a, const double b, const double x,
                     gsl_sf_result * result)
{
  gsl_sf_result lg_b;
  gsl_sf_result lg_bma;
  double sgn_b;
  double sgn_bma;

  int stat_b   = gsl_sf_lngamma_sgn_e(b,   &lg_b,   &sgn_b);
  int stat_bma = gsl_sf_lngamma_sgn_e(b-a, &lg_bma, &sgn_bma);

  if(stat_b == GSL_SUCCESS && stat_bma == GSL_SUCCESS) {
    gsl_sf_result F;
    int stat_F = gsl_sf_hyperg_2F0_series_e(a, 1.0+a-b, -1.0/x, -1, &F);
    if(F.val != 0) {
      double ln_term_val = a*log(-x);
      double ln_term_err = 2.0 * GSL_DBL_EPSILON * (fabs(a) + fabs(ln_term_val));
      double ln_pre_val = lg_b.val - lg_bma.val - ln_term_val;
      double ln_pre_err = lg_b.err + lg_bma.err + ln_term_err;
      int stat_e = gsl_sf_exp_mult_err_e(ln_pre_val, ln_pre_err,
                                            sgn_bma*sgn_b*F.val, F.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_e, stat_F);
    }
    else {
      result->val = 0.0;
      result->err = 0.0;
      return stat_F;
    }
  }
  else {
    DOMAIN_ERROR(result);
  }
}

/********************************************************************/
/* Asymptotic result for 1F1(a, b, x)  x -> +Infinity Assumes b != neg integer and a != neg integer */
static
int
hyperg_1F1_asymp_posx(const double a, const double b, const double x,
                      gsl_sf_result * result)
{
  gsl_sf_result lg_b;
  gsl_sf_result lg_a;
  double sgn_b;
  double sgn_a;

  int stat_b = gsl_sf_lngamma_sgn_e(b, &lg_b, &sgn_b);
  int stat_a = gsl_sf_lngamma_sgn_e(a, &lg_a, &sgn_a);

  if(stat_a == GSL_SUCCESS && stat_b == GSL_SUCCESS) {
    gsl_sf_result F;
    int stat_F = gsl_sf_hyperg_2F0_series_e(b-a, 1.0-a, 1.0/x, -1, &F);
    if(stat_F == GSL_SUCCESS && F.val != 0) {
      double lnx = log(x);
      double ln_term_val = (a-b)*lnx;
      double ln_term_err = 2.0 * GSL_DBL_EPSILON * (fabs(a) + fabs(b)) * fabs(lnx)
                         + 2.0 * GSL_DBL_EPSILON * fabs(a-b);
      double ln_pre_val = lg_b.val - lg_a.val + ln_term_val + x;
      double ln_pre_err = lg_b.err + lg_a.err + ln_term_err + 2.0 * GSL_DBL_EPSILON * fabs(x);
      int stat_e = gsl_sf_exp_mult_err_e(ln_pre_val, ln_pre_err,
                                            sgn_a*sgn_b*F.val, F.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_e, stat_F);
    }
    else {
      result->val = 0.0;
      result->err = 0.0;
      return stat_F;
    }
  }
  else {
    DOMAIN_ERROR(result);
  }
}

/********************************************************************/
/* Asymptotic result for x < 2b-4a, 2b-4a large. [Abramowitz+Stegun, 13.5.21]
 * assumes 0 <= x/(2b-4a) <= 1 */
static
int
hyperg_1F1_large2bm4a(const double a, const double b, const double x, gsl_sf_result * result)
{
  double eta    = 2.0*b - 4.0*a;
  double cos2th = x/eta;
  double sin2th = 1.0 - cos2th;
  double th = acos(sqrt(cos2th));
  double pre_h  = 0.25*M_PI*M_PI*eta*eta*cos2th*sin2th;
  gsl_sf_result lg_b;
  int stat_lg = gsl_sf_lngamma_e(b, &lg_b);
  double t1 = 0.5*(1.0-b)*log(0.25*x*eta);
  double t2 = 0.25*log(pre_h);
  double lnpre_val = lg_b.val + 0.5*x + t1 - t2;
  double lnpre_err = lg_b.err + 2.0 * GSL_DBL_EPSILON * (fabs(0.5*x) + fabs(t1) + fabs(t2));
  double s1 = sin(a*M_PI);
  double s2 = sin(0.25*eta*(2.0*th - sin(2.0*th)) + 0.25*M_PI);
  double ser_val = s1 + s2;
  double ser_err = 2.0 * GSL_DBL_EPSILON * (fabs(s1) + fabs(s2));
  int stat_e = gsl_sf_exp_mult_err_e(lnpre_val, lnpre_err,
                                        ser_val, ser_err,
                                        result);
  return GSL_ERROR_SELECT_2(stat_e, stat_lg);
}

/********************************************************************/
/* Luke's rational approximation.
 * See [Luke, Algorithms for the Computation of Mathematical Functions, p.182]
 * Like the case of the 2F1 rational approximations, these are probably guaranteed to converge for x < 0, barring gross
 * numerical instability in the pre-asymptotic regime.  */
static
int
hyperg_1F1_luke(const double a, const double c, const double xin,
                gsl_sf_result * result)
{
  const double RECUR_BIG = 1.0e+50;
  const int nmax = 5000;
  int n = 3;
  const double x  = -xin;
  const double x3 = x*x*x;
  const double t0 = a/c;
  const double t1 = (a+1.0)/(2.0*c);
  const double t2 = (a+2.0)/(2.0*(c+1.0));
  double F = 1.0;
  double prec;

  double Bnm3 = 1.0;                                  /* B0 */
  double Bnm2 = 1.0 + t1 * x;                         /* B1 */
  double Bnm1 = 1.0 + t2 * x * (1.0 + t1/3.0 * x);    /* B2 */
 
  double Anm3 = 1.0;                                                      /* A0 */
  double Anm2 = Bnm2 - t0 * x;                                            /* A1 */
  double Anm1 = Bnm1 - t0*(1.0 + t2*x)*x + t0 * t1 * (c/(c+1.0)) * x*x;   /* A2 */

  while(1) {
    double npam1 = n + a - 1;
    double npcm1 = n + c - 1;
    double npam2 = n + a - 2;
    double npcm2 = n + c - 2;
    double tnm1  = 2*n - 1;
    double tnm3  = 2*n - 3;
    double tnm5  = 2*n - 5;
    double F1 =  (n-a-2) / (2*tnm3*npcm1);
    double F2 =  (n+a)*npam1 / (4*tnm1*tnm3*npcm2*npcm1);
    double F3 = -npam2*npam1*(n-a-2) / (8*tnm3*tnm3*tnm5*(n+c-3)*npcm2*npcm1);
    double E  = -npam1*(n-c-1) / (2*tnm3*npcm2*npcm1);

    double An = (1.0+F1*x)*Anm1 + (E + F2*x)*x*Anm2 + F3*x3*Anm3;
    double Bn = (1.0+F1*x)*Bnm1 + (E + F2*x)*x*Bnm2 + F3*x3*Bnm3;
    double r = An/Bn;

    prec = fabs((F - r)/F);
    F = r;

    if(prec < GSL_DBL_EPSILON || n > nmax) break;

    if(fabs(An) > RECUR_BIG || fabs(Bn) > RECUR_BIG) {
      An   /= RECUR_BIG;
      Bn   /= RECUR_BIG;
      Anm1 /= RECUR_BIG;
      Bnm1 /= RECUR_BIG;
      Anm2 /= RECUR_BIG;
      Bnm2 /= RECUR_BIG;
      Anm3 /= RECUR_BIG;
      Bnm3 /= RECUR_BIG;
    }
    else if(fabs(An) < 1.0/RECUR_BIG || fabs(Bn) < 1.0/RECUR_BIG) {
      An   *= RECUR_BIG;
      Bn   *= RECUR_BIG;
      Anm1 *= RECUR_BIG;
      Bnm1 *= RECUR_BIG;
      Anm2 *= RECUR_BIG;
      Bnm2 *= RECUR_BIG;
      Anm3 *= RECUR_BIG;
      Bnm3 *= RECUR_BIG;
    }

    n++;
    Bnm3 = Bnm2;
    Bnm2 = Bnm1;
    Bnm1 = Bn;
    Anm3 = Anm2;
    Anm2 = Anm1;
    Anm1 = An;
  }

  result->val  = F;
  result->err  = 2.0 * fabs(F * prec);
  result->err += 2.0 * GSL_DBL_EPSILON * (n-1.0) * fabs(F);

  return GSL_SUCCESS;
}

/********************************************************************/
/* Series for 1F1(1,b,x) b > 0 */
static
int
hyperg_1F1_1_series(const double b, const double x, gsl_sf_result * result)
{
  double sum_val = 1.0;
  double sum_err = 0.0;
  double term = 1.0;
  double n    = 1.0;
  while(fabs(term/sum_val) > 2.0*GSL_DBL_EPSILON) {
    term *= x/(b+n-1);
    sum_val += term;
    sum_err += 2.0 * 4.0 * GSL_DBL_EPSILON * fabs(term);
    n += 1.0;
  }
  result->val  = sum_val;
  result->err  = sum_err;
  result->err += 2.0 * fabs(term);
  return GSL_SUCCESS;
}

/********************************************************************/
/* 1F1(1,b,x) b >= 1, b integer */
static
int
hyperg_1F1_1_int(const int b, const double x, gsl_sf_result * result)
{
  if(b < 1) {
    DOMAIN_ERROR(result);
  }
  else if(b == 1) {
    return gsl_sf_exp_e(x, result);
  }
  else if(b == 2) {
    return gsl_sf_exprel_e(x, result);
  }
  else if(b == 3) {
    return gsl_sf_exprel_2_e(x, result);
  }
  else {
    return gsl_sf_exprel_n_e(b-1, x, result);
  }
}

/********************************************************************/
/* 1F1(1,b,x)
 * b >=1, b real
 * checked OK: [GJ] Thu Oct  1 16:46:35 MDT 1998  */
static
int
hyperg_1F1_1(const double b, const double x, gsl_sf_result * result)
{
  double ax = fabs(x);
  double ib = floor(b + 0.1);

  if(b < 1.0) {
    DOMAIN_ERROR(result);
  }
  else if(b == 1.0) {
    return gsl_sf_exp_e(x, result);
  }
  else if(b >= 1.4*ax) {
    return hyperg_1F1_1_series(b, x, result);
  }
  else if(fabs(b - ib) < _1F1_INT_THRESHOLD && ib < INT_MAX) {
    return hyperg_1F1_1_int((int)ib, x, result);
  }
  else if(x > 0.0) {
    if(x > 100.0 && b < 0.75*x) {
      return hyperg_1F1_asymp_posx(1.0, b, x, result);
    }
    else if(b < 1.0e+05) {
      /* Recurse backward on b, from a chosen offset point. For x > 0, which holds here, this should be a stable direction.  */
      const double off = ceil(1.4*x-b) + 1.0;
      double bp = b + off;
      gsl_sf_result M;
      int stat_s = hyperg_1F1_1_series(bp, x, &M);
      const double err_rat = M.err / fabs(M.val);
      while(bp > b+0.1) {
        /* M(1,b-1) = x/(b-1) M(1,b) + 1 */
        bp -= 1.0;
        M.val  = 1.0 + x/bp * M.val;
      }
      result->val  = M.val;
      result->err  = err_rat * fabs(M.val);
      result->err += 2.0 * GSL_DBL_EPSILON * (fabs(off)+1.0) * fabs(M.val);
      return stat_s;
    }
    else {
      return hyperg_1F1_large2bm4a(1.0, b, x, result);
    }
  }
  else {
    /* x <= 0 and b not large compared to |x| */
    if(ax < 10.0 && b < 10.0) {
      return hyperg_1F1_1_series(b, x, result);
    }
    else if(ax >= 100.0 && GSL_MAX_DBL(fabs(2.0-b),1.0) < 0.99*ax) {
      return hyperg_1F1_asymp_negx(1.0, b, x, result);
    }
    else {
      return hyperg_1F1_luke(1.0, b, x, result);
    }
  }
}

/********************************************************************/
/* 1F1(a,b,x)/Gamma(b) for b->0 [limit of Abramowitz+Stegun 13.3.7] */
static
int
hyperg_1F1_renorm_b0(const double a, const double x, gsl_sf_result * result)
{
  double eta = a*x;
  if(eta > 0.0) {
    double root_eta = sqrt(eta);
    gsl_sf_result I1_scaled;
    int stat_I = gsl_sf_bessel_I1_scaled_e(2.0*root_eta, &I1_scaled);
    if(I1_scaled.val <= 0.0) {
      result->val = 0.0;
      result->err = 0.0;
      return GSL_ERROR_SELECT_2(stat_I, GSL_EDOM);
    }
    else {
      const double lnr_val = 0.5*x + 0.5*log(eta) + fabs(x) + log(I1_scaled.val);
      const double lnr_err = GSL_DBL_EPSILON * (1.5*fabs(x) + 1.0) + fabs(I1_scaled.err/I1_scaled.val);
      return gsl_sf_exp_err_e(lnr_val, lnr_err, result);
    }
  }
  else if(eta == 0.0) {
    result->val = 0.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else {
    /* eta < 0 */
    double root_eta = sqrt(-eta);
    gsl_sf_result J1;
    int stat_J = gsl_sf_bessel_J1_e(2.0*root_eta, &J1);
    if(J1.val <= 0.0) {
      result->val = 0.0;
      result->err = 0.0;
      return GSL_ERROR_SELECT_2(stat_J, GSL_EDOM);
    }
    else {
      const double t1 = 0.5*x;
      const double t2 = 0.5*log(-eta);
      const double t3 = fabs(x);
      const double t4 = log(J1.val);
      const double lnr_val = t1 + t2 + t3 + t4;
      const double lnr_err = GSL_DBL_EPSILON * (1.5*fabs(x) + 1.0) + fabs(J1.err/J1.val);
      gsl_sf_result ex;
      int stat_e = gsl_sf_exp_err_e(lnr_val, lnr_err, &ex);
      result->val = -ex.val;
      result->err =  ex.err;
      return stat_e;
    }
  }
}

/********************************************************************/
/* 1F1'(a,b,x)/1F1(a,b,x)
 * Uses Gautschi's series transformation of the continued fraction. This is apparently the best
 * method for getting this ratio in the stable region. The convergence is monotone and supergeometric
 * when b > x. Assumes a >= -1. */
static
int
hyperg_1F1_CF1_p_ser(const double a, const double b, const double x, double * result)
{
  if(a == 0.0) {
    *result = 0.0;
    return GSL_SUCCESS;
  }
  else {
    const int maxiter = 5000;
    double sum  = 1.0;
    double pk   = 1.0;
    double rhok = 0.0;
    int k;
    for(k=1; k<maxiter; k++) {
      double ak = (a + k)*x/((b-x+k-1.0)*(b-x+k));
      rhok = -ak*(1.0 + rhok)/(1.0 + ak*(1.0+rhok));
      pk  *= rhok;
      sum += pk;
      if(fabs(pk/sum) < 2.0*GSL_DBL_EPSILON) break;
    }
    *result = a/(b-x) * sum;
    if(k == maxiter)
      GSL_ERROR ("error", GSL_EMAXITER);
    else
      return GSL_SUCCESS;
  }
}

/********************************************************************/
/* 1F1(a,b,x)  |a| <= 1, b > 0 */
static
int
hyperg_1F1_small_a_bgt0(const double a, const double b, const double x, gsl_sf_result * result)
{
  const double bma = b-a;
  const double oma = 1.0-a;
  const double ap1mb = 1.0+a-b;
  const double abs_bma = fabs(bma);
  const double abs_oma = fabs(oma);
  const double abs_ap1mb = fabs(ap1mb);

  const double ax = fabs(x);

  if(a == 0.0) {
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(a == 1.0 && b >= 1.0) {
    return hyperg_1F1_1(b, x, result);
  }
  else if(a == -1.0) {
    result->val  = 1.0 + a/b * x;
    result->err  = GSL_DBL_EPSILON * (1.0 + fabs(a/b * x));
    result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
    return GSL_SUCCESS;
  }
  else if(b >= 1.4*ax) {
    return gsl_sf_hyperg_1F1_series_e(a, b, x, result);
  }
  else if(x > 0.0) {
    if(x > 100.0 && abs_bma*abs_oma < 0.5*x) {
      return hyperg_1F1_asymp_posx(a, b, x, result);
    }
    else if(b < 5.0e+06) {
      /* Recurse backward on b from a suitably high point. */
      const double b_del = ceil(1.4*x-b) + 1.0;
      double bp = b + b_del;
      gsl_sf_result r_Mbp1;
      gsl_sf_result r_Mb;
      double Mbp1;
      double Mb;
      double Mbm1;
      int stat_0 = gsl_sf_hyperg_1F1_series_e(a, bp+1.0, x, &r_Mbp1);
      int stat_1 = gsl_sf_hyperg_1F1_series_e(a, bp,     x, &r_Mb);
      const double err_rat = fabs(r_Mbp1.err/r_Mbp1.val) + fabs(r_Mb.err/r_Mb.val);
      Mbp1 = r_Mbp1.val;
      Mb   = r_Mb.val;
      while(bp > b+0.1) {
        /* Do backward recursion. */
        Mbm1 = ((x+bp-1.0)*Mb - x*(bp-a)/bp*Mbp1)/(bp-1.0);
        bp -= 1.0;
        Mbp1 = Mb;
        Mb   = Mbm1;
      }
      result->val  = Mb;
      result->err  = err_rat * (fabs(b_del)+1.0) * fabs(Mb);
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(Mb);
      return GSL_ERROR_SELECT_2(stat_0, stat_1);
    }
    else {
      return hyperg_1F1_large2bm4a(a, b, x, result);
    }
  }
  else {
    /* x < 0 and b not large compared to |x| */
    if(ax < 10.0 && b < 10.0) {
      return gsl_sf_hyperg_1F1_series_e(a, b, x, result);
    }
    else if(ax >= 100.0 && GSL_MAX(abs_ap1mb,1.0) < 0.99*ax) {
      return hyperg_1F1_asymp_negx(a, b, x, result);
    }
    else {
      return hyperg_1F1_luke(a, b, x, result);
    }
  }
}

/********************************************************************/
/* 1F1(b+eps,b,x) |eps|<=1, b > 0  */
static
int
hyperg_1F1_beps_bgt0(const double eps, const double b, const double x, gsl_sf_result * result)
{
  if(b > fabs(x) && fabs(eps) < GSL_SQRT_DBL_EPSILON) {
    /* If b-a is very small and x/b is not too large we can
     * use this explicit approximation.
     *
     * 1F1(b+eps,b,x) = exp(ax/b) (1 - eps x^2 (v2 + v3 x + ...) + ...)
     *
     *   v2 = a/(2b^2(b+1))
     *   v3 = a(b-2a)/(3b^3(b+1)(b+2))
     *   ...
     *
     * See [Luke, Mathematical Functions and Their Approximations, p.292]
     *
     * This cannot be used for b near a negative integer or zero.
     * Also, if x/b is large the deviation from exp(x) behaviour grows. */
    double a = b + eps;
    gsl_sf_result exab;
    int stat_e = gsl_sf_exp_e(a*x/b, &exab);
    double v2 = a/(2.0*b*b*(b+1.0));
    double v3 = a*(b-2.0*a)/(3.0*b*b*b*(b+1.0)*(b+2.0));
    double v  = v2 + v3 * x;
    double f  = (1.0 - eps*x*x*v);
    result->val  = exab.val * f;
    result->err  = exab.err * fabs(f);
    result->err += fabs(exab.val) * GSL_DBL_EPSILON * (1.0 + fabs(eps*x*x*v));
    result->err += 4.0 * GSL_DBL_EPSILON * fabs(result->val);
    return stat_e;
  }
  else {
    /* Otherwise use a Kummer transformation to reduce it to the small a case. */
    gsl_sf_result Kummer_1F1;
    int stat_K = hyperg_1F1_small_a_bgt0(-eps, b, -x, &Kummer_1F1);
    if(Kummer_1F1.val != 0.0) {
      int stat_e = gsl_sf_exp_mult_err_e(x, 2.0*GSL_DBL_EPSILON*fabs(x),
                                            Kummer_1F1.val, Kummer_1F1.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_e, stat_K);
    }
    else {
      result->val = 0.0;
      result->err = 0.0;
      return stat_K;
    }
  }
}

/********************************************************************/
/* 1F1(a,2a,x) = Gamma(a + 1/2) E(x) (|x|/4)^(-a+1/2) scaled_I(a-1/2,|x|/2)
 * E(x) = exp(x) x > 0
 *      = 1      x < 0
 * a >= 1/2  */
static
int
hyperg_1F1_beq2a_pos(const double a, const double x, gsl_sf_result * result)
{
  if(x == 0.0) {
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else {
    gsl_sf_result I;
    int stat_I = gsl_sf_bessel_Inu_scaled_e(a-0.5, 0.5*fabs(x), &I);
    gsl_sf_result lg;
    int stat_g = gsl_sf_lngamma_e(a + 0.5, &lg);
    double ln_term   = (0.5-a)*log(0.25*fabs(x));
    double lnpre_val = lg.val + GSL_MAX_DBL(x,0.0) + ln_term;
    double lnpre_err = lg.err + GSL_DBL_EPSILON * (fabs(ln_term) + fabs(x));
    int stat_e = gsl_sf_exp_mult_err_e(lnpre_val, lnpre_err,
                                          I.val, I.err,
                                          result);
    return GSL_ERROR_SELECT_3(stat_e, stat_g, stat_I);
  }
}

/********************************************************************/
/* Handle the case of a and b both positive integers. Assumes a > 0 and b > 0. */
static
int
hyperg_1F1_ab_posint(const int a, const int b, const double x, gsl_sf_result * result)
{
  double ax = fabs(x);

  if(a == b) {
    return gsl_sf_exp_e(x, result);             /* 1F1(a,a,x) */
  }
  else if(a == 1) {
    return gsl_sf_exprel_n_e(b-1, x, result);   /* 1F1(1,b,x) */
  }
  else if(b == a + 1) {
    gsl_sf_result K;
    int stat_K = gsl_sf_exprel_n_e(a, -x, &K);  /* 1F1(1,1+a,-x) */
    int stat_e = gsl_sf_exp_mult_err_e(x, 2.0 * GSL_DBL_EPSILON * fabs(x),
                                          K.val, K.err,
                                          result);
    return GSL_ERROR_SELECT_2(stat_e, stat_K);
  }
  else if(a == b + 1) {
    gsl_sf_result ex;
    int stat_e = gsl_sf_exp_e(x, &ex);
    result->val  = ex.val * (1.0 + x/b);
    result->err  = ex.err * (1.0 + x/b);
    result->err += ex.val * GSL_DBL_EPSILON * (1.0 + fabs(x/b));
    result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
    return stat_e;
  }
  else if(a == b + 2) {
    gsl_sf_result ex;
    int stat_e = gsl_sf_exp_e(x, &ex);
    double poly  = (1.0 + x/b*(2.0 + x/(b+1.0)));
    result->val  = ex.val * poly;
    result->err  = ex.err * fabs(poly);
    result->err += ex.val * GSL_DBL_EPSILON * (1.0 + fabs(x/b) * (2.0 + fabs(x/(b+1.0))));
    result->err += 2.0 * GSL_DBL_EPSILON * fabs(result->val);
    return stat_e;
  }
  else if(b == 2*a) {
    return hyperg_1F1_beq2a_pos(a, x, result);  /* 1F1(a,2a,x) */
  }
  else if(   ( b < 10 && a < 10 && ax < 5.0 )
          || ( b > a*ax )
          || ( b > a && ax < 5.0 )
    ) {
    return gsl_sf_hyperg_1F1_series_e(a, b, x, result);
  }
  else if(b > a && b >= 2*a + x) {
    /* Use the Gautschi CF series, then recurse backward to a=0 for normalization.
     * This will work for either sign of x. */
    double rap;
    int stat_CF1 = hyperg_1F1_CF1_p_ser(a, b, x, &rap);
    double ra = 1.0 + x/a * rap;
    double Ma   = GSL_SQRT_DBL_MIN;
    double Map1 = ra * Ma;
    double Mnp1 = Map1;
    double Mn   = Ma;
    double Mnm1;
    int n;
    for(n=a; n>0; n--) {
      Mnm1 = (n * Mnp1 - (2*n-b+x) * Mn) / (b-n);
      Mnp1 = Mn;
      Mn   = Mnm1;
    }
    result->val = Ma/Mn;
    result->err = 2.0 * GSL_DBL_EPSILON * (fabs(a) + 1.0) * fabs(Ma/Mn);
    return stat_CF1;
  }
  else if(b > a && b < 2*a + x && b > x) {
    /* Use the Gautschi series representation of the continued fraction. Then recurse forward
     * to the a=b line for normalization. This will work for either sign of x, although we do need
     * to check for b > x, for when x is positive. */
    double rap;
    int stat_CF1 = hyperg_1F1_CF1_p_ser(a, b, x, &rap);
    double ra = 1.0 + x/a * rap;
    gsl_sf_result ex;
    int stat_ex;

    double Ma   = GSL_SQRT_DBL_MIN;
    double Map1 = ra * Ma;
    double Mnm1 = Ma;
    double Mn   = Map1;
    double Mnp1;
    int n;
    for(n=a+1; n<b; n++) {
      Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
      Mnm1 = Mn;
      Mn   = Mnp1;
    }

    stat_ex = gsl_sf_exp_e(x, &ex);  /* 1F1(b,b,x) */
    result->val  = ex.val * Ma/Mn;
    result->err  = ex.err * fabs(Ma/Mn);
    result->err += 4.0 * GSL_DBL_EPSILON * (fabs(b-a)+1.0) * fabs(result->val);
    return GSL_ERROR_SELECT_2(stat_ex, stat_CF1);
  }
  else if(x >= 0.0) {
    if(b < a) {
      /* The point b,b is below the b=2a+x line. Forward recursion on a from b,b+1 is possible.
       * Note that a > b + 1 as well, since we already tried a = b + 1.  */
      if(x + log(fabs(x/b)) < GSL_LOG_DBL_MAX-2.0) {
        double ex = exp(x);
        int n;
        double Mnm1 = ex;                 /* 1F1(b,b,x)   */
        double Mn   = ex * (1.0 + x/b);   /* 1F1(b+1,b,x) */
        double Mnp1;
        for(n=b+1; n<a; n++) {
          Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
          Mnm1 = Mn;
          Mn   = Mnp1;
        }
        result->val  = Mn;
        result->err  = (x + 1.0) * GSL_DBL_EPSILON * fabs(Mn);
        result->err *= fabs(a-b)+1.0;
        return GSL_SUCCESS;
      }
      else {
        OVERFLOW_ERROR(result);
      }
    }
    else {
      /* b > a
       * b < 2a + x 
       * b <= x (otherwise we would have finished above)
       *
       * Gautschi anomalous convergence region. However, we can recurse forward all the way from a=0,1 
	   * because we are always underneath the b=2a+x line. */
      gsl_sf_result r_Mn;
      double Mnm1 = 1.0;    /* 1F1(0,b,x) */
      double Mn;            /* 1F1(1,b,x)  */
      double Mnp1;
      int n;
      gsl_sf_exprel_n_e(b-1, x, &r_Mn);
      Mn = r_Mn.val;
      for(n=1; n<a; n++) {
        Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
        Mnm1 = Mn;
        Mn   = Mnp1;
      }
      result->val  = Mn;
      result->err  = fabs(Mn) * (1.0 + fabs(a)) * fabs(r_Mn.err / r_Mn.val);
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(Mn);
      return GSL_SUCCESS;
    }
  }
  else {
    /* x < 0
     * b < a (otherwise we would have tripped one of the above)  */
    if(a <= 0.5*(b-x) || a >= -x) {
      /* Gautschi continued fraction is in the anomalous region, so we must find another way. 
	   * We recurse down in b,from the a=b line.  */
      double ex = exp(x);
      double Manp1 = ex;
      double Man   = ex * (1.0 + x/(a-1.0));
      double Manm1;
      int n;
      for(n=a-1; n>b; n--) {
        Manm1 = (-n*(1-n-x)*Man - x*(n-a)*Manp1)/(n*(n-1.0));
        Manp1 = Man;
        Man = Manm1;
      }
      result->val  = Man;
      result->err  = (fabs(x) + 1.0) * GSL_DBL_EPSILON * fabs(Man);
      result->err *= fabs(b-a)+1.0;
      return GSL_SUCCESS;
    }
    else {
      /* Pick a0 such that b ~= 2a0 + x, then recurse down in b from a0,a0 to determine
       * the values near the line b=2a+x. Then recurse forward on a from a0. */
      int a0 = ceil(0.5*(b-x));
      double Ma0b;    /* M(a0,b)   */
      double Ma0bp1;  /* M(a0,b+1) */
      double Ma0p1b;  /* M(a0+1,b) */
      double Mnm1;
      double Mn;
      double Mnp1;
      int n;
      {
        double ex = exp(x);
        double Ma0np1 = ex;
        double Ma0n   = ex * (1.0 + x/(a0-1.0));
        double Ma0nm1;
        for(n=a0-1; n>b; n--) {
          Ma0nm1 = (-n*(1-n-x)*Ma0n - x*(n-a0)*Ma0np1)/(n*(n-1.0));
          Ma0np1 = Ma0n;
          Ma0n = Ma0nm1;
        }
        Ma0bp1 = Ma0np1;
        Ma0b   = Ma0n;
        Ma0p1b = (b*(a0+x)*Ma0b + x*(a0-b)*Ma0bp1)/(a0*b);
      }
      /* Initialise the recurrence correctly BJG */
      if (a0 >= a)
        { 
          Mn = Ma0b;
        }
      else if (a0 + 1>= a)
        {
          Mn = Ma0p1b;
        }
      else
        {
          Mnm1 = Ma0b;
          Mn   = Ma0p1b;
          for(n=a0+1; n<a; n++) {
            Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
            Mnm1 = Mn;
            Mn   = Mnp1;
          }
        }
      result->val  = Mn;
      result->err  = (fabs(x) + 1.0) * GSL_DBL_EPSILON *  fabs(Mn);
      result->err *= fabs(b-a)+1.0;
      return GSL_SUCCESS;
    }
  }
}

/********************************************************************/
/* Evaluate a <= 0, a integer, cases directly. (Polynomial; Horner)
 * When the terms are all positive, this
 * must work. We will assume this here. */
static
int
hyperg_1F1_a_negint_poly(const int a, const double b, const double x, gsl_sf_result * result)
{
  if(a == 0) {
    result->val = 1.0;
    result->err = 1.0;
    return GSL_SUCCESS;
  }
  else {
    int N = -a;
    double poly = 1.0;
    int k;
    for(k=N-1; k>=0; k--) {
      double t = (a+k)/(b+k) * (x/(k+1));
      double r = t + 1.0/poly;
      if(r > 0.9*GSL_DBL_MAX/poly) {
        OVERFLOW_ERROR(result);
      }
      else {
        poly *= r;  /* P_n = 1 + t_n P_{n-1} */
      }
    }
    result->val = poly;
    result->err = 2.0 * (sqrt(N) + 1.0) * GSL_DBL_EPSILON * fabs(poly);
    return GSL_SUCCESS;
  }
}

/********************************************************************/
/* Evaluate negative integer a case by relation
 * to Laguerre polynomials. This is more general than
 * the direct polynomial evaluation, but is safe
 * for all values of x.
 * 1F1(-n,b,x) = n!/(b)_n Laguerre[n,b-1,x]
 *             = n B(b,n) Laguerre[n,b-1,x]
 * assumes b is not a negative integer */
static
int
hyperg_1F1_a_negint_lag(const int a, const double b, const double x, gsl_sf_result * result)
{
  const int n = -a;

  gsl_sf_result lag;
  const int stat_l = gsl_sf_laguerre_n_e(n, b-1.0, x, &lag);
  if(b < 0.0) {
    gsl_sf_result lnfact;
    gsl_sf_result lng1;
    gsl_sf_result lng2;
    double s1, s2;
    const int stat_f  = gsl_sf_lnfact_e(n, &lnfact);
    const int stat_g1 = gsl_sf_lngamma_sgn_e(b + n, &lng1, &s1);
    const int stat_g2 = gsl_sf_lngamma_sgn_e(b, &lng2, &s2);
    const double lnpre_val = lnfact.val - (lng1.val - lng2.val);
    const double lnpre_err = lnfact.err + lng1.err + lng2.err
      + 2.0 * GSL_DBL_EPSILON * fabs(lnpre_val);
    const int stat_e = gsl_sf_exp_mult_err_e(lnpre_val, lnpre_err,
                                                s1*s2*lag.val, lag.err,
                                                result);
    return GSL_ERROR_SELECT_5(stat_e, stat_l, stat_g1, stat_g2, stat_f);
  }
  else {
    gsl_sf_result lnbeta;
    gsl_sf_lnbeta_e(b, n, &lnbeta);
    if(fabs(lnbeta.val) < 0.1) {
      /* As we have noted, when B(x,y) is near 1, evaluating log(B(x,y)) is not accurate.
       * Instead we evaluate B(x,y) directly. */
      const double ln_term_val = log(1.25*n);
      const double ln_term_err = 2.0 * GSL_DBL_EPSILON * ln_term_val;
      gsl_sf_result beta;
      int stat_b = gsl_sf_beta_e(b, n, &beta);
      int stat_e = gsl_sf_exp_mult_err_e(ln_term_val, ln_term_err,
                                            lag.val, lag.err,
                                            result);
      result->val *= beta.val/1.25;
      result->err *= beta.val/1.25;
      return GSL_ERROR_SELECT_3(stat_e, stat_l, stat_b);
    }
    else {
      /* B(x,y) was not near 1, so it is safe to use the logarithmic values.  */
      const double ln_n = log(n);
      const double ln_term_val = lnbeta.val + ln_n;
      const double ln_term_err = lnbeta.err + 2.0 * GSL_DBL_EPSILON * fabs(ln_n);
      int stat_e = gsl_sf_exp_mult_err_e(ln_term_val, ln_term_err,
                                            lag.val, lag.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_e, stat_l);
    }
  }
}

/* Assumes a <= -1,  b <= -1, and b <= a. */
static
int
hyperg_1F1_ab_negint(const int a, const int b, const double x, gsl_sf_result * result)
{
  if(x == 0.0) {
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(x > 0.0) {
    return hyperg_1F1_a_negint_poly(a, b, x, result);
  }
  else {
    /* Apply a Kummer transformation to make x > 0 so
     * we can evaluate the polynomial safely. Of course,
     * this assumes b <= a, which must be true for
     * a<0 and b<0, since otherwise the thing is undefined. */
    gsl_sf_result K;
    int stat_K = hyperg_1F1_a_negint_poly(b-a, b, -x, &K);
    int stat_e = gsl_sf_exp_mult_err_e(x, 2.0 * GSL_DBL_EPSILON * fabs(x),
                                          K.val, K.err,
                                          result);
    return GSL_ERROR_SELECT_2(stat_e, stat_K);
  }
}

/********************************************************************/
/* [Abramowitz+Stegun, 13.1.3]
 *
 * M(a,b,x) = Gamma(1+a-b)/Gamma(2-b) x^(1-b) *
 *            { Gamma(b)/Gamma(a) M(1+a-b,2-b,x) - (b-1) U(1+a-b,2-b,x) }
 *
 * b not an integer >= 2
 * a-b not a negative integer */
/*static
int
hyperg_1F1_U(const double a, const double b, const double x, gsl_sf_result * result)
{
  const double bp = 2.0 - b;
  const double ap = a - b + 1.0;

  gsl_sf_result lg_ap, lg_bp;
  double sg_ap;
  int stat_lg0 = gsl_sf_lngamma_sgn_e(ap, &lg_ap, &sg_ap);
  int stat_lg1 = gsl_sf_lngamma_e(bp, &lg_bp);
  int stat_lg2 = GSL_ERROR_SELECT_2(stat_lg0, stat_lg1);
  double t1 = (bp-1.0) * log(x);
  double lnpre_val = lg_ap.val - lg_bp.val + t1;
  double lnpre_err = lg_ap.err + lg_bp.err + 2.0 * GSL_DBL_EPSILON * fabs(t1);

  gsl_sf_result lg_2mbp, lg_1papmbp;
  double sg_2mbp, sg_1papmbp;
  int stat_lg3 = gsl_sf_lngamma_sgn_e(2.0-bp,    &lg_2mbp,    &sg_2mbp);
  int stat_lg4 = gsl_sf_lngamma_sgn_e(1.0+ap-bp, &lg_1papmbp, &sg_1papmbp);
  int stat_lg5 = GSL_ERROR_SELECT_2(stat_lg3, stat_lg4);
  double lnc1_val = lg_2mbp.val - lg_1papmbp.val;
  double lnc1_err = lg_2mbp.err + lg_1papmbp.err
                    + GSL_DBL_EPSILON * (fabs(lg_2mbp.val) + fabs(lg_1papmbp.val));

  gsl_sf_result M;
  gsl_sf_result_e10 U;
  int stat_F = gsl_sf_hyperg_1F1_e(ap, bp, x, &M);
  int stat_U = gsl_sf_hyperg_U_e10_e(ap, bp, x, &U);
  int stat_FU = GSL_ERROR_SELECT_2(stat_F, stat_U);

  gsl_sf_result_e10 term_M;
  int stat_e0 = gsl_sf_exp_mult_err_e10_e(lnc1_val, lnc1_err, sg_2mbp*sg_1papmbp*M.val, M.err, &term_M);
	
  const double ombp = 1.0 - bp;
  const double Uee_val = U.e10*M_LN10;
  const double Uee_err = 2.0 * GSL_DBL_EPSILON * fabs(Uee_val);
  const double Mee_val = term_M.e10*M_LN10;
  const double Mee_err = 2.0 * GSL_DBL_EPSILON * fabs(Mee_val);
  int stat_e1;
  /* Do a little dance with the exponential prefactors
   * to avoid overflows in intermediate results. */ /*
  if(Uee_val > Mee_val) {
	const double factorM_val = exp(Mee_val-Uee_val);
    const double factorM_err = 2.0 * GSL_DBL_EPSILON * (fabs(Mee_val-Uee_val)+1.0) * factorM_val;
    const double inner_val = term_M.val*factorM_val - ombp*U.val;
    const double inner_err =
        term_M.err*factorM_val + fabs(ombp) * U.err
      + fabs(term_M.val) * factorM_err
      + GSL_DBL_EPSILON * (fabs(term_M.val*factorM_val) + fabs(ombp*U.val));
    stat_e1 = gsl_sf_exp_mult_err_e(lnpre_val+Uee_val, lnpre_err+Uee_err, sg_ap*inner_val, inner_err, result);
	printf("\nOK Here %g (%g)(%g) [%g][%g]",(*result).val,lnpre_val+Uee_val, lnpre_err+Uee_err, sg_ap*inner_val, inner_err);  
	stat_e1 = gsl_sf_exp_mult_err_e(2447.43,0,1665.1,0,result);
	printf("\nOK Here %g",(*result).val);	
  }
  else {
    const double factorU_val = exp(Uee_val - Mee_val);
    const double factorU_err = 2.0 * GSL_DBL_EPSILON * (fabs(Mee_val-Uee_val)+1.0) * factorU_val;
    const double inner_val = term_M.val - ombp*factorU_val*U.val;
    const double inner_err =
        term_M.err + fabs(ombp*factorU_val*U.err)
      + fabs(ombp*factorU_err*U.val)
      + GSL_DBL_EPSILON * (fabs(term_M.val) + fabs(ombp*factorU_val*U.val));
    stat_e1 = gsl_sf_exp_mult_err_e(lnpre_val+Mee_val, lnpre_err+Mee_err,
                                       sg_ap*inner_val, inner_err,
                                       result);
  }
  return GSL_ERROR_SELECT_5(stat_e1, stat_e0, stat_FU, stat_lg5, stat_lg2);
} */

/********************************************************************/
/* Handle case of generic positive a, b.
 * Assumes b-a is not a negative integer. */
static
int
hyperg_1F1_ab_pos(const double a, const double b,
                  const double x,
                  gsl_sf_result * result)
{
  const double ax = fabs(x);

  if(   ( b < 10.0 && a < 10.0 && ax < 5.0 ) || ( b > a*ax ) || ( b > a && ax < 5.0 ) ) {
    return gsl_sf_hyperg_1F1_series_e(a, b, x, result);
  }
  else if(   x < -100.0 && GSL_MAX_DBL(fabs(a),1.0)*GSL_MAX_DBL(fabs(1.0+a-b),1.0) < 0.7*fabs(x) ) {
    /* Large negative x asymptotic.  */
    return hyperg_1F1_asymp_negx(a, b, x, result);
  }
  else if(   x > 100.0 && GSL_MAX_DBL(fabs(b-a),1.0)*GSL_MAX_DBL(fabs(1.0-a),1.0) < 0.7*fabs(x) ) {
    /* Large positive x asymptotic. */
    return hyperg_1F1_asymp_posx(a, b, x, result);
  }
  else if(fabs(b-a) <= 1.0) {
    /* Directly handle b near a. */
    return hyperg_1F1_beps_bgt0(a-b, b, x, result);  /* a = b + eps */
  }

  else if(b > a && b >= 2*a + x) {
    /* Use the Gautschi CF series, then recurse backward to a near 0 for normalization.
     * This will work for either sign of x. */ 
    double rap;
    int stat_CF1 = hyperg_1F1_CF1_p_ser(a, b, x, &rap);
    double ra = 1.0 + x/a * rap;

    double Ma   = GSL_SQRT_DBL_MIN;
    double Map1 = ra * Ma;
    double Mnp1 = Map1;
    double Mn   = Ma;
    double Mnm1;
    gsl_sf_result Mn_true;
    int stat_Mt;
    double n;
    for(n=a; n>0.5; n -= 1.0) {
      Mnm1 = (n * Mnp1 - (2.0*n-b+x) * Mn) / (b-n);
      Mnp1 = Mn;
      Mn   = Mnm1;
    }

    stat_Mt = hyperg_1F1_small_a_bgt0(n, b, x, &Mn_true);

    result->val  = (Ma/Mn) * Mn_true.val;
    result->err  = fabs(Ma/Mn) * Mn_true.err;
    result->err += 2.0 * GSL_DBL_EPSILON * (fabs(a)+1.0) * fabs(result->val);
    return GSL_ERROR_SELECT_2(stat_Mt, stat_CF1);
  }
  else if(b > a && b < 2*a + x && b > x) {
    /* Use the Gautschi series representation of
     * the continued fraction. Then recurse forward
     * to near the a=b line for normalization. This will
     * work for either sign of x, although we do need
     * to check for b > x, which is relevant when x is positive. */
    gsl_sf_result Mn_true;
    int stat_Mt;
    double rap;
    int stat_CF1 = hyperg_1F1_CF1_p_ser(a, b, x, &rap);
    double ra = 1.0 + x/a * rap;
    double Ma   = GSL_SQRT_DBL_MIN;
    double Mnm1 = Ma;
    double Mn   = ra * Mnm1;
    double Mnp1;
    double n;
    for(n=a+1.0; n<b-0.5; n += 1.0) {
      Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
      Mnm1 = Mn;
      Mn   = Mnp1;
    }
    stat_Mt = hyperg_1F1_beps_bgt0(n-b, b, x, &Mn_true);
    result->val  = Ma/Mn * Mn_true.val;
    result->err  = fabs(Ma/Mn) * Mn_true.err;
    result->err += 2.0 * GSL_DBL_EPSILON * (fabs(b-a)+1.0) * fabs(result->val);
    return GSL_ERROR_SELECT_2(stat_Mt, stat_CF1);
  }
  else if(x >= 0.0) {

    if(b < a) {
      /* Forward recursion on a from a=b+eps-1,b+eps. */
      double N   = floor(a-b);
      double eps = a - b - N;
      gsl_sf_result r_M0;
      gsl_sf_result r_M1;
      int stat_0 = hyperg_1F1_beps_bgt0(eps-1.0, b, x, &r_M0);
      int stat_1 = hyperg_1F1_beps_bgt0(eps,     b, x, &r_M1);
      double M0 = r_M0.val;
      double M1 = r_M1.val;

      double Mam1 = M0;
      double Ma   = M1;
      double Map1;
      double ap;
      double start_pair = fabs(M0) + fabs(M1);
      double minim_pair = GSL_DBL_MAX;
      double pair_ratio;
      double rat_0 = fabs(r_M0.err/r_M0.val);
      double rat_1 = fabs(r_M1.err/r_M1.val);
      for(ap=b+eps; ap<a-0.1; ap += 1.0) {
        Map1 = ((b-ap)*Mam1 + (2.0*ap-b+x)*Ma)/ap;
        Mam1 = Ma;
        Ma   = Map1;
        minim_pair = GSL_MIN_DBL(fabs(Mam1) + fabs(Ma), minim_pair);
      }
      pair_ratio = start_pair/minim_pair;
      result->val  = Ma;
      result->err  = 2.0 * (rat_0 + rat_1 + GSL_DBL_EPSILON) * (fabs(b-a)+1.0) * fabs(Ma);
      result->err += 2.0 * (rat_0 + rat_1) * pair_ratio*pair_ratio * fabs(Ma);
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(Ma);
      return GSL_ERROR_SELECT_2(stat_0, stat_1);
    }
    else {
      /* b > a
       * b < 2a + x 
       * b <= x
       *
       * Recurse forward on a from a=eps,eps+1. */
      double eps = a - floor(a);
      gsl_sf_result r_Mnm1;
      gsl_sf_result r_Mn;
      int stat_0 = hyperg_1F1_small_a_bgt0(eps,     b, x, &r_Mnm1);
      int stat_1 = hyperg_1F1_small_a_bgt0(eps+1.0, b, x, &r_Mn);
      double Mnm1 = r_Mnm1.val;
      double Mn   = r_Mn.val;
      double Mnp1;

      double n;
      double start_pair = fabs(Mn) + fabs(Mnm1);
      double minim_pair = GSL_DBL_MAX;
      double pair_ratio;
      double rat_0 = fabs(r_Mnm1.err/r_Mnm1.val);
      double rat_1 = fabs(r_Mn.err/r_Mn.val);
      for(n=eps+1.0; n<a-0.1; n++) {
        Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
        Mnm1 = Mn;
        Mn   = Mnp1;
        minim_pair = GSL_MIN_DBL(fabs(Mn) + fabs(Mnm1), minim_pair);
      }
      pair_ratio = start_pair/minim_pair;
      result->val  = Mn;
      result->err  = 2.0 * (rat_0 + rat_1 + GSL_DBL_EPSILON) * (fabs(a)+1.0) * fabs(Mn);
      result->err += 2.0 * (rat_0 + rat_1) * pair_ratio*pair_ratio * fabs(Mn);
      result->err += 2.0 * GSL_DBL_EPSILON * fabs(Mn);
      return GSL_ERROR_SELECT_2(stat_0, stat_1);
    }
  }
  else {
    /* x < 0
     * b < a  */

    if(a <= 0.5*(b-x) || a >= -x) {
      /* Recurse down in b, from near the a=b line, b=a+eps,a+eps-1.
       */
      double N   = floor(a - b);
      double eps = 1.0 + N - a + b;
      gsl_sf_result r_Manp1;
      gsl_sf_result r_Man;
      int stat_0 = hyperg_1F1_beps_bgt0(-eps,    a+eps,     x, &r_Manp1);
      int stat_1 = hyperg_1F1_beps_bgt0(1.0-eps, a+eps-1.0, x, &r_Man);
      double Manp1 = r_Manp1.val;
      double Man   = r_Man.val;
      double Manm1;

      double n;
      double start_pair = fabs(Manp1) + fabs(Man);
      double minim_pair = GSL_DBL_MAX;
      double pair_ratio;
      double rat_0 = fabs(r_Manp1.err/r_Manp1.val);
      double rat_1 = fabs(r_Man.err/r_Man.val);
      for(n=a+eps-1.0; n>b+0.1; n -= 1.0) {
        Manm1 = (-n*(1-n-x)*Man - x*(n-a)*Manp1)/(n*(n-1.0));
        Manp1 = Man;
        Man = Manm1;
        minim_pair = GSL_MIN_DBL(fabs(Manp1) + fabs(Man), minim_pair);
      }

      /* FIXME: this is a nasty little hack; there is some
         (transient?) instability in this recurrence for some
         values. I can tell when it happens, which is when
         this pair_ratio is large. But I do not know how to
         measure the error in terms of it. I guessed quadratic
         below, but it is probably worse than that.
         */
      pair_ratio = start_pair/minim_pair;
      result->val  = Man;
      result->err  = 2.0 * (rat_0 + rat_1 + GSL_DBL_EPSILON) * (fabs(b-a)+1.0) * fabs(Man);
      result->err *= pair_ratio*pair_ratio + 1.0;
      return GSL_ERROR_SELECT_2(stat_0, stat_1);
    }
    else {
      /* Pick a0 such that b ~= 2a0 + x, then
       * recurse down in b from a0,a0 to determine
       * the values near the line b=2a+x. Then recurse
       * forward on a from a0.
       */
      double epsa = a - floor(a);
      double a0   = floor(0.5*(b-x)) + epsa;
      double N    = floor(a0 - b);
      double epsb = 1.0 + N - a0 + b;
      double Ma0b;
      double Ma0bp1;
      double Ma0p1b;
      int stat_a0;
      double Mnm1;
      double Mn;
      double Mnp1;
      double n;
      double err_rat;
      {
        gsl_sf_result r_Ma0np1;
        gsl_sf_result r_Ma0n;
        int stat_0 = hyperg_1F1_beps_bgt0(-epsb,    a0+epsb,     x, &r_Ma0np1);
        int stat_1 = hyperg_1F1_beps_bgt0(1.0-epsb, a0+epsb-1.0, x, &r_Ma0n);
        double Ma0np1 = r_Ma0np1.val;
        double Ma0n   = r_Ma0n.val;
        double Ma0nm1;

        err_rat = fabs(r_Ma0np1.err/r_Ma0np1.val) + fabs(r_Ma0n.err/r_Ma0n.val);

        for(n=a0+epsb-1.0; n>b+0.1; n -= 1.0) {
          Ma0nm1 = (-n*(1-n-x)*Ma0n - x*(n-a0)*Ma0np1)/(n*(n-1.0));
          Ma0np1 = Ma0n;
          Ma0n = Ma0nm1;
        }
        Ma0bp1 = Ma0np1;
        Ma0b   = Ma0n;
        Ma0p1b = (b*(a0+x)*Ma0b+x*(a0-b)*Ma0bp1)/(a0*b); /* right-down hook */
        stat_a0 = GSL_ERROR_SELECT_2(stat_0, stat_1);
      }

      /* Initialise the recurrence correctly BJG */

      if (a0 >= a - 0.1)
        { 
          Mn = Ma0b;
        }
      else if (a0 + 1>= a - 0.1)
        {
          Mn = Ma0p1b;
        }
      else
        {
          Mnm1 = Ma0b;
          Mn   = Ma0p1b;

          for(n=a0+1.0; n<a-0.1; n += 1.0) {
            Mnp1 = ((b-n)*Mnm1 + (2*n-b+x)*Mn)/n;
            Mnm1 = Mn;
            Mn   = Mnp1;
          }
        }

      result->val = Mn;
      result->err = (err_rat + GSL_DBL_EPSILON) * (fabs(b-a)+1.0) * fabs(Mn);
      return stat_a0;
    }
  }
}

/********************************************************************/
/* Assumes b != integer
 * Assumes a != integer when x > 0
 * Assumes b-a != neg integer when x < 0 */
static
int
hyperg_1F1_ab_neg(const double a, const double b, const double x,
                  gsl_sf_result * result)
{
  const double bma = b - a;
  const double abs_x = fabs(x);
  const double abs_a = fabs(a);
  const double abs_b = fabs(b);
  const double size_a = GSL_MAX(abs_a, 1.0);
  const double size_b = GSL_MAX(abs_b, 1.0);
  const int bma_integer = ( bma - floor(bma+0.5) < _1F1_INT_THRESHOLD );

  if(   (abs_a < 10.0 && abs_b < 10.0 && abs_x < 5.0) || (b > 0.8*GSL_MAX(fabs(a),1.0)*fabs(x))  ) {
	return gsl_sf_hyperg_1F1_series_e(a, b, x, result);
  }
  else if(   x > 0.0 && size_b > size_a && size_a*log(M_E*x/size_b) < GSL_LOG_DBL_EPSILON+7.0 ) {
    /* Series terms are positive definite up until there is a sign change. But by then the
     * terms are small due to the last condition. */
	return gsl_sf_hyperg_1F1_series_e(a, b, x, result);
  }
  else if(   (abs_x < 5.0 && fabs(bma) < 10.0 && abs_b < 10.0) || (b > 0.8*GSL_MAX_DBL(fabs(bma),1.0)*abs_x)    ) {
    /* Use Kummer transformation to render series safe. */
    gsl_sf_result Kummer_1F1;
	int stat_K = gsl_sf_hyperg_1F1_series_e(bma, b, -x, &Kummer_1F1);
    int stat_e = gsl_sf_exp_mult_err_e(x, GSL_DBL_EPSILON * fabs(x),
                                      Kummer_1F1.val, Kummer_1F1.err,
                                      result);
    return GSL_ERROR_SELECT_2(stat_e, stat_K);
  }
  else if(   x < -30.0  && GSL_MAX_DBL(fabs(a),1.0)*GSL_MAX_DBL(fabs(1.0+a-b),1.0) < 0.99*fabs(x)   ) {
    /* Large negative x asymptotic.
     * Note that we do not check if b-a is a negative integer. */
    return hyperg_1F1_asymp_negx(a, b, x, result);
  }
  else if( x > 100.0 && GSL_MAX_DBL(fabs(bma),1.0)*GSL_MAX_DBL(fabs(1.0-a),1.0) < 0.99*fabs(x) ) {
    /* Large positive x asymptotic.
     * Note that we do not check if a is a negative integer. */
    return hyperg_1F1_asymp_posx(a, b, x, result);
  }
  else if(x > 0.0 && !(bma_integer && bma > 0.0)) {
    // JOHN CHANGED THIS SECTION
	// return hyperg_1F1_U(a, b, x, result);
  
	// BEGINNING OF ADDED
	gsl_sf_result myxresult;
	
	int stat_F = hyperg_1F1_luke(b-a, b, -x, result);
	int stat_e = gsl_sf_exp_e(x, &myxresult);
    
	if(stat_F == GSL_SUCCESS && stat_e == GSL_SUCCESS) {
      double lnr = log(fabs(result->val)) + x;
      if(lnr < GSL_LOG_DBL_MAX) {
        result->val *= myxresult.val;
		//printf("\n%g=M(%g,%g,%g)\n",result->val,a, b, x);	
		return GSL_SUCCESS;
      }
      else {
        //exit(1);
		result->val = GSL_POSINF; 
        GSL_ERROR ("overflow", GSL_EOVRFLW);
      }
    }
    else if(stat_F != GSL_SUCCESS) {
      return stat_F;
    }
    else {
      return stat_e;
    }
	// END OF ADDED THING 
  }
  else {
    /* FIXME:  if all else fails, try the series... BJG */
    if (x < 0.0) {
      /* Apply Kummer Transformation */
      int status = gsl_sf_hyperg_1F1_series_e(b-a, b, -x, result);
      double K_factor = exp(x);
      result->val *= K_factor;
      result->err *= K_factor;
      return status;
    } else {
	  int status = gsl_sf_hyperg_1F1_series_e(a, b, x, result);
      return status;
    }
  }
}

/********************************************************************/
int
jsl_sf_hyperg_1F1_int_e(const int a, const int b, const double x, gsl_sf_result * result)
{
  /* CHECK_POINTER(result) */
  if(x == 0.0) {
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(a == b) {
    return gsl_sf_exp_e(x, result);
  }
  else if(b == 0) {
    DOMAIN_ERROR(result);
  }
  else if(a == 0) {
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(b < 0 && (a < b || a > 0)) {
    /* Standard domain error due to singularity. */
    DOMAIN_ERROR(result);
  }
  else if(x > 100.0  && GSL_MAX_DBL(1.0,fabs(b-a))*GSL_MAX_DBL(1.0,fabs(1-a)) < 0.5 * x) {
    /* x -> +Inf asymptotic */
    return hyperg_1F1_asymp_posx(a, b, x, result);
  }
  else if(x < -100.0 && GSL_MAX_DBL(1.0,fabs(a))*GSL_MAX_DBL(1.0,fabs(1+a-b)) < 0.5 * fabs(x)) {
    /* x -> -Inf asymptotic */
    return hyperg_1F1_asymp_negx(a, b, x, result);
  }
  else if(a < 0 && b < 0) {
    return hyperg_1F1_ab_negint(a, b, x, result);
  }
  else if(a < 0 && b > 0) {
    /* Use Kummer to reduce it to the positive integer case.
     * Note that b > a, strictly, since we already trapped b = a. */
    gsl_sf_result Kummer_1F1;
    int stat_K = hyperg_1F1_ab_posint(b-a, b, -x, &Kummer_1F1);
    int stat_e = gsl_sf_exp_mult_err_e(x, GSL_DBL_EPSILON * fabs(x),
                                      Kummer_1F1.val, Kummer_1F1.err,
                                      result); 
    return GSL_ERROR_SELECT_2(stat_e, stat_K);
  }
  else {
    /* a > 0 and b > 0 */
    return hyperg_1F1_ab_posint(a, b, x, result);
  }
}
/************************************************************************************************************
 *								This is the main function actually called									*
 ************************************************************************************************************/
int
jsl_sf_hyperg_1F1_e(const double a, const double b, const double x,
                       gsl_sf_result * result )
{
  const double bma = b - a;
  const double rinta = floor(a + 0.5);
  const double rintb = floor(b + 0.5);
  const double rintbma = floor(bma + 0.5);
  const int a_integer   = ( fabs(a-rinta) < _1F1_INT_THRESHOLD && rinta > INT_MIN && rinta < INT_MAX );
  const int b_integer   = ( fabs(b-rintb) < _1F1_INT_THRESHOLD && rintb > INT_MIN && rintb < INT_MAX );
  const int bma_integer = ( fabs(bma-rintbma) < _1F1_INT_THRESHOLD && rintbma > INT_MIN && rintbma < INT_MAX );
  const int b_neg_integer   = ( b < -0.1 && b_integer );
  const int a_neg_integer   = ( a < -0.1 && a_integer );
  const int bma_neg_integer = ( bma < -0.1 &&  bma_integer );

  if(x == 0.0) {
    /* Testing for this before testing a and b
     * is somewhat arbitrary. The result is that
     * we have 1F1(a,0,0) = 1. */
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(b == 0.0) {
    DOMAIN_ERROR(result);
  }
  else if(a == 0.0) {
    result->val = 1.0;
    result->err = 0.0;
    return GSL_SUCCESS;
  }
  else if(a == b) {
    /* case: a==b; exp(x)
     * It's good to test exact equality now. We also test approximate equality later. */
    return gsl_sf_exp_e(x, result);
  }
  else if(fabs(b) < _1F1_INT_THRESHOLD) {
    /* Note that neither a nor b is zero, since we eliminated that with the above tests. */
    if(fabs(a) < _1F1_INT_THRESHOLD) {
      /* a and b near zero: 1 + a/b (exp(x)-1) */
      gsl_sf_result exm1;
      int stat_e = gsl_sf_expm1_e(x, &exm1);
      double sa = ( a > 0.0 ? 1.0 : -1.0 );
      double sb = ( b > 0.0 ? 1.0 : -1.0 );
      double lnab = log(fabs(a/b)); /* safe */
      gsl_sf_result hx;
      int stat_hx = gsl_sf_exp_mult_err_e(lnab, GSL_DBL_EPSILON * fabs(lnab),
                                             sa * sb * exm1.val, exm1.err,
                                             &hx);
      result->val = (hx.val == GSL_DBL_MAX ? hx.val : 1.0 + hx.val);  /* FIXME: excessive paranoia ? what is DBL_MAX+1 ?*/
      result->err = hx.err;
      return GSL_ERROR_SELECT_2(stat_hx, stat_e);
    }
    else {
      /* b near zero and a not near zero */
      const double m_arg = 1.0/(0.5*b);
      gsl_sf_result F_renorm;
      int stat_F = hyperg_1F1_renorm_b0(a, x, &F_renorm);
      int stat_m = gsl_sf_multiply_err_e(m_arg, 2.0 * GSL_DBL_EPSILON * m_arg,
                                            0.5*F_renorm.val, 0.5*F_renorm.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_m, stat_F);
    }
  }
  else if(a_integer && b_integer) {
    /* Check for reduction to the integer case. Relies on the arbitrary "near an integer" test. */
    return gsl_sf_hyperg_1F1_int_e((int)rinta, (int)rintb, x, result);
  }
  else if(b_neg_integer && !(a_neg_integer && a > b)) {
    /* Standard domain error due to uncancelled singularity. */
    DOMAIN_ERROR(result);
  }
  else if(a_neg_integer) {
    return hyperg_1F1_a_negint_lag((int)rinta, b, x, result);
  }
  else if(b > 0.0) {
    if(-1.0 <= a && a <= 1.0) {
      /* Handle small a explicitly. */
      return hyperg_1F1_small_a_bgt0(a, b, x, result);
    }
    else if(bma_neg_integer) {
      /* Catch this now, to avoid problems in the generic evaluation code. */
      gsl_sf_result Kummer_1F1;
      int stat_K = hyperg_1F1_a_negint_lag((int)rintbma, b, -x, &Kummer_1F1);
      int stat_e = gsl_sf_exp_mult_err_e(x, GSL_DBL_EPSILON * fabs(x),
                                            Kummer_1F1.val, Kummer_1F1.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_e, stat_K);
    }
    else if(a < 0.0) {
      /* Use Kummer to reduce it to the generic positive case. Note that b > a, strictly, since we already 
	  trapped b = a. Also b-(b-a)=a, and a is not a negative integer here, so the generic evaluation is safe. */
      gsl_sf_result Kummer_1F1;
	  int stat_K = hyperg_1F1_ab_pos(b-a, b, -x, &Kummer_1F1);
      int stat_e = gsl_sf_exp_mult_err_e( x, GSL_DBL_EPSILON * fabs(x), Kummer_1F1.val, Kummer_1F1.err, result );
      return GSL_ERROR_SELECT_2(stat_e, stat_K);
    }
    else {
      /* a > 0.0 */
      return hyperg_1F1_ab_pos(a, b, x, result);
    }
  }
  else { /* b < 0.0 */
    if(bma_neg_integer && x < 0.0) {
      /* Handle this now to prevent problems in the generic evaluation. */
      gsl_sf_result K;
      int stat_K;
      int stat_e;
      if(a < 0.0) {
        /* Kummer transformed version of safe polynomial. The condition a < 0 is equivalent to b < b-a,
         * which is the condition required for the series to be positive definite here. */
		stat_K = hyperg_1F1_a_negint_poly((int)rintbma, b, -x, &K);
      }
      else {
        /* Generic eval for negative integer a. */
        stat_K = hyperg_1F1_a_negint_lag((int)rintbma, b, -x, &K);
      }
      stat_e = gsl_sf_exp_mult_err_e(x, GSL_DBL_EPSILON * fabs(x),
                                        K.val, K.err,
                                        result);
      return GSL_ERROR_SELECT_2(stat_e, stat_K);
    }
    else if(a > 0.0) {
      /* Use Kummer to reduce it to the generic negative case.*/
      gsl_sf_result K;
      int stat_K = hyperg_1F1_ab_neg(b-a, b, -x, &K);
      int stat_e = gsl_sf_exp_mult_err_e(x, GSL_DBL_EPSILON * fabs(x),
                                            K.val, K.err,
                                            result);
      return GSL_ERROR_SELECT_2(stat_e, stat_K);
    }
    else {
      // This is where we used to end up with the error JOHN
	  return hyperg_1F1_ab_neg(a, b, x, result);
    }
  }
}


#if 0  
    /* Luke in the canonical case. */
  if(x < 0.0 && !a_neg_integer && !bma_neg_integer) {
    double prec;
    return hyperg_1F1_luke(a, b, x, result, &prec);
  }

  /* Luke with Kummer transformation.  */
  if(x > 0.0 && !a_neg_integer && !bma_neg_integer) {
    double prec;
    double Kummer_1F1;
    double ex;
    int stat_F = hyperg_1F1_luke(b-a, b, -x, &Kummer_1F1, &prec);
    int stat_e = gsl_sf_exp_e(x, &ex);
    if(stat_F == GSL_SUCCESS && stat_e == GSL_SUCCESS) {
      double lnr = log(fabs(Kummer_1F1)) + x;
      if(lnr < GSL_LOG_DBL_MAX) {
        *result = ex * Kummer_1F1;
        return GSL_SUCCESS;
      }
      else {
        *result = GSL_POSINF; 
        GSL_ERROR ("overflow", GSL_EOVRFLW);
      }
    }
    else if(stat_F != GSL_SUCCESS) {
      *result = 0.0;
      return stat_F;
    }
    else {
      *result = 0.0;
      return stat_e;
    }
  }
#endif

/* PART 4 - main */
/****************************************************************************************************************************************
 ****************************************************************************************************************************************
 ****************************************************************************************************************************************/
int main(int argc, char *argv[])
{
    /* For command line arguments */
	int choice;
    char *end;
    FILE *in;
	FILE *out;
	
	/* For permutation (if necessary)*/
	const gsl_rng_type * my_gsl_rng_type;
    gsl_rng * my_gsl_rng;
	
	/* mcmc parameters and default values */ 
	long max_rounds = 20000;	/* Max number of rounds */
    double t0 = 1.0;			/* initial temperture */
	double dt = 0.5;			/* Cooling factor */ 
	int tscale = 500;			/* Cool after this many rounds */
	
    struct dataset data, boot_data;
    double *params, *step_sizes;
    double *mins, *maxs;
    int *indices;
    mcmc_log_likelihood_func_t likelihood_func = calc_log_likelihood;
    double param_start_mins[NUM_PARAM_TYPES];
	double param_start_maxs[NUM_PARAM_TYPES];
	double param_mins[NUM_PARAM_TYPES];
	double param_maxs[NUM_PARAM_TYPES];
	int * new_n_vals;
	
	/* Permutation */
	long num_permuts;
	int * flags_original;
	int * flags_randomised;
	int * class_size;
	int ** class_list;
	
	/* Flags and other params */
	double fixed_alpha_val = 1.0; /* Only acceptable if < 1.0 */
	int D_EXCLUDE_FLAG;			  /* Exclude all genes with this flag */
	int D_ONLY_FLAG;			  /* Keep only genes with this flag */
	int x = -1;					  /* exclude all genes with <= x neutral polymorphisms */
	int max_observed = 0;
	int max_alleles = 0;
    int i,j;
	int no_conv = 0; /* Have we converged OK? */
			
	/* Determine the name of the executable */
    if ((progname = strrchr(argv[0], '/')) == NULL) {
        progname = argv[0];
    } else {
        progname++;
    }

	/* turn off gsl's error handler */
	gsl_set_error_handler_off();
	
	/* Set up the permutation random number generator*/
	gsl_rng_env_setup();
    my_gsl_rng_type = gsl_rng_default;
    my_gsl_rng = gsl_rng_alloc (my_gsl_rng_type);
	
	/* Set the output defaults */ 
    out = stdout;
    
	/* likelihood model defaults */ 
	data.models[P_THETA]	= MODEL_UNIVERSAL;
	data.models[P_TIME]		= MODEL_UNIVERSAL;
	data.models[P_OMEGA]	= MODEL_UNIVERSAL;
	data.models[P_GAMMA_D]	= MODEL_NONE;
	data.models[P_ALPHA]	= ALPHA_FREE;
	
	M_BETA_IS_ONE = 1;					/* Shape parameter = 1 (i.e. exponential) */ 
	M_DIST_IS_REFLECTED = 1;			/* Partially reflect dist. of fitness effects */
	M_CORRECTION_COEFF = 1;				/* Measure divergence between a pair of extant taxa */
	D_ONLY_FLAG = D_EXCLUDE_FLAG = -1;	/* Read in all genes */
	BEW = 0;							/* Include sampling corrections as a default */
	A_NOT_ALPHA = 0;					/* Calculate alpha as default */
	num_permuts = 0;					/* Used for comaprison of models a=4 and a=1 */
	
    /* Parse the command-line options */
    while ((choice = getopt(argc, argv, OPTIONS)) != EOF) {
		switch (choice)
        {
            /*case 'G':
				M_BETA_IS_ONE = 0;
				break;*/
			case 'A':
				A_NOT_ALPHA = 1;
				break;
			case 'K':
				D_ONLY_FLAG = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid flag entered -O: %s\n", progname, optarg);
                }
                break;
			case 'B':
				BEW = 1;
				break;
			case 'P':
				data.models[P_ALPHA] = ALPHA_MANY;
				num_permuts = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid number of flag permutations specified: %s\n", progname, optarg);
                }
				break;
			case 'X':
				D_EXCLUDE_FLAG = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid flag to exclude -X: %s\n", progname, optarg);
                }
                break;
			case 'L':
				M_CORRECTION_COEFF = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid correction for common ancestor -L: %s\n", progname, optarg);
                }
				break;
			case 'R':
				M_DIST_IS_REFLECTED = 0;
				break;
			case 'h':
                usage();
                exit(0);
			case 'g':
				data.models[P_GAMMA_D] = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid gamma model number: %s\n", progname, optarg);
                }
                break;
            case 'f':
				data.models[P_OMEGA] = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid constraint model number: %s\n", progname, optarg);
                }
                break;
			case 't':
				data.models[P_TIME] = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid synonymous divergence model number: %s\n", progname, optarg);
                }
                break;
			case 'p':
				data.models[P_THETA] = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid synonymous diversity model number: %s\n", progname, optarg);
                }
                break;
			case 'a':
				data.models[P_ALPHA] = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: invalid alpha model number: %s\n", progname, optarg);
                }
                break;
            case 'k':
				single_gene = strtol(optarg, &end, 10);
				if (*end != '\0') {
                    fprintf(stderr, "%s: odd gene number entered: %s\n", progname, optarg);
                }
                break;
			case 'x':
				single_gene = strtol(optarg, &end, 10);
				single_gene = -single_gene;
				if (*end != '\0') {
                    fprintf(stderr, "%s: odd gene number entered: %s\n", progname, optarg);
                }
                break;
			case 'c':
				fixed_alpha_val = strtod(optarg, &end);
                if (*end != '\0') {
                    fprintf(stderr, "%s: invalid fixed alpha value: %s\n", progname, optarg);
                    exit(1);
                }
				break;
			case 'o':
                /* Open the output file */
                if ((out = fopen(optarg, "w")) == NULL) {
                    fprintf(stderr, "%s: unable to write to file %s: %s\n",
                            progname, optarg, strerror(errno));
                    exit(1);
                }
                break;
            case 'i':
                max_rounds = strtol(optarg, &end, 0);
                if (*end != '\0') {
                    fprintf(stderr, "%s: invalid maximum number of rounds: %s\n", progname, optarg);
                    exit(1);
                }
                break;
           case 'm':
                x = strtol(optarg, &end, 0);
                if (*end != '\0') {
                    fprintf(stderr, "%s: invalid exclusion limit: %s\n", progname, optarg);
                    exit(1);
                }
                break;
            case 'r':
                dt = strtod(optarg, &end);
                if (*end != '\0' || dt > 1.0 || dt < 0.01) {
                    fprintf(stderr, "%s: invalid cooling constant: %s\n", progname, optarg);
                    exit(1);
                }
                break;
            case 'v':
                printf("%s (" PACKAGE " version " VERSION ")\n", progname);
                exit(0);
            case '?':
                usage();
                exit(1);
           default:
                fprintf(stderr, "internal error: unhandle option -%c\n",choice);
                exit(1);
        }
    }

    /* Make sure an input file is given */
    if (optind >= argc) {
        fprintf(stderr, "%s: no input file provided\n", progname);
        usage();
        exit(1);
    }
    /* Try to open it */
    if ((in = fopen(argv[optind], "r")) == NULL) {
        fprintf(stderr, "%s: unable to open input file %s: %s\n", progname, argv[optind], strerror(errno));
        exit(1);
    }
    
	/* Read our data from the input file */
	read_data(in, argv[optind], &data.count, &data.items, &data.flag_count, &data.flag_list, x, D_ONLY_FLAG, D_EXCLUDE_FLAG);
    printf("\nread data from %d loci from %s", data.count, argv[optind]);
    if(x>=0)
		printf(" with %d or more segregating synonymous polymorphisms",x);
	if(data.models[P_ALPHA] == ALPHA_MANY) {
		printf(", placed in %d different classes: ", data.flag_count);
		for(i=0;i<data.flag_count;i++) {
			printf(" %d ", data.flag_list[i]);
		}
	}
	printf("\n");
	
	/* And then close it */
    if (fclose(in) == EOF) {
        fprintf(stderr, "%s: error closing %s: %s\n", progname, argv[optind], strerror(errno));
        exit(1);
    }
    optind++;

    /* Complain about extra arguments */
    if (optind < argc) {
        fprintf(stderr, "%s: ignoring superfluous arguments: %s...\n",
                progname, argv[optind]);
    }

	/* Seed the random number generator with the process id and current time */
   srand48((long)getpid() * (long)time(NULL));
	
	/* Do SEW estimators etc if required then exit */
	if ( data.models[P_ALPHA] == 999 ) {
		do_Smith_Eyre_Walker( &data );
		exit(0);
	}
	
	/* Note largest value for Pn/Pi/Dn/Di so we can build a log factorial table */
    for (i = 0; i < data.count; i++) {
        if (max_observed < data.items[i].measured[P_Pi]) {
            max_observed = data.items[i].measured[P_Pi];
        }
        if (max_observed < data.items[i].measured[P_Pn]) {
            max_observed = data.items[i].measured[P_Pn];
        }
		if (max_observed < data.items[i].measured[P_Di]) {
            max_observed = data.items[i].measured[P_Di];
        }
		if (max_observed < data.items[i].measured[P_Dn]) {
            max_observed = data.items[i].measured[P_Dn];
        }
		if (max_alleles < data.items[i].alleles) {
			max_alleles = data.items[i].alleles;
		}
	}

    /* Allocate room for the log factorial table and create */
    if ((log_factorial = calloc(max_observed + 1, sizeof(double))) == NULL) {
        perror("Ack: unable to allocate log factorial table\n");
        exit(1);
    }
    log_factorial[0] = 0.0;
    for (i = 1; i <= max_observed; i++) {
        log_factorial[i] = log_factorial[i - 1] + log(i);
    }
	
	/* Allocate room for sampling constant tables and create */
    if ((neutral_P_k = calloc(max_alleles - 1, sizeof(double))) == NULL) {
        perror("Ack: unable to allocate sampling constant table neutral_P_k\n");
        exit(1);
    }
    /* Allocate room for sampling constant table and create */
    if ((neutral_D_k1 = calloc(max_alleles - 1, sizeof(double))) == NULL) {
        perror("Ack: unable to allocate sampling constant table neutral_D_k1\n");
        exit(1);
    }
	neutral_D_k1[0]  = 1.0/(double)M_CORRECTION_COEFF + 0.5; /* = 1/m + 1/n {n=2} */ 
	neutral_P_k [0]  = 1.0; /* = Sum(i^-1,i=1..n-1) {n=2}*/
    for (i = 1; i < max_alleles-1; i++) {
        neutral_D_k1[i] = 1.0/(double)M_CORRECTION_COEFF + 1.0/(double)(i+2); 
		neutral_P_k[i] = neutral_P_k[i - 1] + 1.0/(double)(i+1);
    }
	/* Allocate room for sampling psi table and create */
    /*if ((selected_sampling_constant = calloc(max_alleles - 1, sizeof(double))) == NULL) {
        perror("Ack: unable to allocate sampling psi table\n");
        exit(1);
    }
    for (i = 0; i < max_alleles-1; i++) {
        selected_sampling_constant[i] = 2.0 * ( gsl_sf_psi_int(i+2) + M_EULER );
    }*/
									
	/* Do global variables for integration */
	w = gsl_integration_workspace_alloc (1000);
	FPn.params = &i_params;
	FDn.params = &i_params;
	if ( M_BETA_IS_ONE && M_DIST_IS_REFLECTED ) {	// Reflected exponential
		/*printf("\nThe distribution of fitness effects is a reflected exponential");*/
		FPn.function = &Pn_d_k_integrand_exp_reflect;
		FDn.function = &Dn_d_k1_integrand_exp_reflect;
	} else if ( M_BETA_IS_ONE ) {					// Exponential shift
		printf("\nThe distribution of fitness effects is Exponential shift");
		FPn.function = &Pn_d_k_integrand_exp_shift;
		FDn.function = &Dn_d_k1_integrand_exp_shift;
	} 
	/*else if ( M_DIST_IS_REFLECTED ) {				// Reflected Gamma
		printf("\nThe distribution of fitness effects is a reflected Gamma, shape param 1/2\n");
		FPn.function = &Pn_d_k_integrand_gamma_reflect;
	} else {										// Gamma shift
		printf("\nThe distribution of fitness effects is Gamma shift, shape param 1/2\n");
		FPn.function = &Pn_d_k_integrand_gamma_shift;
	}*/
	
	/* Initial values for each parameter type */
	param_start_mins[P_THETA]=0.01;				param_start_maxs[P_THETA]=0.1;
	param_start_mins[P_REG_INTERCEPT]=0.0;		param_start_maxs[P_REG_INTERCEPT]=1.0;
	param_start_mins[P_REG_SLOPE]=0.01;			param_start_maxs[P_REG_SLOPE]=1.0;
	param_start_mins[P_TIME]=0.1;				param_start_maxs[P_TIME]=0.4;
	param_start_mins[P_OMEGA]=0.0;				param_start_maxs[P_OMEGA]=1.0;
	param_start_mins[P_GAMMA_D]=1.0;			param_start_maxs[P_GAMMA_D]=5.0;	
	param_start_mins[P_ALPHA]=0.1;				param_start_maxs[P_ALPHA]=0.6;
	param_start_mins[P_BETA_A]=0.1;				param_start_maxs[P_BETA_A]=0.5;
	param_start_mins[P_BETA_B]=0.1;				param_start_maxs[P_BETA_B]=0.4;
	param_start_mins[P_DELTA_A0]=0.0;			param_start_maxs[P_DELTA_A0]=0.4;
	param_start_mins[P_DELTA_A1]=0.4;			param_start_maxs[P_DELTA_A1]=0.7;
	param_start_mins[P_DELTA_P]=0.5;			param_start_maxs[P_DELTA_P]=0.5;
		
	/* Global bounds for each parameter type */
	param_mins[P_THETA]=0.0;		param_maxs[P_THETA]=DBL_MAX;
	param_mins[P_REG_INTERCEPT]=0.0;param_maxs[P_REG_INTERCEPT]=DBL_MAX;
	param_mins[P_REG_SLOPE]=0.0;	param_maxs[P_REG_SLOPE]=10.0;
	param_mins[P_TIME]=0.0;			param_maxs[P_TIME]=DBL_MAX;
	param_mins[P_OMEGA]=0.0;		param_maxs[P_OMEGA]=1.0;
	param_mins[P_GAMMA_D]=0.0;		param_maxs[P_GAMMA_D]=2e6; 
	param_mins[P_ALPHA]=-10.0;		param_maxs[P_ALPHA]=10.0;	
	param_mins[P_BETA_A]=0.0;		param_maxs[P_BETA_A]=DBL_MAX; 	
	param_mins[P_BETA_B]=0.0;		param_maxs[P_BETA_B]=10000; 	
	param_mins[P_DELTA_A0]=-10.0;	param_maxs[P_DELTA_A0]=5.0;
	param_mins[P_DELTA_A1]=-10.0;	param_maxs[P_DELTA_A1]=5.0;
	param_mins[P_DELTA_P]=0.0;		param_maxs[P_DELTA_P]=1.0;
	/* Fix alpha if necessary */
	if ( fixed_alpha_val != 1.0 ) {
		param_mins[P_ALPHA]       = param_maxs[P_ALPHA]       = fixed_alpha_val;
		param_start_mins[P_ALPHA] = param_start_maxs[P_ALPHA] = fixed_alpha_val;	
	}
	if( data.models[P_ALPHA] == ALPHA_FULL ) { // Take these parameters out of action, and fix them later...
		param_mins[P_ALPHA]       = param_maxs[P_ALPHA]       = 0;
		param_start_mins[P_ALPHA] = param_start_maxs[P_ALPHA] = 0;	
	}
	
	/* Allocate room for all arrays whose size varies according to the model */
	setup_model(&data,&params,&mins,&maxs,&step_sizes,&indices,param_mins,param_maxs);
	
	/* Redefine round */
	max_rounds *= param_count;
	/* Set up the signal handler to die gracefully */
#ifdef HAVE_SIGNAL
    signal(SIGINT, handle_sigint);
#endif /* HAVE_SIGNAL */
	
	/* If gamma_d is a universal parameter, create table of unique n values */
	if( data.models[P_GAMMA_D] == MODEL_UNIVERSAL ) {
		for( i=0; i<data.count; i++) {
			for(j=0; j < n_values_count;j++) { 
				if ( data.items[i].alleles == n_values[j] ) {
					j = n_values_count + 10;
				}
			}
			if ( j < n_values_count + 10 ) {
				n_values_count++;
				if( (new_n_vals = realloc(n_values, n_values_count * sizeof(int))) == NULL ) {
					free(n_values);
					fprintf(stderr, "%s: out of memory\n", progname);
					exit(1);
				}
				n_values = new_n_vals;
				n_values[n_values_count-1] = data.items[i].alleles;
			}
		}
		// Allocate room for Pn_k and Dn_k1 integrals and populate
		if ((old_univ_Pn_d_k  = calloc(n_values_count, sizeof(double *))) == NULL ||
			(univ_Pn_d_k      = calloc(n_values_count, sizeof(double *))) == NULL ||
			(old_univ_Dn_d_k1 = calloc(n_values_count, sizeof(double *))) == NULL ||
			(univ_Dn_d_k1     = calloc(n_values_count, sizeof(double *))) == NULL ) {
				perror("Ack: unable to allocate univ_Dn_d_k1\n");
				exit(1);
		}
		for(j = 0; j < n_values_count; j++) {
			univ_Pn_d_k[j]      = calloc(2, sizeof(double));
			old_univ_Pn_d_k[j]  = calloc(2, sizeof(double));
			univ_Dn_d_k1[j]     = calloc(2, sizeof(double));
			old_univ_Dn_d_k1[j] = calloc(2, sizeof(double));
		}
	}

	/* The simulated annealing */
    fflush(stderr);
    initialise_params(&data, &params, &step_sizes, param_start_mins, param_start_maxs );
	no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &data);
	output_header(&data, param_count, out);
	output_line(no_conv*tally_log_likelihood(&data), param_count, params, out);
	disp_to_screen( no_conv, tally_log_likelihood(&data), data.count, param_count, params, data.models[P_ALPHA], data.flag_list, data.flag_count);
	
	/* permutation*/
	if( num_permuts > 0) {
	    flags_original	 = calloc(data.count, sizeof(int));
		flags_randomised = calloc(data.count, sizeof(int));	
		for(i=0; i<data.count; i++)
			flags_original[i] = flags_randomised[i] = data.items[i].flag;
		
		for(i=0; i < num_permuts; i++) {
			printf("\npermutation %d of %d",i+1,num_permuts);
			gsl_ran_shuffle(my_gsl_rng, flags_randomised, data.count, sizeof(int));
			for(j=0; j<data.count; j++) {
				data.items[j].flag = flags_randomised[j];
				flags_randomised[j] = flags_original[j];
			}
			initialise_params(&data, &params, &step_sizes,param_start_mins, param_start_maxs );
			no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &data);
			output_line(no_conv*tally_log_likelihood(&data), param_count, params, out);
			
			/*Repeat each condition as a check - How often are we failing to get the correct answer??*/ 
			initialise_params(&data, &params, &step_sizes, param_start_mins, param_start_maxs );
			no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &data);
			output_line(no_conv*tally_log_likelihood(&data), param_count, params, out);
		}			
	} else if( num_permuts < 0) { // Bootstrap with replacement
		
		num_permuts = -num_permuts;
		
		/* Initialise boot_data dataset and class sizes */
		boot_data.count		 = data.count;
		boot_data.flag_count = data.flag_count;
		boot_data.flag_list  = calloc(data.flag_count, sizeof(int)); 
		boot_data.items      = calloc(data.count, sizeof(struct item));
		class_size = calloc(data.flag_count, sizeof(int));
		class_list = calloc(data.flag_count,sizeof(int*));
		for(i=0; i < 5; i++)
			boot_data.models[i] = data.models[i];
		for(i=0; i < data.flag_count; i++) {
			boot_data.flag_list[i] = data.flag_list[i];
			class_size[i] = 0;
			class_list[i] = NULL;
		}	
		for(i=0; i < data.count; i++ ) {
			class_size[data.items[i].flag]++; 
			class_list[data.items[i].flag] = realloc(class_list[data.items[i].flag], class_size[data.items[i].flag] * sizeof(int));
			class_list[data.items[i].flag][class_size[data.items[i].flag]-1] = i; 
		}
		for(i=0; i < num_permuts; i++) {
			printf("\nbootstrap %d of %d",i+1,num_permuts);
			create_bootstrapped_dataset(&boot_data, &data, class_size, class_list);
			
			/*boot_data.models[P_ALPHA] = ALPHA_FREE; 
			setup_model(&boot_data,&params,&mins,&maxs,&step_sizes,&indices,param_mins,param_maxs);
			initialise_params(&boot_data, &params, &step_sizes,param_start_mins, param_start_maxs );
			no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &boot_data);
			output_line(no_conv*tally_log_likelihood(&boot_data), param_count, params, out);
			
			initialise_params(&boot_data, &params, &step_sizes, param_start_mins, param_start_maxs );
			no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &boot_data);
			output_line(no_conv*tally_log_likelihood(&boot_data), param_count, params, out);

			boot_data.models[P_ALPHA] = ALPHA_MANY;
			setup_model(&boot_data,&params,&mins,&maxs,&step_sizes,&indices,param_mins,param_maxs);*/
			initialise_params(&boot_data, &params, &step_sizes,param_start_mins, param_start_maxs );
			no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &boot_data);
			output_line(no_conv*tally_log_likelihood(&boot_data), param_count, params, out);
			
			initialise_params(&boot_data, &params, &step_sizes, param_start_mins, param_start_maxs );
			no_conv = mcmc(t0, dt, tscale, max_rounds, param_count,params, mins, maxs, step_sizes, indices, likelihood_func, &boot_data);
			output_line(no_conv*tally_log_likelihood(&boot_data), param_count, params, out);
		}
	}
			
	if(approx_ctr>0){
		printf("\n (Kummer's U crashed %d times)\n",approx_ctr);
	}
    if (fclose(out) < 0) {
        fprintf(stderr, "%s: unable to close output file: %s\n", progname, strerror(errno));
        exit(1);
    }
    exit(0);
}